﻿using System.Collections.Generic;
using System.Collections;
using System.Reflection;
using System;
using System.Text;
using System.IO;
using System.Linq;
using System.Data;
using System.Diagnostics;

namespace VisTool
{
    public enum PathState { // Not only path state, but entire state of the expression. Important for Intellisense to make suggestions.
        Init, // Beginning of formula or element in expression list. Allowed: "Init" and what is allowed after "Init".
        Not,  // Beginning of expression after an Init.  Allowed: "Not" and what is allowed after "Not".
        Unary, // Beginning of expression after Not.  Allowed: +, - and the beginning of a primary.
        ExpectPrimary, // After a unary operator (+, -). Allowed: A primary, e.g. "(", constant, prefix, identifier.
        Rows, // Initially in the Rows formula.
        Unknown, // After an unrecognized or cut identifier at the start of the path. 
        Forms, // After the prefix "Forms". Allowed: dot+form name (short file name).
        Control, // After a reference to a control. Allowed: dot+field (may have a table prefix), dot+aggregate, 
        // dot+property, dot+relation, dot+template, dot+function, [ ], operator, EOL. 
        Property, // After a property name, indirect property or Param reference. Allowed: [ ], dot+property {indirect reference}, operator, EOL.
        // Indirect, // After an indirect property reference, e.g. Me ! Target. Allowed: [ ], operator, EOL
        Function, // After a reference to a function. Allowed: ( )
        Map, // After the prefix "Map": Allowed: table.
        RecordSet, // After a table or join expression. Allowed: dot+relation (maybe prefixed), Where, Group By, etc., EOL.
        AscDesc, // After an order-by expression in a Rows or Find formula.
        Record, // After a record reference (me>-relation). Allowed: dot+field (maybe prefixed), dot+aggregate, dot+relation.
        FieldPrefix, // After a table name. Allowed: dot+field
        RelationPrefix, // After join and relationName. Allowed: dot+relation
        RecordPrefix, // After >- tableName. Allowed: dot+ manyOneRelation.
        Field, // After a reference to a field. Allowed: operator, EOL.
        Aggregate, // After an aggregate name (Sum, Count, etc.). Allowed: ( ).
        Templates, // After the prefix "Templates". Allowed: form.
        FormTpl, // After Templates ! form. Allowed: template name or property name.
        Template, // After Templates ! form ! template. Allowed: property name, Index, Last, [i].

        Find, // After Find. Allowed: Template name in this form.
        FindTpl, // After Find tplName. Allowed: On.
        FindOn, // After Find ... On key = key. Allowed: Order By, EOL.
        BundleSource, // After B! Bundle( . Allowed: Find properties that refer to B.
        System, // After the System prefix. Allowed: system function.
        Value, // After another kind of operand, e.g. 123, "abc", a+b, (a+b), Sin(a), Color.Red. Allowed: binary operator, ), ], EOL 
        FreeText, // Inside a comment or an unfinished string, e.g. "abc. Allowed: any char. 
        AfterError, // An error before this point makes suggestions impossible. 
        Enum, // After an enumerator type, e.g. Color
        SystemVariable, // After a system variable, e.g. MouseX.
        Else, // After If (). Allowed: Else, and as ExpectStatement.
        ExpectStatement, // At the beginning of a statement. Allowed: For, If, Function, a = b.
        EndStatement, // After a statement. Allowed: Default, ; or end-of-line.
        ExpectSymbol // When only a specific token is allowed, e.g. ( after For or ident after For (.  
    }

    /* The operands in formulas are constants and paths. A path specifies how to walk from Me to the target.
     * Information about the current path is kept in PathInfo pi. However, after the path the walk has to be reset: Tpl, RowTpl and ForeignForm.   
     * Path(out ...) does this. The Parser can use the same pi throughout the formula. LeftNode links the parts together. 
     * In the path something may have a parameter list or a subscript []. Here the Parser stacks the current pi in old, and creates a new one. 
     */
    public class PathInfo {
        public PathState state;
        
        public string foreignForm; // Null unless the path refers to another form than the one where the formula is.
        public Template Tpl = null; // The template that the path has reached. Initially the template being compiled.
        public Template RowTpl = null; // The template where Tpl's data rows are defined. Tpl if there is no Rows.
        public Template block = null; // A template that cannot be used now. Used in Find target On exp1... = exp2... to ensure that exp2 doesn't use target, etc. 
        public TplProperty tplProp; // The property being referenced now. Used by a findProperty reference to suggest properties.
        public Expression leftNode=null; // The path becomes a chain of expressions. This is the previous expression.

        public Token dot=null; // The dot operator after the last ident (pi.ident). Null when no dot-operator and at path start. 
        // Make it null after an accepted ident?? Will simplify Intellisense (method NoDot) and PropertyGridLine, InsertItmes
        public Token la = null; // The look-ahead token (la) at return. Intellisense needs it to check for cutTokens.
        public Token tblPrefix=null; // A token used as prefix, i.e. a table name or relation name. Null if no prefix token.
        public string toTableName = null; // The table name where a record path ends, e.g. "Patient" for Me >- Patient as well as for Me >- Spouse.
        // After a tblPrefix, toTableName is important for intellisense.
        public Token ident=null; // The operand name in the last step, e.g. field name, relation name, control name. 
        // After an operator and a path, the state will normally be Value. When the last identifier isn't OK, it remains Control, etc. to help Intellisense.
        public Type operandType; // Defined for property, field, aggregate and value: int, double, string, dateTime. Not used.
        public Stack<string> brackets = new Stack<string>(); // A list of closing brackets, e.g. ")", "]". Allows Intellisense to propose closing brackets.
        public bool inTailExpr = false; // True when we compile expressions in query tail (Where, etc.). Generates SQL snippets plus expressions.
        //public bool inTailuVis = false; // True when we compile expressions in query tail that are uVis stuff (txtA[7]!Top, MouseX, etc.). Doesn't allow field refs.
        public bool allowSQLkeyword = false; // True when Where, etc. is allowed in addition to what the path state allows. Only to help Intellisense.
        public string aggregateName = null; // Holds "Max", "First" etc. when inside an aggregate call. Tells Parser whether fields may be referenced without Max( ), etc.
        public RecordRefExpr tempRecordRefExpr; // ??

        // A list of relations for a sequence of join-one steps, e.g. me >- relName_1 >- relName_2.
        // Don't make a join for this until it accesses a field, e.g. me >- relName_1.field
        public List<VisTool.Relationship> JoinOne = new List<Relationship>();

        public PathInfo(Template tpl, PathState init) {
            this.state = init;
            this.Tpl = tpl; this.RowTpl = tpl.RowTpl;
        }
    }

    public partial class Parser {
        bool CheckStep() { // Updates pi and returns true if this is a legal step. Returns false if this is not a legal step.
            if (pi.Tpl == null) 
                pi.Tpl = tpl; // Occurs?

            switch (pi.state) {
            case PathState.Rows:
                if (IsTemplate()) { Expect(TkKind.EOL); return true; }
                if (IsSet()) { pi.state = PathState.RecordSet; return true; }
                return false;
            //case PathState.ParentTemplate:
            //    if (IsRelation()) { pi.state = PathState.RecordSet; return true; }
            //    return false;
            case PathState.Forms:
                if (IsDot(pi.dot, ".", "!") && this.IsForm()) {
                    pi.state = PathState.Control; return true; // pi.foreignForm set by IsForm. 
                }
                return false;
            case PathState.Unknown: // An identifier has been read and assumed to be the start of a path.
                // IsA: Fields, relations and aggregate functions.
                // IsB: Properties, templates and control joins.
                // IsC: Tables.
                // System functions may have the same name as other things, e.g. Left(s, 10) and Me ! Left. 
                // We check for system functions first, including checking for ( ). In this way we can avoid checking that everything else is not followed by ( )..
                if (IsDot(pi.dot, "!")) return IsSystemFunction() || IsB() || IsA() || IsC() || IsEnum() || isSystemVariable();
                return IsSystemFunction() || IsA() || IsB() || IsC() || IsEnum() || isSystemVariable();
            case PathState.Control:
                if (IsDot(pi.dot, "!")) return IsB() || IsA();
                return IsA() || IsB();
            case PathState.Property: // Indirect reference
                if (IsDot(pi.dot, ".", "!") && IsIndirect()) return true; //{ pi.state = PathState.Indirect; return true; }
                return false;
            case PathState.Map:
                if (IsDot(pi.dot, ".", "!") && IsSet()) { pi.state = PathState.RecordSet; return true; }
                return false;
            case PathState.RecordSet:
                return IsRelation();
            case PathState.Record:
                if (IsRelationField()) { pi.state = PathState.Field; return true; }
                //if (IsRecordOneRelation()) return true;
                //if (IsControlJoin()) { pi.state = PathState.Control; return true; }
                return false;
            case PathState.Templates:
                if (IsDot(pi.dot, ".", "!") && IsForm()) {
                    pi.state = PathState.FormTpl; return true;
                }
                return false;
            case PathState.FormTpl:
                if (IsDot(pi.dot, ".", "!") && IsTemplate()) { pi.state = PathState.Template; return true; }
                goto case PathState.Template;
            case PathState.Template:
                TplProperty tplProp = null;
                var componentInfo = pi.Tpl.Info;
                string formula = null;
                if (!IsDot(pi.dot, ".", "!")) return false;
                // Check for design property, value property event handlers (with a formula), or function property (without a formula).
                PropInfo propInfo;
                tplProp = TplProperty.GetTemplateProperty(tpl, pi.ident.val, out propInfo);
                if (tplProp != null) formula = tplProp.Formula;
                pi.leftNode = new FormulaExpr(pi.Tpl.FormTpl.Name, pi.Tpl.Name, pi.ident.val, formula);
                pi.state = PathState.Value; 
                return true;
            case PathState.System:
                return (IsDot(pi.dot, ".", "!")) && (IsSystemFunction() || IsEnum() || isSystemVariable());
            default: // Covers terminal states such as Field and Indirect.
                return false;
            }
        }

        void CompileTarget(Template target) { // Compile pi.RowTpl when neded. Used at cross references.
            if (target.RowsCompiled || target.propRows.Formula == null) return; // StringReader rejects null.

            // The formula may consist of just a comment. Parser handles this.
            Parser parser = new Parser(target, new StringReader(target.propRows.Formula), target.propRows, false);
            parser.rowsFormula(target.propRows); // Sets tpl.Rows. Catches errors.
            target.RowsCompiled = true; pi.Tpl.RowsCompiled = true;
            //Debug.Print("Prepared " + tpl.Name + " RowTpl: " + (pi.RowTpl == null ? "null" : pi.RowTpl.Name));
        }

        bool IsA() {
            if (IsField()) { pi.state = PathState.Field; return true; }
            //if (IsRecordOneRelation()) { pi.state = PathState.Record; return true; }
            //if (IsRelation()) { pi.state = PathState.RecordSet; return true; } // Only allowed in Rows.

            if (IsAggregate()) { pi.state = PathState.Aggregate; return true; }
            return false;
        }

        bool IsB() {
            if (IsTemplate()) { pi.state = PathState.Control; return true; }
            Type returnType;
            if (IsProperty(out returnType)) { // Return type is the final result. Not used.
                return true;
            }
            return false;
        }

        bool IsC() {
            if (IsTableName(pi.ident.val) != null)
                    error("Use the Rows property to access tables", token, true);
            return false;
        }

        bool isSystemVariable() {
            SystemFieldInfo fieldInfo = null; // SystemItems.CheckVariable(pi.ident.val);
            if (VisEngine.SystemFieldByName.TryGetValue(pi.ident.val, out fieldInfo)) {
                pi.leftNode = new SystemVariableExpr(fieldInfo.PropertySetter);
                pi.state = PathState.Value; 
                return true;
            }
            return false;
        }

        bool IsEnum() { // E.g. Color . enumId
            SystemMethodInfo m;
            if (!VisEngine.SystemMethodByName.TryGetValue(pi.ident.val, out m)) return false;
            if (m.attr.itemType != VisSystemAtt.ItemType.UvisEnum) return false;
            pi.state = PathState.Enum; 
            if (pi.ident.next == null || (pi.ident.next.val != "." && pi.ident.next.val != "!")) {
                error("Expects a dot", la); return true;
            }

            // We have: Color dot
            Get(); // Get dot
            pi.dot = token; 
            if (La.kind != TkKind.Ident) { error("Expects an identifier", la); return true; }
            object[] parameters = new object[] { la.val, null }; // 
            object value = m.PropertySetter.Invoke(null, parameters); // EnumMethod.Invoke(null, parameters);
            if (value == null) { error(la.val + " is not a possible " + pi.ident.val, la); return true; }
            pi.state = PathState.Value; 
            pi.leftNode = new EnumExpr(value);
            Get();
            return true;
        }

        bool IsSet() { // If it is a Set (a data table), creates a RecordSetExpr.
            if (pi.inTailExpr) return false; // Sets not allowed in query tail.
            Table table = IsTableName(pi.ident.val);
            if (table !=null) {
                TableExpr tableExpr = new TableExpr(pi.ident.val, table);
                RecordSetExpr rsExpr = new RecordSetExpr(tpl.FormTpl.Name + "_" + tpl.FormTpl.cloneID.ToString() + "_" + tpl.Name, tpl);
                rsExpr.JoinChain = tableExpr;
                Join newJoin = new Join(table.Name, table, null, true);
                rsExpr.joinList.Add(newJoin);
                rsExpr.primaryTable = rsExpr.lastTable = pi.ident.val;
                pi.leftNode = rsExpr;
                pi.RowTpl.RowsExpr = rsExpr; // For Intellisense.
                return true;
            }
            return false;
        }

        bool IsSystemFunction() {
            string functionName = pi.ident.val;
            SystemMethodInfo f=null;
            if (VisEngine.SystemMethodByName.TryGetValue(functionName, out f)) {
                pi.state = PathState.Function; // Helps Intellisense.
                if (La.val != "(") return false; // The name could also be a property, e.g.Left or design prop date.
                if (!IsStatement && f.attr.itemType == VisSystemAtt.ItemType.Statement) {
                    error("Cannot be used in expressions", pi.ident); throw new LoggedException(); 
                }
                if (f.attr.itemType == VisSystemAtt.ItemType.SystemMethod || f.attr.itemType == VisSystemAtt.ItemType.Statement) {
                    MethodInfo m = typeof(VisTool.SystemItems).GetMethod(f.Name); // f.Name has the correct capitalization.
                    functionCall(m, f.attr.SQL);
                    return true;
                }
            }
            return false;
        }

        bool IsIndirect() {
            // E.g.: StartControl ! Color or StartControl.tblMedOrder.medName, where StartControl: me-=MedOrder
            // Assumes that pi.leftNode is a property reference, e.g. StartControl
            // Assumes: pi.dot is the dot-operator at the beginning of the path step: . or !
            // Assumes: pi.ident is the potential set name.
            Token tLa = La; // Must be a .-token if there is a prefix.
            ResetPeek();
            Token tP1 = Peek(); // Should be the identifier token after the prefix, e.g. medName
            string setName = pi.ident.val; // E.g. Color or tblMedOrder
            string fieldName = tP1.val;
            pi.tblPrefix = null;

            if (IsDot(tLa, ".")) { // && tP1.kind==tk.Ident
                // The dot operators and idents are ok. This looks like a prefixed field name. Will be checked at run time.
                pi.tblPrefix = pi.ident; pi.toTableName = pi.ident.val; // Cannot be sure, but gives some Intellisense help.
                //pi.state = PathState.FieldPrefix; // Path engine sets PathState to Indirect.
                dotOps(); // Sets pi.dot
                Expect(TkKind.Ident); // Get the field name
                pi.leftNode = new IndirectPrefixedField(pi.leftNode, setName, fieldName, tP1);
                return true;
            }
            pi.leftNode = new IndirectPropOrField(pi.leftNode, pi.dot.val, pi.ident.val, pi.ident); // Will be checked at run time.
            return true;
        }

        bool IsField() { // field = [ . ] [ setName . ] fieldName
            // E.g. tblName.fieldName, me.tblName.fieldName, me >- relName.tblName.fieldName
            // Assumes: pi.dot is the dot-operator at the beginning of the path step. Null of no dot operator.
            // Assumes: pi.ident = token, the potential setName or fieldName. 
            if (pi.RowTpl == null) return false; // No fields
            if (pi.leftNode is TemplateRefExpr) { // Reference to a field in another template. 
                // Make sure the rows are compiled. Otherwise the fields and Group By cannot be seen.
                PathInfo oldPi = pi; pi = new PathInfo(pi.RowTpl, PathState.Rows);
                CompileTarget(pi.RowTpl); // pi.RowTpl must match TemplateRef.
                pi = oldPi;
            }
            Token initialDot = pi.dot;
            if (!(initialDot == null || IsDot(initialDot, ".", "!"))) return false; // The initial dot operator must be null, dot or bang.

            ResetPeek();
            Token fieldToken = Peek(); // The fieldName token assuming there is a prefix..
            string prefix = pi.ident.val; // assuming there is a prefix.
            Join join = null;
            string fieldName = fieldToken.val; // assuming there is a prefix.
            Template target = pi.Tpl;
            pi.tblPrefix = null;

            if (IsDot(la, ".", "!")) { // The assumed dot after the prefix must be . or !  Don't care about cut idents.
            if (!(pi.leftNode is RecordSetExpr
                && (pi.leftNode as RecordSetExpr).HasPrefix(prefix, out join, pi))) // Look for relation in the path.
                if (target.RowTpl.RowsExpr != null) target.RowTpl.RowsExpr.HasPrefix(prefix, out join, pi);
            }

            if (join != null) {
                // The relationName is ok. This is a prefix.
                pi.tblPrefix = pi.ident;
                pi.state = PathState.FieldPrefix; // Support Intellisense.
                pi.toTableName = (join.relation == null ? join.prefix : join.relation.ToTableName); // Either the left-side table name or the target of the join.

                if (pi.inTailExpr && pi.dot != null) { error("Inside queries, a field cannot start with dot or bang", pi.ident); throw new LoggedException(); }
                Expect("."); pi.dot = token;
                Expect(TkKind.Ident); pi.ident = token; pi.dot = null;
                if (!IsFieldName(join.table, fieldName)) { error(fieldName + " is not a field in " + prefix, token); throw new LoggedException(); }
                if (pi.inTailExpr) { // Set brackets around the prefix and field name in sql tails to avoid conflicts with Sql keywords, e.g. End.
                    pi.tblPrefix.sqlBracket = Token.SqlBracket.left; token.sqlBracket = Token.SqlBracket.right;
                }
            } else {

                // There is no table prefix (setName).
                fieldName = prefix; fieldToken = pi.ident; // For error reporting.
                target = pi.Tpl;
                int fieldCount = 0;
                if (!(pi.leftNode is RecordSetExpr
                    && (fieldCount = (pi.leftNode as RecordSetExpr).HasField(fieldName, out join)) == 0)) // Look for field in the path.
                    if (target.RowTpl.RowsExpr != null) {
                        fieldCount = target.RowTpl.RowsExpr.HasField(fieldName, out join);
                    }
                if (fieldCount == 0) return false;
                if (fieldCount > 1) { error("Field name occurs in more than one table", token); throw new LoggedException(); }

                // The field name is okay.
                prefix = join.prefix;
                if (pi.inTailExpr && pi.dot != null) { error("Inside queries, a field cannot start with dot or bang", pi.ident); throw new LoggedException(); }
                if (pi.inTailExpr) pi.ident.sqlBracket = (Token.SqlBracket.left | Token.SqlBracket.right); // Set brackets around the field name in sql tails.
            }

            // Common part ----------------------------------------
            if (initialDot != null) initialDot.val = "."; // The initial dot-operator must be "."
            Field field = new Field() { fieldName = fieldName, prefix = prefix, tableName = join.table.Name };
            if (pi.foreignForm == null) { // Add the field to the Rows. Not possible for a foreign form.
                if (pi.aggregateName == null) { // pi.state != PathState.Aggregate
                    if (target.RowTpl.RowsExpr.NeedsGroupBy(field)) {
                        error("Field " + fieldToken.val + " must be Group By, First( ), etc.", fieldToken); throw new LoggedException();
                    }
                    if (!pi.inTailExpr) { // Don't add fields inside query tail. 
                        target.RowTpl.RowsExpr.AddField(field);
                    }
                }

                //if (pi.aggregateName == null && pi.inTailExpr && target.Rows.NeedsGroupBy(field)) {
                //    error("Field " + fieldToken.val + " must be Group By, First( ), etc.", fieldToken); throw new LoggedException();
                //}
                //if (!pi.inTailExpr) { // Don't add fields inside query tail. 
                //    if (target.Rows.NeedsGroupBy(field)) {
                //        error("Field " + fieldToken.val + " must be Group By, First( ), etc.", fieldToken); throw new LoggedException();
                //    }
                //    target.Rows.AddField(field);
                //}
            }
            pi.leftNode = new FieldExpr(pi.leftNode, join, fieldName, target, 0, token);
            return true;
        }

        bool IsRelationField() { // After:  >- relation; check for:  . fieldName
            // To be called in PathState.Record. Doesn't need to look for ancestors. Just look for the fields in relation.

            // Assumes: pi.dot is the dot-operator before the potential fieldName.
            // Assumes: pi.ident is token, the potential fieldName.
            string fieldName = pi.ident.val;
            Join join;
            string prefix;

            if (!IsDot(pi.dot, ".", "!")) return false;
            // The initial dot operator is dot or bang.
            // Check that the field is in the last relation. For me >-relA >-relB.field, the field must be in the table that relB points to.
            Relationship relation = pi.JoinOne.Last();
            prefix = relation.RelationName;
            if (!IsFieldName(relation.ToTable, fieldName)) {
                error(fieldName + " is not a field in " + relation.RelationName, token); throw new LoggedException();
            }
            if (pi.foreignForm != null) { // Add the field to the Rows. Not possible for a foreign form.
                error(fieldName + " is in another Form. It cannot be referenced through a join-one relation", token); throw new LoggedException();
            }
            pi.dot.val = "."; // The initial dot-operator must be ".".

            // A join-one relation doesn't make a join in itself. It just provides keys for a record in a foreign table, e.g. me>-relA.
            // However, when a field in the record is addressed, the relA join must be added to Rows.
            // Check whether the join exists already in Rows.JoinList, because another formula addressed a field in the record.
            // In general there may be several join-one steps. Ensure that all of them get a join.
            join = null;
            foreach (Relationship rel in pi.JoinOne) {
                foreach (Join j in pi.RowTpl.RowsExpr.joinList) 
                    if (j.prefix.ToUpper() == rel.RelationName.ToUpper()) {
                        join = j;
                        j.leftJoin = true;
                        break;
                }
                // Find target Rows, e.g. pi.RowTpl.Rows.
                //RelationExpr relExpr = pi.RowTpl.Rows.JoinChain as RelationExpr;
                //while (relExpr != null) {
                //    if (relExpr.rel.RelationName.ToUpperInvariant() == rel.RelationName.ToUpperInvariant()
                //        && relExpr.rel.FromTableName.ToUpperInvariant() == rel.FromTableName.ToUpperInvariant() ) {  // The same relation
                //        exists = true;
                //        relExpr.Op = Operator.LJ; // The Rows join may have inner join, but a JoinOne field enforces left join.
                //        break;
                //    }
                //    relExpr = relExpr.next as RelationExpr;
                //}
                if (join==null) {
                    pi.RowTpl.RowsExpr.JoinChain = new RelationExpr(pi.RowTpl.RowsExpr.JoinChain, Operator.JO, relation, token);
                    //pi.RowTpl.Rows.AddSetName(rel.ToTableName);
                    join = new Join(rel.RelationName, rel.ToTable, rel, true);
                    pi.RowTpl.RowsExpr.joinList.Add(join);
                }
            }
            Field field = new Field() { fieldName = fieldName, prefix = prefix, tableName= join.table.Name };
            pi.RowTpl.RowsExpr.AddField(field);
            if (pi.RowTpl.RowsExpr.RecordSet != null && pi.RowTpl.RowsExpr.exprGroupby.Count > 0) {
                error("IsRelationField - should be Group By: " + pi.RowTpl.Name + "!" + VisEngine.curTplProp.Name + " ." + fieldName, token); throw new LoggedException();
            }
            pi.leftNode = new FieldExpr(pi.leftNode, join, fieldName, pi.RowTpl, 0, token); // 0 = parentCount
            return true;
        }

        bool IsRelation() { // relation =  ( -< | =< | >- | >= ) [ setName . ] relationName
            // E.g.: tblName.relName, me.tblName.fieldName, me >- relName.tblName.fieldName
            // Assumes: pi.dot is the dot-operator at the beginning of the path step. Null initially.
            // Assumes: pi.ident is token, the potential setName or relationName. 
            if (pi.inTailExpr) return false; // Relations not allowed in query tail.
            if (!IsDot(pi.dot, "-<", "=<", ">-", ">=")) return false;

            Token initialDot = pi.dot;
            ResetPeek();
            Token relToken = Peek(); // The relation token assuming there is a prefix..
            string prefix = pi.ident.val; // assuming there is a prefix.
            string relName = relToken.val;
            Join join=null;
            Relationship relation = null;
            pi.tblPrefix = null;
            RecordSetExpr expr = pi.leftNode as RecordSetExpr;

            if ( expr == null || !expr.HasPrefix(prefix, out join, pi)) { 
                // If not in the path, Check in RowTpl.
                if (pi.RowTpl != null && pi.RowTpl.RowsExpr != null) pi.RowTpl.RowsExpr.HasPrefix(prefix, out join, pi);
            }
            if (join != null) { // The relationName is ok. This is a prefix.
                pi.tblPrefix = pi.ident; // pi.toTableName = relation.ToTableName;
                pi.state = PathState.RelationPrefix; // Support Intellisense when there is a prefix but no relation name.

                Expect(".", pi.tblPrefix.val + " is already a relation in the path"); pi.dot = token;
                Expect(TkKind.Ident); pi.ident = token; pi.dot = null;
                if (!IsRelationName(relName, prefix, out relation)) { error(relName + " is not a relation in " + prefix, token); throw new LoggedException(); }
            } else {

                // There is no prefix.
                relName = prefix;
                int relationCount = 0;
                if (expr != null) relationCount = expr.HasRelation(relName, out relation);
                if (relationCount > 1) { error("Relation name occurs in more than one table", token); return false; }
                if (relationCount == 0 && pi.RowTpl != null) 
                    relationCount = pi.RowTpl.RowsExpr.HasRelation(relName, out relation);
                if (relationCount > 1) { error("Relation name occurs in more than one table", token); return false; }
                if (relationCount == 0) return false;
            }

            // Common part ----------------------------------------------
            string strDot = initialDot.val;
            CardinalityType car = relation.Cardinality;
            Operator op = Operator.LJ;
            bool isLeftJoin = true;
            switch (strDot) {
            case "-<": if (car != CardinalityType.onemany) strDot = ">-"; break;
            case "=<": if (car != CardinalityType.onemany) strDot = ">="; op = Operator.IJ; isLeftJoin = false; break;
            case ">-": if (car != CardinalityType.manyone) strDot = "-<"; break;
            case ">=": if (car != CardinalityType.manyone) strDot = "=<"; op = Operator.IJ; isLeftJoin = false; break;
            }
            initialDot.val = strDot;

            Expression next = null;
            RecordSetExpr rsExpr = pi.leftNode as RecordSetExpr;
            if (rsExpr == null) 
                rsExpr = new RecordSetExpr(tpl.FormTpl.Name + "_" + tpl.FormTpl.cloneID.ToString() + "_" + tpl.Name, tpl);
            else next = rsExpr.JoinChain;

        if (!(pi.leftNode is RecordSetExpr)) {
                error("Cannot join with the left side", token); throw new LoggedException();
            }
            RelationExpr relationExpr = new RelationExpr(next, op, relation, token);
            rsExpr.lastTable = relation.ToTableName;
            if (car==CardinalityType.onemany) rsExpr.primaryTable = relation.ToTableName; // Primary table is the many table.
            rsExpr.JoinChain = relationExpr;
            Join newJoin = new Join(relation.RelationName, relation.ToTable, relation, isLeftJoin);
            rsExpr.joinList.Add(newJoin);
            pi.leftNode = rsExpr;
            pi.state = PathState.RecordSet; 
            pi.dot = null; // pi.ident set above.
            pi.toTableName = relation.ToTableName;
            return true;
        }

        //bool IsRecordOneRelation() { // record = [ component ] >- [ setName . ] relationName
        // Deleted 24-04-17 by SL. Too hard to understand, particularly when field or relation prefixes were needed,
        // Prefixes didn't work. Work around: Add the join in the rows formula.

        //    // E.g. >- tblName.relName, Me >- relName, box >- tblName.relname
        //    // Assumes: pi.dot is the dot-operator at the beginning of the path step. 
        //    // Assumes: pi.ident is the potential set name.
        //    if (pi.inTailExpr) return false; // Relations not allowed in query tail.
        //    if (pi.RowTpl == null) return false; // No relations
        //    if (!IsDot(pi.dot, "-<", "=<", ">-", ">=")) return false;

        //    Token initialDot = pi.dot;
        //    ResetPeek();
        //    Token relToken = Peek(); // The relation token assuming there is a prefix..
        //    string prefix = pi.ident.val; // assuming there is a prefix.
        //    string relName = relToken.val;
        //    //Join join = null;
        //    Relationship relation=null;
        //    Template target = pi.Tpl;
        //    pi.tblPrefix = null;

        //    if (pi.tblPrefix != null) {
        //        // The setname is ok. This is a prefix.
        //        Expect("."); pi.dot = token;
        //        Expect(TkKind.Ident); pi.ident = token; pi.dot = null;
        //        if (!IsRelationName(relName, prefix, out relation)) { error(relName + " is not a relation in " + prefix, token); throw new LoggedException(); }
        //    } else {

        //        // There is no prefix.
        //        relName = prefix;
        //        target = pi.Tpl;
        //        if (target.RowsExpr != null) {
        //            int relationCount = target.RowsExpr.HasRelation(relName, out relation);
        //            if (relationCount > 1) { error("Relation name occurs in more than one table", token); throw new LoggedException(); }
        //            if (relationCount == 0) return false; // Not a relation.
        //        }

        //        // The relation name is okay.
        //        relName = pi.ident.val;
        //    }

        //    // Common part ----------------------------------------
        //    pi.state = PathState.Record; 
        //    pi.toTableName = relation.ToTableName;
        //    string strDot = initialDot.val;
        //    CardinalityType car = relation.Cardinality;
        //    if (car != CardinalityType.manyone) { error("The relationship must be many-to-one", token); throw new LoggedException(); }
        //    if (initialDot != null) initialDot.val = ">-";
        //    if (pi.foreignForm != null) { error("Cannot join with Rows in another Form", token); throw new LoggedException(); }

        //    foreach (Field foreignKey in relation.FromKeyFields) {
        //        pi.RowTpl.RowsExpr.AddField(foreignKey);
        //        if (pi.RowTpl.RowsExpr.RecordSet != null && pi.RowTpl.RowsExpr.exprGroupby.Count > 0) {
        //            error("Should be Group By: " + pi.RowTpl.Name + "!" + VisEngine.curTplProp.Name + " ." + foreignKey.fieldName, token);
        //            throw new LoggedException();
        //        }
        //    }
        //    pi.tempRecordRefExpr = new RecordRefExpr(pi.leftNode, relation, token);
        //    pi.JoinOne.Add(relation);
        //    return true;
        //}

        bool IsFieldName(Table tbl, string fieldName) {
            return tbl.Columns.ContainsKey(fieldName);
        }

        bool IsAggregate( ) {
            //  The parameters are read at end of Primary. The syntax slightly 
            if ((pi.dot == null || IsDot(pi.dot, "!", ".")) && (pi.ident.keyType & KeyType.Aggregate) > 0) {
                if (pi.foreignForm != null) { // Not possible for a foreign form.
                    error("Not possible to aggregate data from another form. Use a Designer property in the other form.", token); 
                    throw new LoggedException();
                }
                if (pi.aggregateName != null) { 
                    error("Not possible to have nested aggregate functions.", token);
                    return false;
                }
                if (pi.RowTpl == null || pi.RowTpl.RowsExpr == null || pi.RowTpl.RowsExpr.exprGroupby.Count == 0) { 
                    error(pi.ident.val + " needs Rows and Group By", token); throw new LoggedException(); 
                }
                pi.aggregateName = pi.ident.val;
                if (pi.inTailExpr) return true; // Aggregate embedded in SQL.
                // The AggregateCall generates the sql string at run time. 
                pi.leftNode = new AggregateCall (pi.leftNode, "Aggr" + (pi.RowTpl.RowsExpr.aggregateFields.Count + 1).ToString(), token);
                return true;
            }
            return false;
        }

        bool IsRelationName(string relationName, string prefix, out VisTool.Relationship relation) {
            relation = VisEngine.DataMaps[0].getRelationship(relationName, prefix);
            return relation != null;
        }

        bool IsProperty(out Type resultType) {
            resultType = null; // Not used
            if (pi.dot != null && !IsDot(pi.dot, "!", ".")) return false;
            Template currentTpl = tpl;  
            if (pi.state == PathState.Control) currentTpl = pi.Tpl;
            
            string propNameUpper = pi.ident.val.ToUpperInvariant();
            switch (propNameUpper) {
            case "INDEX": // Don't enforce a leftNode MeExpr to allow the special scope rule for Index inside a subscript.
                pi.leftNode = new IndexExpr(pi.leftNode, token);
                pi.state = PathState.Value; 
                return true;
            case "LAST": // Don't enforce a leftNode MeExpr to allow the special scope rule for Last inside a subscript.
                pi.leftNode = new LastExpr(pi.leftNode, token);
                pi.state = PathState.Value; 
                return true;
            case "NAME": // No reason to enforce a leftNode. NameExpr knows what to do.
                pi.leftNode = new NameExpr(pi.leftNode, token);
                pi.state = PathState.Value;
                return true;
            case "PARAM": // Param [ '['paramIndex ']' [ '['component index ']' ] ]
                Expression paramIndex = null;
                pi.state = PathState.Property; // Also before [ ] to help Intellisense. 
                if (La.val == "[") {
                    pi.brackets.Push("]"); // Stack expected closing character.
                    Get();
                    defaultExpr(ref paramIndex);
                    Expect("]");
                    pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
                    pi.state = PathState.Property;
                }
                pi.leftNode = new ParamExpr(pi.leftNode, paramIndex, token);
                if (La.val == "[") {
                    subscript(); // Sets PathState.
                }
                return true;

            default:
                MethodInfo funcPropInfo = IsFuncPropName(currentTpl, pi.ident.val);
                if (funcPropInfo != null) {
                    resultType = funcPropInfo.ReturnType;
                    pi.state = PathState.Function; // Helps Intellisense.
                    if (pi.ident.val.ToUpper() == "BUNDLE") // e.g. Me! Bundle(sourceTpl ! findProp)
                        sourceBundle(funcPropInfo);
                    else
                        functionCall(funcPropInfo, false); // Creates parameters and a FunctionCall.
                    if (resultType == typeof(VisComponent) || resultType == typeof(BundleRef)) pi.state = PathState.Control;
                    else pi.state = PathState.Value;
                    return true;
                }

                TplProperty tplProp; PropInfo propInfo;
                tplProp = TplProperty.GetTemplateProperty(currentTpl, pi.ident.val, out propInfo);
                if (tplProp == null && propInfo == null) return false; // No such property.

                pi.tplProp = tplProp;
                bool bundleRef = false; // When true, refer to the bundle rather than the component.
                ResetPeek();
                if (IsDot(la, ".", "!") && Peek().val.ToUpper() == "BUNDLE") { // e.g. findProp! Bundle [Index+1]
                    if (tplProp.startChain.val.ToUpper() != "FIND") // Can be checked without having to compile target.
                        error("The target property must have a Find formula.", la);
                    bundleRef = true; Get(); Get();
                    pi.state = PathState.Control; // Reference to a sibling component in the bundle.
                } else 
                    pi.state = PathState.Property; // May become a value or an indirect reference.
                
                if (propInfo == null) { // Designer property.
                    int propIndex = currentTpl.DesignProperties[pi.ident.val].cmpPropIndex;
                    pi.leftNode = new DesignPropRef(pi.leftNode, propIndex, tplProp, bundleRef, token);
                    resultType = typeof(object);
                    return true;
                }

                PropertyInfo guiPropInfo;
                if (propInfo is ValueInfo) { // Value property
                    guiPropInfo = (propInfo as ValueInfo).PropertySetter;
                    pi.leftNode = new ValPropertyRef(pi.leftNode, propInfo.propIndex, guiPropInfo, tplProp, bundleRef, token);
                    resultType = guiPropInfo.PropertyType;
                    return true;
                }
                resultType = null;
                return false;
            }
        }

        public MethodInfo IsFuncPropName(Template tpl, string propName) {  // out List<MethodInfo> funcPropInfo) {
            //In this step, we do not have parameter information such as the number of parameters and their types
            //this method is also used by VisParser

            FunctInfo fi;
            if (tpl.Info.FunctionByName.TryGetValue(propName, out fi)) 
                return fi.PropertySetter;
            return null;

            // Part of a solution for overloaded functions.
            //funcPropInfo = null;
            //string propNameUpper = propName.ToUpperInvariant();
            //funcPropInfo = new List<MethodInfo>();
            //PropInfo p = null;
            //tpl.Info.PropertyByName.TryGetValue(propName, out p);
            //if (p == null || p.VisAttribute.PropType != VisPropAtt.PropertyType.Method) return false;
            //MethodInfo[] ctlInstFuncProps = tpl.typeVisComponent.GetMethods(BindingFlags.Public | BindingFlags.Instance);
            //foreach (MethodInfo f in ctlInstFuncProps) {
            //    if (f.Name.ToUpperInvariant() == propNameUpper) {
            //        funcPropInfo.Add(f);
            //    }
            //}
            //return funcPropInfo.Count > 0;
        }

        bool IsForm() { // Forms ! formName [ '(' formID ')' ] [ '['formIndex ']' ]
            // Looks up the formName in VisEngine.Forms. Fast - doesn't use the file directory.
            string frmName = pi.ident.val;
            if (! IsFormName(ref frmName)) return false;
            //FormTemplate frmTpl = VisEngine.GetFormTemplate(frmName, -1); // Any clone will do.
            //if (frmTpl == null)  // Load the form, but don't instantiate it
            //    frmTpl = VisEngine.LoadForm(frmName); // Loading compiles it and gives confusing error messages.
            //if (frmTpl == null) return false;

            Expression cloneExpr = null;
            if (La.val == "(") {
                pi.brackets.Push(")"); // Stack expected closing character.
                //PathInfo old = pi; pi = PathInfo.CopyPathInfo(old); // Stack path info in order to keep the preceding path
                Get();
                defaultExpr(ref cloneExpr);
                //pi = old; // Unstack
                Expect(")");
                pi.brackets.Pop(); // Unstack only if there is a matching parenthesis
            }
            pi.leftNode = new FormsExpr(frmName, cloneExpr, token);
            //pi.Tpl = frmTpl; // Set Tpl and RowTpl for further steps, because the form has been changed
            //PrepareRowTpl();
            pi.foreignForm = frmName;
            // The formIndex is handled by checkStep in the same way as indexes after other controls.
            return true;
        }

        bool IsTemplate() { // The current template is in pi.T or a foreign form.
            Template target;
            if (pi.dot != null && !IsDot(pi.dot, ".", "!")) // There is a dot, but it is neither . nor !
                return false;
            if (pi.foreignForm != null) { // In a foreign form, we have to look up the template at run time because the foreign form may open and close
                pi.leftNode = new ForeignTplOrPropRef(pi.leftNode, pi.ident.val, token);
                return true;
            }
            // If in this form, we can optimize by providing the template reference.
            if (pi.Tpl.FormTpl.Templates.TryGetValue(pi.ident.val, out target) && !target.Ignore) {
                if (pi.block == target && pi.Tpl != target) // Reference in Find t On exp1 = exp2 from exp2 to t, etc. Allow self-ref.
                    error("A reference to " + target.Name + " must be on the other side of =", token);
                pi.Tpl = target; pi.RowTpl = target.RowTpl; 
                pi.leftNode = new TemplateRefExpr(pi.leftNode, pi.Tpl, token);
                if (pi.state == PathState.Rows) Expect(TkKind.EOL); // A reference to the main Rows.
                return true;
            }
            return false;
        }

        bool IsDot(Token t, params string[] dot) {
            if (t == null || t.kind != TkKind.Operator) return false;
            foreach (string x in dot) if (t.val == x) {
                    //  if (t.val != dot[0]) t.val = dot[0]; tplName.dirty = true; // Correct output to first join operator ??
                    return true;
                }
            return false;
        }

        Table IsTableName(string tableName) {
            string tableNameUpper = tableName.ToUpperInvariant();
            Table table=null;
            VisEngine.DataMaps[0].Tables.TryGetValue(tableName, out table);
            return table;
        }

        public static bool IsFormName(ref string frmName) {
            List<FormTemplate> clones = null;
            if (!VisEngine.Forms.TryGetValue(frmName, out clones)) return false; 
            VisEngine.FormNames.TryGetValue(frmName, out frmName); // Return the true file name. 
            return true;
        }
    }
}
