﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Text;

namespace VisTool {
    /* When parsing a formula, Intellisense can at any point list the possibilities (suggestions) that are available for continuing the text. 
     * During parsing, the parser updates the path information in pi, part of which is the pathState. Calling GetSuggestions will return a 
     * dictionary of the possible names at this point. Each name entry has an associated list of the possible meanings of this name, 
     * including a necessary or optional prefix, tool tips (explanations), etc.
     * 
     * UVis Studio utilizes it in this way: When the designer has put the cursor at a certain position in a formula and clicks Ctrl+J, 
     * Studio cuts out the part of the formula until the cursor. Next it asks the parser to compile this part and gets the final path information. 
     * It then calls ListSuggestions and shows the suggestions as a dropdown list. When the designer chooses one of the suggestions, 
     * Studio inserts it at the end of the cut-out part and adds the rest of the formula.
     * 
     * The parser needs to know that the formula doesn't end normally. For this purpose, when reading a formula in memory, the 
     * scanner terminates the chain of tokens with a token marked cutToken. 
     * On the way, the parser looks at the next token with "la" and can test whether it is a cutToken. 
     * It can also use the method version "La". It checks whether the next token is a cutToken. If so, it sets pi.la to la and throws a 
     * CutException. This aborts the parsing with sufficient information in pi for intellisense to see what can be next in the formula.
     * 
     * The parser uses this basic pattern:
     *           isA();
     *           pi.state = PathState.xyz; // Set pi information that allows Intellisense to work.
     *           if (La.val == "..." ) { // if it is cutToken, sets pi.la = la and throws CutException. Intellisense takes over.
     *               isB(); ...
     *           }
     * The scanner does the same when the parser uses Get() on a cutToken or generates an error.
     */

    public class Suggestion {
        public string name; // The item name properly capitalized.
        public string prefix; // Null if a prefix is meaningless, e.g. after Map.
        public string tblPrefix; // Null if a table prefix is meaningless, e.g. for properties or after a table prefix.
        public string dotOperator; // The correct dot-operator to use.
        public string category; // The property grid category from the annotation attribute. 
        public int position; // The position property from the annotation attribute (position within a grid category).
        public string[] functionParameters;
        public string iconName;
        public string longExplanation;
        public string shortExplanation;

        public Suggestion(string name, string prefix, string tblPrefix, string dotOperator, string category,
            string longExplanation, string shortExplanation, int position, string iconName, string[] functionParameters) {
            // If name isn't in suggestedNames, add it. Create the suggestion and add it to the item list in the suggested name.
            if (name == null) 
                name = "?";
            List<Suggestion> itemList;
            if (VisIntelliSense.possibleNames.TryGetValue(name.ToLower(), out itemList)) { // Check for duplicates
                if (itemList.Find(delegate(Suggestion s) { return s.prefix == prefix && s.tblPrefix == tblPrefix && s.dotOperator == dotOperator; }) != null)
                    return; // Return without inserting the suggestion again;
            } else { // Add an item name.
                itemList = new List<Suggestion>(5);
                VisIntelliSense.possibleNames.Add(name.ToLower(), itemList);
            }
            this.name = name;
            this.prefix = prefix;
            this.tblPrefix = tblPrefix;
            this.dotOperator = dotOperator;
            this.category = category;
            this.position = position;
            this.longExplanation = longExplanation;
            this.shortExplanation = shortExplanation;
            this.iconName = iconName;
            this.functionParameters = functionParameters;
            itemList.Add(this);
        }
    }

    public class VisIntelliSense {
        public static Dictionary<string, List<Suggestion>> possibleNames;

        public static PathInfo currentPi ;
        //public static Template selected;
        public static Template template; // pi.Tpl if not null, selectedTemplate otherwise.
        public static Template rowTpl; // selectedTemplate if it has rows, pi.RowTpl otherwise.
        public static TplProperty property;
        public static string match; // Cut ident to match in lists
        public static string expectedSymbol; // The suggested text in PathState.ExpectSymbol.

        // The suggestions are stored in a dictionary of all the possible identifier names. Each name has a list of the possible items with this name.
        // When there are more than one item in the list, the name is ambiguous and a prefix, ( ) or priority rule is needed to resolve the ambiguity.

        public static Dictionary<string, List<Suggestion>> GetSuggestions(PathInfo pi, Template meTpl, TplProperty selectedProperty, string curText) {
            // Clear previous suggestions.
            possibleNames = new Dictionary<string, List<Suggestion>>();

            currentPi = pi; // The static version will be used in the add methods (e.g. addPrefix, addProperty, etc.)
            template = meTpl;
            if (pi.Tpl != null) template = pi.Tpl; // For instance inside a template ref, inside [ ] or inside Find a On a.b = me.c
            //selected = selectedTemplate;
            //template = (pi.Tpl == null ? selectedTemplate : pi.Tpl); // template will be used in the add methods (e.g. addPrefix, addProperty, etc.)
            rowTpl = template.RowTpl;  // (selectedTemplate.RowsExpr == null ? pi.RowTpl : selectedTemplate);
            property = selectedProperty;

            //string brackets = ""; foreach (var b in pi.brackets) brackets = brackets + " " + b;
            //Debug.Print("state: " + pi.state + (pi.state == PathState.ExpectSymbol ? expectedSymbol : "") + " " + template.Name //+ " rowTpl: " + rowTpl.Name
            //    + " pi.dot: " + (pi.dot != null ? pi.dot.val : "") + " brackets: " + brackets);
            //" allowsqlkeyw:" + pi.allowSQLkeyword + " intailexpr:" + pi.inTailExpr +
            //    " dotop:" + (pi.dot == null ? "null" : pi.dot.val) + " pi.t:" + (pi.Tpl == null ? "null" : pi.Tpl.Name) +
            //    " ident:" + (pi.ident == null ? "_" : pi.ident.val) + " match:" + match + " la: " + (pi.la == null ? "" : (pi.la.val + " " + pi.la.error)));

            switch (pi.state) {
            case PathState.Init: // Beginning of formula or element in expression list. Allowed: "Init" and what is allowed after "Init".
                addKeyword("Init", "Initial value of the property. User interaction may change the value.");
                addKeyword("Find", "Find target On targetKey = MeKey ... Order By ...");
                goto case PathState.Not;
            case PathState.Not: // Beginning of expression after an Init.  Allowed: "Not", sometimes @ and what is allowed after "Not".
                addOperator("Not", "", 10);
                goto case PathState.Unary;
            case PathState.Unary: // Beginning of expression after Not.  Allowed: +, - and the beginning of a primary.
                addOperator("-  +", "", 9);
                goto case PathState.ExpectPrimary;
            case PathState.ExpectPrimary: // After a unary operator (+, -). Allowed: A primary, e.g. "(", constant, prefix, identifier.
                addOperator("(", "", 8);
                addConstant();
                addKeyword("@",
                    "Calculate the expression after @ as a string. Inside Rows insert it as an SQL snippet.\nUvis will not check the syntax, but the database will.");
                goto Common;
            case PathState.Unknown:  // After an unrecognized or cut identifier at the start of the path. 
                addKeyword("Not", "False if the rest is true, and vice versa."); // The indentifier may be Not.
                goto Common;
            case PathState.Else:
                addKeyword("Else", "Execute when condition is false");
                goto case PathState.ExpectStatement;
            case PathState.ExpectStatement: // In start of event handler, etc.
                addOperator("{", "", 0);
                addEndBracket();
                addKeyword("If", "If (b) {c; d} Else {e; f}"); // Appears in alfa sequence. Should start with If, not For, but how?
                addKeyword("For", "Execute statements repeatedly, e.g. repeat for i = 0, 1 ... 9: For (i=0; i<10; i=i+1) {c; d[i] = ... } ");
                addKeyword("Return", "Return from the event handler");
Common:
                string loops = null;
                for (int i = 0; i < Parser.forStack.Count; i++) 
                    loops = loops + Parser.forStack[i] + "  ";
                if (loops != null) addKeyword(loops, "Loop variables");
                addPrefix("Me", "This component");
                addPrefix("Main", "The component that has my row");
                addPrefix("Form", "My Form component");
                addPrefix("Param", "The parameters given to my Form at OpenForm. Ex: Param[0].");
                addPrefix("Forms", @"The collection of all forms, including closed forms. Ex: Forms!Detail: the Detail form. 
Forms!Detail(1): clone(1) of the Detail form. Null if the form is closed or doesn't exist.");
                addPrefix("Templates", "The collection of all form templates. Ex: Templates!Detail. For addressing formulas.");
                addPrefix("System", "For addressing system items. Ex: System.Now().");
                // Other prefixes: Me, Map. The elements in these collections are suggested directly where relevant.

                string prefix = (template == meTpl ? "Me" : template.Name) ;
                if (pi.dot == null || pi.dot.val == ".") addFieldPrefixes(null, "."); 
                if (pi.dot == null || pi.dot.val == ".") addRowFields(prefix, "."); // Including a table prefix
                if (pi.dot == null || pi.dot.val == ".") addAggregates(prefix, ".");
                if ((pi.dot == null || pi.dot.val == "." || pi.dot.val == ">-" || pi.dot.val == ">=") && !pi.inTailExpr) addManyOneRelations(prefix); // Including any necessary table prefixes
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addProperties(prefix, "!"); // In this template, including function properties
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addTemplates("Form", "!"); // In this form
                if (pi.dot == null || pi.dot.val == ".") addSystemItems("System", ".");
                break;

            case PathState.Rows:
                addPrefix("Map", "Select a table in the data map. Ex: Rows: Map.Persons");
                //if (!(template is FormTemplate)) addPrefix("Form", "Create a child bundle in the Form. Ex: Form -< Children");
                if (pi.dot == null || pi.dot.val == ".") addTables("Map", ".");
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addTemplates("Form", "!");
                break;
            case PathState.Forms: // After the prefix "Forms". Allowed: dot+form name (short file name).
                addForms();
                break;
            case PathState.Control:  // After a reference to a control. Allowed: dot+field (may have a table prefix), dot+aggregate, dot+property, dot+template, [ ]
                if (pi.dot == null || pi.dot.val == ".") addFieldPrefixes(null, "."); 
                if (pi.dot == null || pi.dot.val == ".") addRowFields(null, ".");
                if (pi.dot == null || pi.dot.val == ".") addAggregates(null, ".");
                //if (pi.dot == null || pi.dot.val == "." || pi.dot.val == ">-" || pi.dot.val == ">=") addManyOneRelations(null); // Including any necessary table prefixes
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addProperties(null, "!");
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addTemplates(null, "!"); // In the path's template.
                if (pi.dot == null) addKeyword("[", "Refer to a component in the bundle.");
                break;
            case PathState.Property:  // After a reference to a property. Allowed: [ ], dot+property {indirect reference}, operator, EOL.
                TplProperty tplProp = pi.tplProp;
                if (tplProp != null && tplProp.startChain != null && tplProp.startChain.val.ToUpper() == "FIND") {
                    FindExpr find = tplProp.expr as FindExpr;
                    addKeyword("Bundle", "The bundle of my siblings in this Find formula. Ex: " + tplProp.Name + " !Bundle[Index + 1]");
                    template = find.targetTpl;
                    addProperties(null, "!");
                } else if (pi.dot != null && pi.ident.next == pi.dot) // The formula ends with a dot-operator.
                    addKeyword("! property", "Assumes that \"" + pi.ident.val + "\" is a component reference.");
                else {
                    if (pi.dot == null && pi.ident != null) addKeyword("[", "Assumes that \"" + pi.ident.val + "\" is a list property or a component reference.");
                    addEndBracket();
                    addOperator("+  *  ^  &  =  <>  <  >  ?  And  Default Like Mod  . . . ", "", 12);
                    if (pi.ident != null) addKeyword("! property", "Assumes that \"" + pi.ident.val + "\" is a component reference.");
                }
                break;

            case PathState.Function:  // After a reference to a function. Allowed: ( )
                addOperator("(", "", 8);
                break;
            case PathState.BundleSource:
                foreach (TplProperty findRef in template.findProperties)
                    addKeyword(findRef.tpl.Name + "! " + findRef.Name, "A Find bundle from "+findRef.tpl.Name + " that refers to me");
                break;
            case PathState.Map: // After the prefix "Map": Allowed: table.
                addTables("", ".");
                break;
            case PathState.RecordSet: // After a table or join expression. Allowed: dot+relation (maybe prefixed), Where, Group By, etc., EOL.
                if (pi.dot == null && pi.la.kind == Parser.TkKind.Operator) 
                    break; // Deals with first part of -< etc.
                addSQLKeywords(); // Deal with first part of -< etc.
                if (pi.dot != null || pi.la.val == "") addOneManyRelations(null);
                if (pi.dot != null || pi.la.val == "") addManyOneRelations(null);
                break;
            case PathState.RelationPrefix: // After a join-operator and a table name. Allowed: Dot relationName.
                addOneManyRelations(null, pi.tblPrefix.val);
                addManyOneRelations(null, pi.tblPrefix.val);
                break;
            case PathState.AscDesc: // After an order-by value. Allowed: Asc, Desc, ","
                addKeyword("Desc", "Order in descendiing sequence");
                addSQLKeywords();
                goto case PathState.Not;
            case PathState.Record: // After a record reference (me>-relation). Allowed: dot+field (maybe prefixed), dot+aggregate, dot+relation.
                if (pi.dot == null || pi.dot.val == ".") addRecordFields(null, pi.toTableName);
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == ">-" || pi.dot.val == ">=") addManyOneRelations(null); // Including any necessary table prefixes
                break;
            case PathState.FieldPrefix:
                addRecordFields(pi.tblPrefix.val, pi.toTableName);
                break;
            case PathState.Field: // After a reference to a field. Allowed: operator, EOL.
                addEndBracket();
                addOperator("+  *  ^  &  =  <>  <  >  ?  And  Default  Like  Mod  . . . ", "", 12);
                break;
            case PathState.Aggregate: // After an aggregate name (Sum, Count, etc.). Allowed: ( ).
                addOperator("(", "", 8);
                break;
            case PathState.Templates: // After the prefix "Templates". Allowed: form.
                addForms();
                break;
            case PathState.FormTpl: // After Templates ! form. Allowed: template name or property name.
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addTemplates(null, "!"); // In the path's form
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addFormulaProperties(null, "!");
                break;
            case PathState.Template: // After Templates ! form ! template: Allowed: property name.
                if (pi.dot == null || pi.dot.val == "." || pi.dot.val == "!") addFormulaProperties(null, "!");
                break;
            case PathState.System: // After the System prefix. Allowed: system function.
                if (pi.dot == null || pi.dot.val == ".") addSystemItems(null, ".");
                break;

            case PathState.Find: // After Find. Allowed: Template name in this form.
                addTemplates(null, "!"); break;
            case PathState.FindTpl: // After Find tplName. Allowed: On, Order By, End-of-line.
                new Suggestion("On", null, "", "", "Find keyword", "Example: Find heading On headingField = Me.Field And . . . Order By . . .",
                    "Example: Find heading On headingField = Me.Field And . . . Order By . . .", 0, null, null);
                new Suggestion("Order By", null, "", "", "Find keyword", "Example: Find heading Order By . . .",
                    "Example: Find heading Order By . . .", 1, null, null);
                new Suggestion("End-of-line", null, "", "", "Find keyword", "Example: Find heading",
                    "Example: Find heading", 2, null, null);
                break;
            case PathState.FindOn: // After Find ... On key = key. Allowed: Order By, EOL, as in Value state.
                new Suggestion("Order By", null, "", "", "Find keyword", "Example: Find heading . . . Order By . . .",
                    "Example: Find heading . . . Order By . . .", 0, null, null);
                goto case PathState.Value;

            case PathState.Value: // After another kind of operand, e.g. 123, "abc", a+b, (a+b), Sin(a). Allowed: binary operator, ), ], EOL 
                if (pi.la.error == null) {
                    if (expectedSymbol != null) {
                        addOperator(expectedSymbol, "", 6); // E.g. comma at end of an expression.
                        expectedSymbol = null;
                    }
                    addEndBracket(); // Should appear first.
                    addOperator("+  *  ^  &  =  <>  <  >  ?  And  Default  Like  Mod  . . . ", "", 12);
                } else { // Uncompleted string or date 
                    string s = pi.la.kind == Parser.TkKind.String ? "\"" : "#";
                    addComment(s);
                }
                break;
            case PathState.FreeText: // Inside a comment. Allowed: any char. 
                addComment("Comment - no suggestion");
                break; // Shows as "No suggestion".
            case PathState.AfterError:
                addComment("After an error - no suggestion");
                break;
            case PathState.Enum:  // After an enumerator type, e.g. Color
                if (pi.dot == null || pi.dot.val == ".") addEnums();
                break;
            case PathState.SystemVariable: // After a system variable, e.g. MouseX - MOAK 28.01.13
                addEndBracket();
                addOperator("+  *  ^  &  =  <>  <  >  ?  And  Default  Like  Mod  . . . ", "", 12);
                break;
            case PathState.EndStatement:
                addEndBracket();
                addOperator(";  Default  End-of-line", "", 0);
                break;
            case PathState.ExpectSymbol: // When only one symbol is possible at cutToken.
                addOperator(expectedSymbol, "", 0);
                expectedSymbol = null;
                break;
            }
            return possibleNames;
        }

        private static void addEndBracket() {
            if (currentPi.brackets.Count == 0) return;
            addOperator(currentPi.brackets.Peek(), "", 0);
        }

        private static void addAggregates(string prefix, string dot) {
            if (currentPi.RowTpl == null || currentPi.RowTpl.RowsExpr == null) return; // Aggregate only with Rows in template or main templates
            addAggregate(prefix, dot, "Count", "(x) Number of rows in my group, omitting rows with x=null");
            addAggregate(prefix, dot, "Sum", "(x) Sum of the x fields in my group");
            addAggregate(prefix, dot, "Min", "(x) Minimum of the x fields in my group");
            addAggregate(prefix, dot, "Max", "(x) Maximum of the x fields in my group");
            addAggregate(prefix, dot, "Avg", "(x) Average of the x fields in my group");
            addAggregate(prefix, dot, "Var", "(x) Variance of the x fields (divide by Count-1)");
            addAggregate(prefix, dot, "VarP", "(x) Variance of the x fields (divide by Count)");
            addAggregate(prefix, dot, "StDev", "(x) Standard deviation (square root of Var)");
            addAggregate(prefix, dot, "StDevP", "(x) Standard deviation (square root of VarP");
            addAggregate(prefix, dot, "First", "(x) Field x in the first record of my group. May be null.");
        }

        private static void addAggregate(string prefix, string dot, string name, string tip) {
            string descr = null;
            if (currentPi.RowTpl == null) return;
            if (currentPi.RowTpl.RowsExpr.exprGroupby.Count == 0) descr = "( ) Requires Rows with Group By";
            new Suggestion(name, prefix, "", dot, "Aggregate", "", (descr == null ? tip : descr), 0, "", null);
        }

        private static void addConstant() {
            new Suggestion("123, \"abc\", True, False, Null", null, "", "", "KeyWord", "Constants, ", "Constants", -1, "", null);
            new Suggestion("#24-12-2013 14:15#", null, "", "", "KeyWord", "dd-mm-yyyy . . . DateTime in UK English format.", "dd-mm-yyyy, DateTime in UK English format.", 0, "", null);
        }

        private static void addOperator(string op, string tooltip, int position) {
                new Suggestion(op, "", null, "", "Operator", null, tooltip, position, null, null);
                if (currentPi.allowSQLkeyword) addSQLKeywords();
        }

        private static void addKeyword(string op, string tooltip) {
            new Suggestion(op, null, null, null, "Keyword", null, tooltip, 0, null, null);
        }

        private static void addComment(string op) {
            new Suggestion(op, "", null, "", "Operator", null, "", 0, null, null); // Doesn't allow insert suggestion
        }

        private static void addPrefix(string op, string tooltip) {
            new Suggestion(op, null, null, null, "Prefix", null, tooltip, 0, null, null);
        }

        private static void addFieldPrefixes(string prefix, string dot) {
            // prefix is null when a prefix isn't allowed, e.g. after Me. Dot is the suggested dot-operator. 
            if (rowTpl == null || rowTpl.RowsExpr == null) return; // No rows, no fields.
            if (rowTpl.RowsExpr.joinList.Count < 2) return; // No reason to use or show the prefix.
            foreach (Join join in rowTpl.RowsExpr.joinList) { // Look at each table in the Rows join.
                //string type = f.Value.ReadableType;
                new Suggestion(join.prefix, prefix, null, dot, "Prefix", "Type a bit of the field name to see the different meanings of it", "Prefix to fields in table " + join.table.Name, 0, "", null);
                // Should report the relation name because that is what the designer sees.
            }
        }

        private static void addRowFields(string prefix, string dot) {
            // prefix is null when a prefix isn't allowed, e.g. after Me. Dot is the suggested dot-operator. 
            if (rowTpl == null || rowTpl.RowsExpr == null) return; // No rows, no fields.
            string warn = rowTpl.RowsExpr.exprGroupby.Count == 0 ? "" : ". Must be in Group By or First( ) ...";
            foreach (Join j in rowTpl.RowsExpr.joinList) // Look at each table in the Rows join.
                foreach (var f in j.table.Columns) { // Suggest each field name.
                    string type = f.Value.ReadableType;
                    new Suggestion(f.Key, prefix, j.prefix, dot, "Field", j.table.Name + " " + type + warn, j.prefix + ", " + type + warn, 0, "", null);
                    // Should report the relation name because that is what the designer sees.
                }
            }

        private static void addRecordFields(string prefix, string tableName) {
            // prefix is null when a prefix isn't allowed, e.g. after Me. 
            if (rowTpl == null || rowTpl.RowsExpr == null) return; // No rows, no fields.
            string warn = rowTpl.RowsExpr.exprGroupby.Count == 0 ? "" : ". Must be in Group By or First( ) ...";
            foreach (var f in VisEngine.DataMaps[0].Tables[tableName].Columns) { // Suggest each field name.
                string type = f.Value.ReadableType;
                new Suggestion(f.Key, "", prefix, ".", "Field", tableName + " " + type + warn, tableName + ", " + type + warn, 0, "", null);
                // Should report the original table name because this is not part of a join.
            }
        }

        private static void addTables(string prefix, string dot) {
            // prefix is null when a prefix isn't allowed, e.g. after Map. Dot is the suggested dot-operator. 
            foreach (var t in VisEngine.DataMaps[0].Tables) {
                string keyFields="";
                foreach (var k in t.Value.Keys) keyFields = keyFields+ (keyFields==""? "" : ", ") + k.Key;
                new Suggestion(t.Key, prefix, null, dot, "Table", "Table with key: " + keyFields, "Table with key: " + keyFields, 0, "", null); 
            }
        }

        private static void addManyOneRelations(string prefix) { // OK in an ordinary formula. Will be included in the Row joins
            if (rowTpl == null) return; // No rows, no fields.
            RecordSetExpr rows = rowTpl.RowsExpr;
            if (rows == null) return; // No rows
            VisTool.Relationship rel;

            // Look at "table" in all joins in the list, except when Group By. For joins starting with a table, the first join is the table with a null relation.
            if (!currentPi.inTailExpr && rows.exprGroupby.Count > 0) { // Group By. Look only at the last relationship. 
                Join j = rows.joinList[rows.joinList.Count - 1];
                foreach (var r in j.table.Relationships) {
                    rel = r.Value;
                    if (rel.Cardinality == CardinalityType.manyone && NotInRows(rel.RelationName))
                        new Suggestion(rel.RelationName, prefix, rel.FromTableName, ">-", "Relation", "", "Add fields from " + rel.ToTableName, 0, null, null);
                }
                return;
            }

            foreach (Join j in rows.joinList) // Look at each table in the Rows join.
                foreach (var r in j.table.Relationships) {
                    rel = r.Value;
                    if (rel.Cardinality == CardinalityType.manyone && NotInRows(rel.RelationName))
                        new Suggestion(rel.RelationName, prefix, rel.RelationName, ">-", "Relation", "", "Add fields from " + rel.RelationName, 0, null, null);
                }
        }

        private static void addManyOneRelations(string prefix, string tableName) {
                Table table = VisEngine.DataMaps[0].Tables[tableName];
                foreach (var r in table.Relationships) {
                    VisTool.Relationship rel = r.Value;
                    string m = "For each " + rowTpl.Name + " get all rows from " + rel.ToTableName;
                    if (rel.Cardinality == CardinalityType.manyone && NotInRows(rel.RelationName))
                        new Suggestion(rel.RelationName, prefix, rel.FromTableName, ">-", "Relation", m, "Get related rows from " + rel.ToTableName, 0, null, null);
                }
        }

        private static void addOneManyRelations(string prefix) { // Not allowed in ordinary formulas. Differs from addManyOneRelations also in help texts.
            if (rowTpl == null) return; // No rows, no fields.
            RecordSetExpr rows = rowTpl.RowsExpr;
            if (rows == null) return; // No rows
            VisTool.Relationship rel;

            // Look at "table" in all joins in the list, except when Group By. For joins starting with a table, the first join is the table with a null relation.
            if (!currentPi.inTailExpr && rows.exprGroupby.Count > 0) { // Group By. Look only at the last relationship. 
                Join j = rows.joinList[rows.joinList.Count - 1];
                foreach (var r in j.table.Relationships) {
                    rel = r.Value;
                    string m = "For each " + rowTpl.Name + " get all rows from " + rel.ToTableName;
                    if (rel.Cardinality == CardinalityType.onemany && NotInRows(rel.RelationName))
                        new Suggestion(rel.RelationName, prefix, rel.FromTableName, "-<", "Relation", m, "Get related rows from " + rel.ToTableName, 0, null, null);
                }
                return;
            }

            foreach (Join j in rows.joinList) // Look at each table in the Rows join.
                foreach (var r in j.table.Relationships) {
                    rel = r.Value;
                    string m = "For each " + rowTpl.Name + " get all rows from " + rel.ToTableName;
                    if (rel.Cardinality == CardinalityType.onemany && NotInRows(rel.RelationName))
                        new Suggestion(rel.RelationName, prefix, rel.FromTableName, "-<", "Relation", m, "Get related rows from " + rel.ToTableName, 0, null, null);
                }
        }

        private static void addOneManyRelations(string prefix, string tableName) {
            Table table = VisEngine.DataMaps[0].Tables[tableName];
            foreach (var r in table.Relationships) {
                VisTool.Relationship rel = r.Value;
                string m = "For each " + rowTpl.Name + " get all rows from " + rel.ToTableName;
                if (rel.Cardinality == CardinalityType.onemany && NotInRows(rel.RelationName))
                    new Suggestion(rel.RelationName, prefix, rel.FromTableName, "-<", "Relation", m, "Get related rows from " + rel.ToTableName, 0, null, null);
            }
        }

        private static bool NotInRows(string relationName) { // Check whether the relation is already in the join list being compiled.
            if (rowTpl.RowsExpr == null) return true;
            foreach (Join j in rowTpl.RowsExpr.joinList)
                if (j.prefix.ToLower() == relationName.ToLower()) return false; // Name is a table name or relation name in the join list.
            return true;
        }

        private static void addProperties(string prefix, string dot) { // All instance properties. Excludes event handlers.
            new Suggestion("Index", prefix, null, dot, "Common", "My number in my bundle", "My number in my bundle", 0, "", null);
            new Suggestion("Last", prefix, null, dot, "Common", "Index of the last component in my bundle", "Index of the last component in my bundle", 0, "", null);

            //Add all control properities
            ComponentInfo c = template.Info; //VisEngine.ComponentByName[template.typeVisComponent.Name];
            foreach (var p in c.PropertyByName) {
                VisPropAtt attr = p.Value.VisAttribute;
                if (attr.InPopup && attr.PropType != VisPropAtt.PropertyType.Eventhandler) { // Event handlers cannot be addressed from a formula.
                    string tip = attr.ShortHelp;
                    new Suggestion(p.Key, prefix, null, dot, attr.PropertyCategory, attr.LongHelp == null ? attr.ShortHelp : attr.LongHelp, tip, attr.Position, "", null);
                }
            }
            foreach (var d in template.DesignProperties)
                //if (d.Value.PropKind == PropertyKind.Designer)
                    new Suggestion(d.Value.Name, prefix, null, dot, "Designer", "", "", 0, null, null);
        }

        private static void addFormulaProperties(string prefix, string dot) { // All template properties, e.g. to get their formula. Includes event handlers. 
            ComponentInfo c = template.Info; //VisEngine.ComponentByName[template.typeVisComponent.Name];
            foreach (var p in c.PropertyByName) {
                VisPropAtt attr = p.Value.VisAttribute;
                    string tip = attr.PropType == VisPropAtt.PropertyType.Method ? "( ) " + attr.ShortHelp : attr.ShortHelp;
                    new Suggestion(p.Key, prefix, null, dot, attr.PropertyCategory, attr.LongHelp == null ? attr.ShortHelp : attr.LongHelp, tip, attr.Position, "", null);
            }
            foreach (var d in template.DesignProperties)
                if (d.Value.PropKind == PropertyKind.Designer)
                    new Suggestion(d.Value.Name, prefix, null, dot, "Designer", "", "", 0, null, null);
        }

        private static void addSystemItems(string prefix, string dot) {
            foreach (var d in VisEngine.SystemMethodByName) {
                SystemMethodInfo s = d.Value;
                if (s.attr.itemType != VisSystemAtt.ItemType.Statement || property.PropKind == PropertyKind.EventHandler)
                    new Suggestion(s.Name, prefix, null, dot, "System", s.attr.LongHelp, s.attr.ShortHelp, s.attr.Position, null, null);
            }
             //Get System Variables
            foreach (var v in VisEngine.SystemFieldByName) {
                SystemFieldInfo p = v.Value;
                new Suggestion(p.Name, prefix, null, dot, "System", p.attr.LongHelp, p.attr.ShortHelp, p.attr.Position, null, null);
            }
        }

        private static void addForms() {
            foreach (var v in VisEngine.SortedForms) {
                int cloneID = 0;
                string frmName = v.Key;
                FormTemplate frmTpl = VisEngine.GetFormTemplate(frmName, ref cloneID); // Any clone will do.
                string c = frmTpl != null ? "Form, open now" : "Form, closed now";
                new Suggestion(v.Key, "", null, "!", "Form", c, c, 0, null, null);
            }
        }

        private static void addTemplates(string prefix, string dot) {
            if (currentPi.Tpl == null) return;
            foreach (KeyValuePair<string, Template> kvp in currentPi.Tpl.FormTpl.Templates) { // template.FormTpl.Templates
                if (kvp.Value.Ignore) continue;
                if (kvp.Value == currentPi.block)
                    continue;
                new Suggestion(kvp.Key, prefix, "", dot, "Template", "", kvp.Value.Info.Type.Name, 0, null, null);
            }
        }

        private static void addEnums() { // pi.ident is an enum name
            var enumMethod = VisEngine.SystemMethodByName[currentPi.ident.val].PropertySetter;
            List<string> suggestions = null;
            object[] parameters = new object[] { null, suggestions };
            enumMethod.Invoke(null, parameters);
            suggestions = (List<string>)parameters[1];
            foreach (object o in suggestions)
                new Suggestion(o.ToString(), currentPi.ident.val, "", ".", "Enum", "", "", 0, null, null);
        }

        private static void addSQLKeywords() { // Used only in PathState.QueryTail.
            //RecordSetExpr rs = currentPi.leftNode as RecordSetExpr;
            if (rowTpl == null || rowTpl.RowsExpr == null) return;
            RecordSetExpr rs = rowTpl.RowsExpr;
            if (rs.exprCreate == null)
                new Suggestion("Create", null, "", "", "SQLkeyword", "Add n empty rows at the end of the rows.", "Add n empty rows at the end of the rows.", 0, null, null);
            if (rs.exprGroupby.Count == 0) 
                new Suggestion("Group By", null, "", "", "SQLkeyword", "Divide rows into groups.", "Divide rows into groups and aggregate each group with Sum(F), etc.", 0, null, null);
            if (rs.exprHaving == null)
                new Suggestion("Having", null, "", "", "SQLkeyword", "Select which groups to include.", "Select which groups to include.", 0, null, null);
            if (rs.exprOrderby == null)
                new Suggestion("Order By", null, "", "", "SQLkeyword", "Order the rows in sequence.", "Order the rows in sequence.", 0, null, null);
            if (rs.exprTop == null)
                new Suggestion("Top", null, "", "", "SQLkeyword", "Include only the first n rows.", "Include only the first n rows.", 0, null, null);
            if (rs.exprWhere == null)
                new Suggestion("Where", null, "", "", "SQLkeyword", "Select which rows to include.", "Select which rows to include.", 0, null, null);
        }
    }
}