﻿using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Data;
using System.ComponentModel;
using System.Reflection;
using System.Data.OleDb;
using VisTool;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace VisTool {

    [VisClassAtt(ToolboxCategory = "Control", Icon = "glyph.png",
        ShortHelp = "Circle, triangle . . . according to the Shape property")]
    public class Glyph : VisComponent {
        public Glyph(int targetCount, DataRow row, Template tpl) : base(targetCount, row, tpl) { }

        public static int ShapeID; // Property index. Set at initialization in ReflectDLL. 
        private int shape = 3; //1 Line, 2 Ellipse,3 Triangle, 4 Box, 5 Pentagon, 6 Hexagon
        [VisPropAtt(VisPropAtt.PropertyType.Property, "3", LongHelp = "The glyph shape. 1: Line, 2: Ellipse, 3: Triangle, 4: Box, 5: Pentagon, 6: Hexagon, 7: TopDownTriangle",
            PropertyCategory = "Special", ShowInitially = true,
            ShortHelp = "The glyph shape. 1: Line, 2: Ellipse, 3: Triangle, 4: Box, 5: Pentagon, 6: Hexagon, 7: TopDownTriangle", Position = 0)]
        public int Shape {
            get {
                CmpProp shapeProp = this.propertyValues[ShapeID];
                if (shapeProp.state >= State.Updated) return shape;
                Shape = toPosInt32(EvalProp(ShapeID, 0), 0);
                return shape;
            }
            set { shape = value; SetUpdated(ref propertyValues[ShapeID]); }
        }

        // A Glyph has no internal dependencies and thus no ResetUpdated().
        // At Paint time, the kernel has evaluated all the properties, and drawing is straight-forward.

        public PointF[] Points;
        public GraphicsPath Path = null;
        public Pen pen;
        public Brush brush;

        private void createLine() {
            Points = new PointF[2];
            Points[0] = new Point(RelativeX(left), RelativeY(top));
            Points[1] = new Point(RelativeX(left) + PhyS(width), RelativeY(top) + PhyS(height));
            Path = new GraphicsPath();
            Path.AddLine(Points[0], Points[1]);
        }

        private void createEllipse() {
            Path = new GraphicsPath();
            Path.AddEllipse(new Rectangle(RelativeX(left), RelativeY(top), PhyS(width), PhyS(height)));
        }

        private void createTriangle() {
            Points = new PointF[3];
            Points[0] = new PointF(RelativeX(left) + (float)PhyS(width) / 2f, RelativeY(top));
            Points[1] = new PointF(RelativeX(left) + PhyS(width), RelativeY(top) + PhyS(height));
            Points[2] = new PointF(RelativeX(left), RelativeY(top) + PhyS(height));
            Path = new GraphicsPath();
            Path.AddPolygon(Points);
            //Debug.Print("Triangle RelativeTop, top: " + Points[0].Y + ", " + top + ", " + this.tpl.Name);
        }

        private void createDownTriangle() {
            Points = new PointF[3];
            Points[0] = new PointF(RelativeX(left) + (float)PhyS(width) / 2f, RelativeY(top) + PhyS(height));
            Points[1] = new PointF(RelativeX(left) + PhyS(width), RelativeY(top));
            Points[2] = new PointF(RelativeX(left), RelativeY(top));
            Path = new GraphicsPath();
            Path.AddPolygon(Points);
        }

        private void createRectangle() {
            Path = new GraphicsPath();
            Path.AddRectangle(new Rectangle(RelativeX(left), RelativeY(top), PhyS(width), PhyS(height)));
        }

        private void createPentagon() {
            Points = new PointF[5];
            Points[0] = new PointF(RelativeX(left) + (float)PhyS(width) / 2f, RelativeY(top));
            Points[1] = new Point(RelativeX(left) + PhyS(width), RelativeY(top) + PhyS(height) * 4 / 10);
            Points[2] = new PointF(RelativeX(left) + (float)PhyS(width) * 8 / 10f, RelativeY(top) + PhyS(height));
            Points[3] = new PointF(RelativeX(left) + (float)PhyS(width) * 2 / 10f, RelativeY(top) + PhyS(height));
            Points[4] = new PointF(RelativeX(left), RelativeY(top) + PhyS(height) * 4 / 10);
            Path = new GraphicsPath();
            Path.AddPolygon(Points);
        }

        private void createHexagon() {
            Points = new PointF[6];
            Points[0] = new PointF(RelativeX(left) + PhyS(width) * 25 / 100f, RelativeY(top));
            Points[1] = new PointF(RelativeX(left) + (float)PhyS(width) * 75 / 100f, RelativeY(top));
            Points[2] = new PointF(RelativeX(left) + PhyS(width), RelativeY(top) + (float)PhyS(height) * 50 / 100f);
            Points[3] = new PointF(RelativeX(left) + (float)PhyS(width) * 75 / 100f, RelativeY(top) + PhyS(height));
            Points[4] = new PointF(RelativeX(left) + (float)PhyS(width) * 25 / 100f, RelativeY(top) + PhyS(height));
            Points[5] = new PointF(RelativeX(left), RelativeY(top) + (float)PhyS(height) * 50 / 100f);
            Path = new GraphicsPath();
            Path.AddPolygon(Points);
        }

        public override void Paint(PaintEventArgs e) {
            if (!visible) return;
            base.Paint(e);
            switch (shape) {
            case 1: createLine(); break;
            case 2: createEllipse(); break;
            case 3: createTriangle(); break;
            case 4: createRectangle(); break;
            case 5: createPentagon(); break;
            case 6: createHexagon(); break;
            case 7: createDownTriangle(); break;
            default: createLine(); break;
            }

            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias; // Otherwise smooth borders look strange.
            pen = new Pen(BorderColor);
            pen.Width = PhyS(borderWeight);
            brush = new SolidBrush(BackColor);
            if (borderWeight > 0) // Windows draws a border even when pen.Width == 0.
                e.Graphics.DrawPath(pen, Path);
            e.Graphics.FillPath(brush, Path);
        }
    }

    //--------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "A polygon with a starting point and a list of increments", Icon = "polygon.png")]
    public class Polygon : VisComponent {
        public Polygon(int targetCount, DataRow row, Template tpl) : base(targetCount, row, tpl) { }
        int PointCount = 0;

        public static int XpositionsID; // Property index. Set at initialization in ReflectDLL. 
        public int[] xpositions = new int[0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 3, // Corresponds to Left
            ShortHelp = "The starting X position and a list of X increments, e.g. a V starting at (40, 30): 40, 10, 10")]
        public int[] Xpositions {
            get {
                CmpProp xpositionsProp = propertyValues[XpositionsID];
                CmpProp ypositionsProp = propertyValues[YpositionsID];
                if (xpositionsProp.state >= State.Updated) return xpositions;
                TplProperty targetProp = tpl.valPropArray[XpositionsID];

                object[] newList = EvalListProp(targetProp, xpositionsProp, xpositions);
                if (newList.Length < 2) {
                    VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "Xpositions",
                    "There must be at least two Xpositions", null, this.Index());
                    xpositions = new int[] { 30, 10, 10 }; // Default value.
                    newList = new object[] { 30, 10, 10 };
                } else if (newList.Length != xpositions.Length)
                    xpositions = new int[newList.Length];
                if (ypositionsProp.state >= State.Updated) { // Both lists are updated.
                    PointCount = Math.Min(xpositions.Length, ypositions.Length); // Use the shortest.
                    if (xpositions.Length != ypositions.Length) {
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "Xpositions",
                        "There are " + ypositions.Length + " Ypositions and " + xpositions.Length + " Xpositions. Must be the same.", null, this.Index());
                    }
                }
                int xPos = xpositions[0] = toInt32(newList[0], 30); // Current x-pos on the canvas.
                left = right = xPos;
                for (int i = 1; i < xpositions.Length; i++) {
                    int xIncr = xpositions[i] = toInt32(newList[i], 10);
                    xPos += xIncr;
                    if (left > xPos) left = xPos; // To define the area to invalidate.
                    if (right < xPos) right = xPos;
                }
                Left = left; Width = right - left; Right = right;
                if (PointCount > xpositions.Length) PointCount = xpositions.Length;
                return xpositions;
            }
            // set {} // Elements can be set, not the entire list.
        }

        public static int YpositionsID; // Property index. Set at initialization in ReflectDLL. 
        public int[] ypositions = new int[0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 2, // Corresponds to Top
            ShortHelp = "The starting Y position and a list of Y increments, e.g. a V starting at (40, 30): 30, 15, -15")]
        public int[] Ypositions {
            get {
                CmpProp xpositionsProp = propertyValues[XpositionsID];
                CmpProp ypositionsProp = propertyValues[YpositionsID];
                if (ypositionsProp.state >= State.Updated) return ypositions;
                TplProperty targetProp = tpl.valPropArray[YpositionsID];

                object[] newList = EvalListProp(targetProp, ypositionsProp, ypositions);
                if (newList.Length < 2) {
                    VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "Ypositions",
                    "There must be at least two Ypositions", null, this.Index());
                    ypositions = new int[] { 30, 15, -15 }; // Default value.
                    newList = new object[] { 30, 15, -15 };
                } else if (newList.Length != ypositions.Length)
                    ypositions = new int[newList.Length];
                if (xpositionsProp.state >= State.Updated) { // Both lists are updated.
                    PointCount = Math.Min(xpositions.Length, ypositions.Length); // Use the shortest.
                    if (xpositions.Length != ypositions.Length) {
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "Xpositions",
                        "There are " + ypositions.Length + " Ypositions and " + xpositions.Length + " Xpositions. Must be the same.", null, this.Index());
                    }
                }
                int yPos = ypositions[0] = toInt32(newList[0], 30); // Current y-pos on the canvas.
                top = bottom = yPos;
                for (int i = 1; i < ypositions.Length; i++) {
                    int yIncr = ypositions[i] = toInt32(newList[i], 15);
                    yPos += yIncr;
                    if (top > yPos) top = yPos; // To define the area to invalidate.
                    if (bottom < yPos) bottom = yPos;
                }
                Top = top; Height = bottom - top; Bottom = bottom; 
                if (PointCount > ypositions.Length) PointCount = ypositions.Length;
                return ypositions;
            }
            // set {} // Elements can be set, not the entire list.
        }

        public static int FillID; // Property index. Set at initialization in ReflectDLL. 
        public bool fill = false;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = true, Position = 4,
            ShortHelp = "When true, I fill the polygon with BackColor.")]
        public virtual bool Fill {
            get {
                CmpProp fillProp = this.propertyValues[FillID];
                if (fillProp.state >= State.Updated) return fill;
                Fill = toBool(EvalProp(FillID, false), false);
                return fill;
            }
            set {
                fill = value; SetUpdated(ref propertyValues[FillID]);
            }
        }


        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The top of the entire polygon (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, Position = 5,
            ShortHelp = "The top of the entire polygon.")]
        public override int Top { // Top must be calculated correctly, including the next point, to allow Click, Paint and Invalidate to work correctly.
            get {
                var ypos = Ypositions; // Will set Top, Height and Bottom.
                return top; 
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The height of the entire polygon (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, Position = 6,
            ShortHelp = "The height of the entire polygon.")]
        public override int Height {
            get {
                var ypos = Ypositions; // Will set Top, Height and Bottom.
                return height;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The bottom of the entire polygon (Internal formula)", PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "The bottom of the entire polygon.", Position = 7)]
        public override int Bottom {
            get {
                var ypos = Ypositions; // Will set Top, Height and Bottom.
                return bottom;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The left of the entire polygon (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
            ShortHelp = "The left of the entire polygon.", Position = 8)]
        public override int Left {
            get {
                var xpos = Xpositions; // Will set Left, Width and Right.
                return left;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The width of the entire polygon (Internal formula)",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "The width of the entire polygon.", Position = 9)]
        public override int Width {
            get {
                var xpos = Xpositions; // Will set Left, Width and Right.
                return width;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' The right of the entire polygon (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
            ShortHelp = "The right of the entire polygon.", Position = 10)]
        public override int Right {
            get {
                var xpos = Xpositions; // Will set Left, Width and Right.
                return right;
            }
        }

        public static new bool SetTopFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            TplProperty tplProp= tpl.valPropArray[YpositionsID];
            if (tplProp.Formula == null) { // Empty formula. Drag and drop a new Glyph. 
                tplProp.Formula = pos.ToString() + ", 15, -15";
            }
            if (!(tplProp.expr is ExpressionList)) return false;
            // Extract tokens of first expression:
            Token curToken = tplProp.startChain;
            while (curToken.next.val != ",") curToken = curToken.next; // NextOnLevel(curToken);
            Token tail = curToken.next;
            curToken.next = new Token(0);
            curToken.next.kind = Parser.TkKind.EOL; // Create artificial end.

            string formula = AdjustConstant(tplProp, pos, oldPos) + Parser.GetFormulaFromTokens(tail);
            VisEngine.CreateOrSetTemplateProperty(tplProp.tpl, tplProp.Name, formula);
            return true;
        }

        public static new bool SetHeightFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            // Has no effect. Might set the second list element, but unnecessary.
            return pos == oldPos;
        }

        public static new bool SetLeftFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            TplProperty tplProp = tpl.valPropArray[XpositionsID];
            if (tplProp.Formula == null) { // Empty formula. Drag and drop a new Glyph. 
                tplProp.Formula = pos.ToString() + ", 10, 10";
                return true;
            }
            if (!(tplProp.expr is ExpressionList)) return false;
            // Extract tokens of first expression:
            Token curToken = tplProp.startChain;
            while (curToken.next.val != ",") curToken = curToken.next; // NextOnLevel(curToken);
            Token tail = curToken.next;
            curToken.next = new Token(0);
            curToken.next.kind = Parser.TkKind.EOL; // Create artificial end.

            string formula = AdjustConstant(tplProp, pos, oldPos) + Parser.GetFormulaFromTokens(tail);
            VisEngine.CreateOrSetTemplateProperty(tplProp.tpl, tplProp.Name, formula);
            return true;
        }

        public static new bool SetWidthFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int pos, int oldPos) {
            // Has no effect. Might set the second list element, but unnecessary.
            return pos == oldPos;
        }

        // At Paint time, the kernel has evaluated all the properties, and drawing is straight-forward.
        public override void Paint(PaintEventArgs e) {
            if (!visible) return;
            PointF[] Points;
            GraphicsPath Path = new GraphicsPath(); 
            Pen pen;
            Brush brush;
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed; // Otherwise smooth borders look strange.
            Points = new PointF[PointCount];
            int x = xpositions[0];
            int y = ypositions[0];
            Points[0] = new Point(RelativeX(x), RelativeY(y));
            for (int i = 1; i < PointCount; i++) {
                x += xpositions[i]; y += ypositions[i];
                Points[i] = new Point(RelativeX(x), RelativeY(y));
            }
            Path.AddLines(Points);

            pen = new Pen(BorderColor);
            pen.Width = PhyS(borderWeight);
            e.Graphics.DrawPath(pen, Path);
            if (fill) {
                brush = new SolidBrush(backColor);
                e.Graphics.FillPath(brush, Path);
            }
        }
    }

    //--------------------------------------------------------------------------
    [VisClassAtt(ToolboxCategory = "Control", Icon = "graphpoint.png",
        ShortHelp = "One or more bundles of points (circles) connected with lines. Invisible points are ignored.")]
    public class GraphPoint : VisComponent {

        public GraphPoint(int targetCount, DataRow row, Template tpl) : base(targetCount, row, tpl) {
            TplProperty targetProp = tpl.valPropArray[RootID];
            if (targetProp.expr == null || targetProp.expr is FindExpr) return;
            VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, targetProp,
"Root must be empty or a Find formula", targetProp.startChain, this.Index()); // A wrong root formula.
        }

        GraphPoint nextVisible() { // Find next visible instance in the bundle.
            if (root == null) return null; // The graphpoint may not belong to a bundle.
            int index = root.index;
            while (++index < root.bundle.Instances.Count) {
                GraphPoint next = (GraphPoint)root.bundle.Instances[index].cmp;
                if (next.Visible) return next;
            }
            return null;
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false, ShortHelp = "(x) The vertical position corresponding to pixel position x on the Canvas.",
            LongHelp = "(x) The vertical position corresponding to pixel position x on the Canvas.\r\nAssumes that HPos increases in the bundle. Uses linear interpolation.")]
        public double GetVPos(double x) {
            object dummy = Root; // Make sure the bundles have been created.
            int upperIndex = root.bundle.Instances.Count - 1;
            return getVP(root.bundle.Instances, 0, upperIndex, x);
        }

        private double getVP(List<CmpProp> list, int first, int last, double x) { // Binary search.
            // Some graph points may be invisible and must be ignored. However, this is not possible if all of them are invisible.
            GraphPoint fst = (GraphPoint)list[first].cmp;
            while (!fst.Visible && first < last) { first++; fst = (GraphPoint)list[first].cmp; }
            GraphPoint lst = (GraphPoint)list[last].cmp;
            while (!lst.Visible && first < last) { last--; lst = (GraphPoint)list[last].cmp; }
            int m = (first + last) / 2;

            if (first == last || fst.HPos == lst.HPos) return (double)(fst.VPos);
            if (first +1 >= last) return fst.VPos + (x - fst.HPos) * (lst.VPos - fst.VPos) / (lst.HPos - fst.HPos);
            if (x < (list[m].cmp as GraphPoint).HPos) return getVP(list, first, m, x);
            return getVP(list, m, last, x);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false, ShortHelp = "(x) The GraphPoint component corresponding to pixel position x on the Canvas.",
            LongHelp = "(x) The last GraphPoint with HPos <= position x on the Canvas.\r\nAssumes that HPos increases in the bundle.")]
        public VisComponent GetPoint(double x) {
            object dummy = Root; // Make sure the bundles have been created.
            int upperIndex = root.bundle.Instances.Count - 1;
            return getPoint(root.bundle.Instances, 0, upperIndex, x);
        }

        private VisComponent getPoint(List<CmpProp> list, int first, int last, double x) { // Binary search.
            // Some graph points may be invisible and must be ignored. However, this is not possible if all of them are invisible.
            if (list.Count == 0) return null;
            GraphPoint fst = (GraphPoint)list[first].cmp;
            while (!fst.Visible && first < last) { first++; fst = (GraphPoint)list[first].cmp; }
            GraphPoint lst = (GraphPoint)list[last].cmp;
            while (!lst.Visible && first < last) { last--; lst = (GraphPoint)list[last].cmp; }
            int m = (first + last + 1) / 2; // Ensure that interval lenght is shrinking.

            if (first >= last || fst.HPos == lst.HPos) return lst;
            if (x < (list[m].cmp as GraphPoint).HPos) return getPoint(list, first, m-1, x);
            return getPoint(list, m, last, x);
        }

        public static int RootID; // Property index. Set at initialization in ReflectDLL. 
        BundleRef root;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, Position = 1,
            ShortHelp = "Empty for a single graph. For several bundles, the root component this point belongs to: Find lblA Where lblA.key = Me key . . . Order By . . .")]
        public BundleRef Root {
            get {
                CmpProp rootProp = this.propertyValues[RootID];
                if (rootProp.state >= State.Updated) return root;
                TplProperty targetProp = tpl.valPropArray[RootID];
                if (targetProp.expr == null) { // No root formula.
                    Root = this.propertyValues[RowsID].Value as BundleRef;
                } else {
                    Root = EvalProp(RootID, null) as BundleRef; // Will be a null root when there is no matching target.
                }
                return root;
            }
            set { root = value; SetUpdated(ref propertyValues[RootID]); }
        }

        public static int VPosID; // Property index. Set at initialization in ReflectDLL. 
        int vPos;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, Position = 9, 
            ShortHelp = "The vertical position of the point.")]
        public virtual int VPos { 
            get {
                CmpProp vPosProp = this.propertyValues[VPosID];
                if (vPosProp.state >= State.Updated) return vPos;
                object dummy = Root; // Make sure the bundles have been created.
                VPos = toInt32(EvalProp(VPosID, 20), 20);
                return vPos; 
            }
            set { vPos = value; SetUpdated(ref propertyValues[VPosID]); } 
        }

        public static int HPosID; // Property index. Set at initialization in ReflectDLL. 
        int hPos;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, Position = 10, 
            ShortHelp = "The horizontal position of the point.")]
        public virtual int HPos { 
            get {
                CmpProp hPosProp = this.propertyValues[HPosID];
                if (hPosProp.state >= State.Updated) return hPos;
                object dummy = Root; // Make sure the bundles have been created.
                HPos = toInt32(EvalProp(HPosID, 30), 30);
                return hPos; 
            }
            set { hPos = value; SetUpdated(ref propertyValues[HPosID]); } 
        }

        public static int LineColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color lineColor; // Not part of internal dependencies. Needs no lineColorUpdated.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", Position = 11,
            ReadOnly = false, ShowInitially = true, ShortHelp = "The color of my line to the next point in the bundle.")]
        public virtual Color LineColor { 
            get {
                CmpProp lineColorProp = this.propertyValues[LineColorID];
                if (lineColorProp.state >= State.Updated) return lineColor;
                object dummy = Root; // Make sure the bundles have been created.
                LineColor = toColor(EvalProp(LineColorID, Color.Black), Color.Black);
                return lineColor; 
            }
            set { lineColor = value; SetUpdated(ref propertyValues[LineColorID]); } 
        }

        public static int LineWeightID; // Property index. Set at initialization in ReflectDLL. 
        int lineWeight; // Not part of internal dependencies. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Special", Position = 12,
            ReadOnly = false, ShowInitially = true, ShortHelp = "Pixels across my line.")]
        public virtual int LineWeight { 
            get {
                CmpProp lineWeightProp = this.propertyValues[LineWeightID];
                if (lineWeightProp.state >= State.Updated) return lineWeight;
                object dummy = Root; // Make sure the bundles have been created.
                LineWeight = toPosInt32(EvalProp(LineWeightID, 1), 1);
                return lineWeight; 
            }
            set { lineWeight = value; SetUpdated(ref propertyValues[LineWeightID]); } 
        }

        public static int PointColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color pointColor; // Not part of internal dependencies. 
        [VisPropAtt(VisPropAtt.PropertyType.Property, null,
            PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, ShortHelp = "The color of the point (circle).", Position = 13)]
        public virtual Color PointColor { 
            get {
                CmpProp pointColorProp = this.propertyValues[PointColorID];
                if (pointColorProp.state >= State.Updated) return pointColor;
                object dummy = Root; // Make sure the bundles have been created.
                PointColor = toColor(EvalProp(PointColorID, Color.Black), Color.Black);
                return pointColor; 
            }
            set { pointColor = value; SetUpdated(ref propertyValues[PointColorID]); } 
        }

        public static int PointSizeID; // Property index. Set at initialization in ReflectDLL. 
        int pointSize = 5;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, 
            PropertyCategory = "Special", ReadOnly = false, ShowInitially = true, ShortHelp = "Pixels across the point.", Position = 14)]
        public virtual int PointSize { 
            get {
                CmpProp pointSizeProp = this.propertyValues[PointSizeID];
                if (pointSizeProp.state >= State.Updated) return pointSize;
                object dummy = Root; // Make sure the bundles have been created.
                PointSize = toPosInt32(EvalProp(PointSizeID, 5), 5);
                return pointSize; 
            }
            set { pointSize = value; SetUpdated(ref propertyValues[PointSizeID]); } 
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Top of the point including the line to the next point (Internal formula)", 
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, Position = 5,
            ShortHelp = "Top of the point including the line to the next point." )]
        public override int Top { // Top must be calculated correctly, including the next point, to allow Click, Paint and Invalidate to work correctly.
            get {
                top = VPos - PointSize / 2;
                GraphPoint next = nextVisible();
                if (next != null) {
                    top = Math.Min(next.VPos + next.PointSize / 3, top); // Dividing by 3 takes care of the rounded corners.
                }
                base.Top = top; return top; // base.Top sets state.
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Height of the point including the line to the next point. (Internal formula)", 
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, Position = 6,
            ShortHelp = "Height of the point including the line to the next point.")]
        public override int Height { 
            get { base.Height = Bottom - Top; return height; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Bottom of the point including the line to the next point. (Internal formula)", PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "Bottom of the point including the line to the next point.", Position = 7)]
        public override int Bottom {
            get {
                bottom = VPos - PointSize / 2 + pointSize; // Handles round-off.
                GraphPoint next = nextVisible();
                if (next != null) {
                    bottom = Math.Max(next.VPos - next.PointSize / 3, bottom);
                }
                base.Bottom = bottom; return bottom;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Left of the point including the line to the next point. (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
            ShortHelp = "Left of the point including the line to the next point.", Position = 8)]
        public override int Left {
            get {
                left = HPos - PointSize / 2;
                GraphPoint next = nextVisible();
                if (next != null) {
                    left = Math.Min(next.HPos + next.PointSize / 3, left);
                }
                base.Left = left; return left;
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Width of the point including the line to the next point. (Internal formula)",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
        ShortHelp = "Width of the point including the line to the next point.", Position = 9)]
        public override int Width {
            get { base.Width = Right - Left; return width; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Right of the point including the line to the next point. (Internal formula)",
            PropertyCategory = "Common", ReadOnly = true, ShowInitially = false,
            ShortHelp = "Right of the point including the line to the next point.", Position = 10)]
        public override int Right {
            get {
                right = HPos - PointSize / 2 + pointSize; // Handles round-off.
                GraphPoint next = nextVisible();
                if (next != null) {
                    right = Math.Max(next.HPos - next.PointSize / 3, right);
                }
                base.Right = right; return right;
            }
        }

        public static new void SetTopFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int dropTop, int old) {
            ChangePositionFormula(tpl.valPropArray[VPosID], oldTpl.valPropArray[VPosID], dropTop, old);
            //VisEngine.CreateOrSetTemplateProperty(tpl, "VPos", dropTop.ToString());
        }

        public static new void SetHeightFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int dropHeight, int old) {
            // Has no effect. Might set the point size, but unnecessary.
        }

        public static new void SetLeftFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int dropLeft, int old) {
            ChangePositionFormula(tpl.valPropArray[HPosID], oldTpl.valPropArray[HPosID], dropLeft, old);
            //VisEngine.CreateOrSetTemplateProperty(tpl, "HPos", dropLeft.ToString());
        }

        public static new void SetWidthFormula(ComponentInfo componentInfo, Template tpl, Template oldTpl, int dropWidth, int old) {
            // Has no effect. Might set the point size, but unnecessary.
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Exposed = false,
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "Not supported. Use PointColor.", Position = 11)]
        public override Color BackColor { get { return Color.White; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Exposed = false,
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "Not supported. Use LineColor.", Position = 12)]
        public override Color BorderColor { get { return Color.Black; } }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, Exposed = false,
            PropertyCategory = "Common", ReadOnly = false, ShowInitially = false, ShortHelp = "Not supported. Use LineWeight.", Position = 13)]
        public override int BorderWeight { get { return lineWeight; } }

        public override void Paint(PaintEventArgs e) {
            if (!visible) return;
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            PointF[] Points = new PointF[2];
            GraphicsPath Path = new GraphicsPath();
            Pen pen = new Pen(lineColor);
            Brush brush = new SolidBrush(pointColor);
            GraphPoint next = nextVisible();
            if (next != null) {
                pen.Width = PhyS(lineWeight);
                // The points may be painted in any order. Reduce the length of the line to keep the next point free.
                int hDiff = next.hPos - hPos, vDiff = next.vPos - vPos;
                double reduce = 1.0;
                double distance = Math.Sqrt(hDiff * hDiff + vDiff * vDiff);
                if (distance > 0) reduce -= pointSize / distance / 2.0; // Deal with points that are very close.
                if (reduce < 0) reduce = 0;
                Path.AddLine(new Point(RelativeX(hPos), RelativeY(vPos)),
                    new Point(next.RelativeX(hPos + (int)(reduce * hDiff)), next.RelativeY(vPos + (int)(reduce * vDiff))));
                e.Graphics.DrawPath(pen, Path);
                Path.Reset();
            }
            if (pointSize > 0) {
                pen.Color = pointColor;
                pen.Width = SystemItems.Zoom2;
                float pSize = PhyS(pointSize);
                Path.AddEllipse(new RectangleF((float)RelativeX(hPos) - pSize / 2f, (float)RelativeY(vPos) - pSize / 2f, pSize, (float)pSize));
                e.Graphics.DrawPath(pen, Path);
                e.Graphics.FillPath(brush, Path);
            }
        }
    }
}


