﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.ComponentModel;
using System.Diagnostics;
using VisTool;

namespace PanelGrid {

    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Grid of n*m cells (canvases)", Icon = "panelgrid.bmp")]
    public class PanelGrid : VisComponent {

        public PanelCell[ , ] panelCells = new PanelCell[0, 0]; //panel[column][row]. Initialize to avoid null refs when initializing other properties.
        public int columns = 0;
        public int rows = 0;
        public ExpressionList rowExpressions = null; // To test for Init during border drag.
        public ExpressionList columnExpressions = null; // To test for Init during border drag.
        public VBorder[] vBorders = new VBorder[0];
        public HBorder[] hBorders = new HBorder[0];

        public PanelGrid(int targetCount, System.Data.DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
        }

        public override void Dispose() {
            for (int i = 0; i < columns; i++)
                for (int j = 0; j < rows; j++)
                    panelCells[i, j].GuiInstance.Dispose();
        }

        public override void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            cellsCreated = vPosUpdated = hPosUpdated = false;
            base.ResetUpdated();
            foreach (Canvas can in panelCells) {
                can.dictComponents.Clear(); // The cells are not in the component tree handled by ResetUpdated. Must be cleared here.
                can.ResetUpdated();
            }
        }

        bool cellsCreated = false;
        void createCells() { // Update the dimensions of panelGrid[ , ] in order for Cell(c, r) to get a cell without evaluating the sizes.
            // Create cells, copy them or dispose them as needed.
            if (cellsCreated) return;
            cellsCreated = true;
            int newColumns = 0, newRows = 0;

            TplProperty tplProp = this.tpl.valPropArray[RowHeightsID]; 
            // When a formula has changed, Refresh doesn't create a new PanelGrid, but the tplProp.expr changes and the refs need update.
            if (tplProp.expr is ExpressionList)
                rowExpressions = (ExpressionList)tplProp.expr;
            else {
                rowExpressions = new ExpressionList(); rowExpressions.Add(new Constant<int>(50)); rowExpressions.Add(new Constant<int>(100));
            }
            tplProp = this.tpl.valPropArray[ColumnWidthsID];
            if (tplProp.expr is ExpressionList) 
                columnExpressions = (ExpressionList)tplProp.expr;
            else {
                columnExpressions = new ExpressionList(); columnExpressions.Add(new Constant<int>(100)); columnExpressions.Add(new Constant<int>(200));
            }

            newRows = rowExpressions.expressions.Count;
            newColumns = columnExpressions.expressions.Count;
            if (newRows == rows && newColumns == columns) return; // No change.

            PanelCell [,] newPanelCells = new PanelCell [newColumns, newRows];
            for (int i = 0; i < newColumns; i++) // Keep old panel cells as far as possible.
                for (int j = 0; j < newRows; j++)
                    if (i < columns && j < rows) //  && panelCells[i, j] != null TreeChange refresh will delete all cells.
                        newPanelCells[i, j] = panelCells[i, j];
                    else {
                        newPanelCells[i, j] = new PanelCell(0, null, this.tpl, i, j, this);
                    }
            for (int i = 0; i < columns; i++) // Dispose of deleted gui canvases.
                for (int j = 0; j < rows; j++)
                    if (i >= newColumns || j >= newRows) panelCells[i, j].Dispose();

            VBorder[] newVBorders = new VBorder[newColumns];
            HBorder[] newHBorders = new HBorder[newRows];

            for (int i = 0; i < newColumns; i++) // Add new borders.
                if (i < columns) newVBorders[i] = vBorders[i];
                else newVBorders[i] = new VBorder(this, i);
            for (int j = 0; j < newRows; j++)
                if (j < rows) newHBorders[j] = hBorders[j];
                else newHBorders[j] = new HBorder(this, j);

            for (int i = newColumns; i < columns; i++) // Delete old borders.
                vBorders[i].Dispose();
            for (int j = newRows; j < rows; j++) // Delete old borders.
                hBorders[j].Dispose();

            if (newColumns < columns || newRows < rows) // If cells removed, invalidate the old area for repaint.
                Canvas.GuiInstance.Invalidate(new System.Drawing.
                    Rectangle(RelativeX(this.left) - PhyS(this.borderWeight), RelativeY(this.top) - PhyS(this.borderWeight),
                    PhyS(this.width + 2 * this.borderWeight), PhyS(this.height + 2 * this.borderWeight)));

            rows = newRows; columns = newColumns; 
            panelCells = newPanelCells;
            vBorders = newVBorders; hBorders = newHBorders;
        }

        public bool vPosUpdated = false;
        void updateVpos(int sourceID) {
            if (vPosUpdated) return;
            vPosUpdated = true;
            CmpProp sourceProp = this.propertyValues[sourceID];
            sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.
            object dummy = RowHeights; dummy = BorderWeight;
            int h = 0;
            createCells(); // Necessary to set rows.
            for (int i = 0; i < rows; i++) h += rowHeights[i];
            base.Height = h + (rows - 1) * BorderWeight;
            this.propertyValues[HeightID].state = State.Updated;
            base.Bottom = base.Top + height;
            this.propertyValues[BottomID].state = State.Updated;
            if (sourceProp.state == State.Visited) // If it hasn't been computed, reset it to Old.
                sourceProp.state = State.Old;
        }

        public bool hPosUpdated = false;
        void updateHpos(int sourceID) {
            if (hPosUpdated) return;
            hPosUpdated = true;
            CmpProp sourceProp = this.propertyValues[sourceID];
            sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.
            object dummy = ColumnWidths; dummy = BorderWeight;
            int w = 0;
            createCells(); // Necessary to set columns.
            for (int i = 0; i < columns; i++) w += columnWidths[i];
            base.Width = w + (columns - 1) * BorderWeight;
            this.propertyValues[WidthID].state = State.Updated;
            base.Right = base.Left + width;
            this.propertyValues[RightID].state = State.Updated;
            if (sourceProp.state == State.Visited) // If it hasn't been computed, reset it to Old.
                sourceProp.state = State.Old;
        }

        public override void ComponentUpdated() {
            base.ComponentUpdated();
            createCells(); // RowHeights and ColumnWidths have been evaluated at this point. Now set the cell positions and sizes.
            //updateHpos(ColumnWidthsID); updateVpos(RowHeightsID);
            int cellLeft = this.left;
            for (int i = 0; i < columns; i++) {
                int cellTop = this.top;
                for (int j = 0; j < rows; j++) {
                    PanelCell c = panelCells[i, j];
                    c.Left = cellLeft; c.Width = columnWidths[i]; c.Right = c.left + c.width; // Sets also gui-properties and updated.
                    c.Top = cellTop; c.Height = rowHeights[j]; c.Bottom = c.top + c.height;
                    cellTop += (rowHeights[j] + borderWeight);
                    c.backColor = backColor;
                    c.Anonymous = true;
                    c.AddToCanvas(this.Canvas);
                }
                cellLeft += (columnWidths[i] + borderWeight);
            }
            for (int i = 0; i < columns; i++) {
                vBorders[i].setPos(panelCells[i, 0].right); // Adjust drag borders.
                vBorders[i].AddToCanvas(this.canvas); // Will be removed at Refresh()
            }
            for (int j = 0; j < rows; j++) {
                hBorders[j].setPos(panelCells[0, j].bottom);
                hBorders[j].AddToCanvas(this.canvas); // Will be removed at Refresh()
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(c, r) The cell in column c, row r. Place a component in a cell like this: Canvas: Cell(0, 2)")]
        public VisTool.Canvas Cell(int column, int row) {
            createCells();
            if (column < 0 || column >= columns) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Column index is " + column + ". Must be between 0 and " + (columns - 1), null, this.Index());
                column = (column < 0 ? 0 : columns-1); // Tempting to return panelCells[0, 0], but cannot change Canvas at this time.
            }
            if (row < 0 || row >= rows) {
                ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, VisEngine.curTemplate, VisEngine.curTplProp,
                    "Row index is " + row + ". Must be between 0 and " + (rows - 1), null, this.Index());
                row = (row < 0 ? 0 : rows-1); 
            }
            return panelCells[column, row];
        }

        public static int ColumnWidthsID; // Property index. Set at initialization in ReflectDLL. 
        public int [] columnWidths = new int [] { 100, 200 };
        [VisPropAtt(VisPropAtt.PropertyType.Property, "100, 200", ShowInitially = true, Position = 1,
            ShortHelp = "Width of my columns excluding the borders. Example with three columns: 20, Form!Width-78, Init 50")]
        public int[] ColumnWidths {
            get {
                CmpProp columnWidthsProp = this.propertyValues[ColumnWidthsID];
                if (columnWidthsProp.state >= State.Updated) return columnWidths;
                TplProperty targetProp = tpl.valPropArray[ColumnWidthsID];

                object[] newList = EvalListProp(targetProp, columnWidthsProp, columnWidths); // Sets state=Updated.
                if (newList.Length < 1)
                    columnWidths = new int[] { 100, 200 }; // Default value. Length = 1 possible.
                else {
                    if (newList.Length != columnWidths.Length)
                        columnWidths = new int[newList.Length];
                    for (int i = 0; i < columnWidths.Length; i++) {
                        columnWidths[i] = toInt32(newList[i], 100);
                        if (columnWidths[i] < MinCellWidth) {
                            ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, this.tpl, "ColumnWidths",
                                "ColumnWidth is " + columnWidths[i] + ". Must be larger than " + minCellWidth, null, this.Index());
                            columnWidths[i] = minCellWidth;
                        }
                    }
                }
                return columnWidths;
            }
            //set { columnWidths = value; } // Elements can be set, not the entire list.
        }

        public static int RowHeightsID; // Property index. Set at initialization in ReflectDLL. 
        public int[] rowHeights = new int [] { 50, 100 };
        [VisPropAtt(VisPropAtt.PropertyType.Property, "50, 100", ShowInitially = true, Position = 2,
            ShortHelp = "Height of my rows excluding the borders. Example with three rows: 20, Form!Height-100, Init 50")]
        public int [] RowHeights {
            get {
                CmpProp rowHeightsProp = propertyValues[RowHeightsID];
                if (rowHeightsProp.state >= State.Updated) return rowHeights;
                TplProperty targetProp = tpl.valPropArray[RowHeightsID];

                object[] newList = EvalListProp(targetProp, rowHeightsProp, rowHeights);
                if (newList.Length < 1) {
                    rowHeights = new int[] { 50, 100 }; // Default value. Length = 1 possible.
                    newList = new object[] { 50, 100 };
                } else if (newList.Length != rowHeights.Length)
                    rowHeights = new int[newList.Length];
                for (int i = 0; i < rowHeights.Length; i++) {
                    rowHeights[i] = toInt32(newList[i], 50);
                    if (rowHeights[i] < MinCellHeight) {
                        ErrorLogging.Log(ErrorMode.RunTime, Priority.Warning, this.tpl, "RowHeights",
                            "RowHeight is " + rowHeights[i] + ". Must be larger than " + minCellHeight, null, this.Index());
                        rowHeights[i] = minCellHeight;
                    }
                }
                return rowHeights;
            }
            //set { columnWidths = value; } // Elements can be set, not the entire list.
        }

        public override int borderWeightDefault { get { return 3; } } // Property to allow overriding.
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "Border thickness around my cells. Default: 3.", Position = 13)]
        public override int BorderWeight {
            get { return base.BorderWeight; }
            //set { base.BorderWeight = value; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
        ShortHelp = "Border color of my cells. Default: Color.Black.", Position = 12)]
        public override Color BorderColor {
            get { return base.BorderColor; }
            set { base.BorderColor = value; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
            ShortHelp = "Inner color of my cells.", Position = 11)]
        public override Color BackColor {
            get { return base.BackColor; }
            set { base.BackColor = value; }
        }

        public static int MinCellWidthID; // Property index. Set at initialization in ReflectDLL. 
        public int minCellWidth;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, 
            ShortHelp = "User cannot drag the columns smaller than this. Example: 4")]
        public int MinCellWidth {
            get {
                CmpProp minCellWidthProp = this.propertyValues[MinCellWidthID];
                if (minCellWidthProp.state >= State.Updated) return minCellWidth;
                MinCellWidth = toPosInt32(EvalProp(MinCellWidthID, 4), 4);
                return minCellWidth;
            }
            set { minCellWidth = value; SetUpdated(ref propertyValues[MinCellWidthID]); }
        }

        public static int MinCellHeightID; // Property index. Set at initialization in ReflectDLL. 
        public int minCellHeight;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, 
            ShortHelp = "User cannot drag the rows smaller than this. Example: 4")]
        public int MinCellHeight {
            get {
                CmpProp minCellHeightProp = this.propertyValues[MinCellHeightID];
                if (minCellHeightProp.state >= State.Updated) return minCellHeight;
                MinCellHeight = toPosInt32(EvalProp(MinCellHeightID, 4), 4);
                return minCellHeight;
            }
            set { minCellHeight = value; SetUpdated(ref propertyValues[MinCellHeightID]); }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Sum of row heights and internal borders (automatic)", PropertyCategory = "Common", ShowInitially = false, ReadOnly = true, Position = 7,
            ShortHelp = "Pixels from the top of the grid area inside the border to the bottom inside the border. Read-only.")]
        public override int Height { // Computed from row heights.
            get { updateVpos(HeightID); return height; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, LongHelp = "Number of pixels from my canvas top to my bottom inside the border.",
    PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
    ShortHelp = "Pixels from my canvas top to my bottom inside the border.", Position = 7)]
        public override int Bottom {
            get { updateVpos(BottomID); return bottom; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Sum of column widths and internal borders (automatic)", PropertyCategory = "Common", ShowInitially = false, ReadOnly = true, Position = 6,
            ShortHelp = "Pixels from the left of the grid area inside the border to the right inside the border. Read-only.")]
        public override int Width { // Computed from cell widths.
            get { updateHpos(WidthID); return width; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, null, PropertyCategory = "Common", ReadOnly = false, ShowInitially = false,
    ShortHelp = "Pixels from my canvas left to my right inside the border.", Position = 10)]
        public override int Right {
            get { updateHpos(RightID); return right; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null,  Position = -9, PropertyCategory = "EventHandler", ReadOnly = false,
    ShortHelp = "Occurs when user has dragged a border.")]
        public void BordersChanged(object sender, EventArgs e) {
            this.ExecuteHandler("BordersChanged", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null,  Position = -10, PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "Occurs repeatedly while user drags a border.")]
        public void BordersChanging(object sender, EventArgs e) {
            //Debug.Print("HTimeScale.BordersChanging");
            this.ExecuteHandler("BordersChanging", true);
        }

        public override void Paint(PaintEventArgs e) { // Paint all the borders.
            if (!visible) return;
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            //for (int i = 0; i < columns; i++) { // Set horizontal scrollbars
            //    int maxWidth = 0;
            //    for (int j = 0; j < rows; j++) maxWidth = Math.Max(panelCells[i, j].virtualWidth, maxWidth);
            //    ScrollableControl sc;
            //    HScrollProperties hs = null;
            //    for (int j = 0; j < rows; j++) {
            //        sc = panelCells[i, j].GuiInstance as ScrollableControl;
            //        hs = sc.HorizontalScroll;
            //        panelCells[i, j].corner.Left = maxWidth + panelCells[i, j].rightMargin - hs.Value;
            //        hs.LargeChange = columnWidths[i]; hs.SmallChange = 10;
            //        hs.Maximum = maxWidth; hs.Minimum = 0;
            //    }
            //    if (maxWidth > columnWidths[i]) { // Horizontal scroll needed.
            //        hs.Visible = true; hs.Enabled = true;
            //    } else { // Remove scroll.
            //        hs.Visible = false; hs.Enabled = false;
            //    }
            //}

            PanelCell c;
            int offset = (PhyS(borderWeight) + 1) / 2; // Make sure the border surrounds the component.
            // Notice that yPos = top is inside the client area and yPos = bottom is just below the client area.

            if (PhyS(borderWeight) > 0) { // Draw border:
                for (int i = 0; i < columns; i++) {
                    c = panelCells[i, 0];
                    e.Graphics.DrawLine(new Pen(BorderColor, PhyS(borderWeight)), RelativeX(c.right - 1) + offset, RelativeY(top), RelativeX(c.right - 1) + offset, RelativeY(bottom - 1) + PhyS(borderWeight));
                }
                for (int j = 0; j < rows; j++) {
                    c = panelCells[0, j];
                    e.Graphics.DrawLine(new Pen(BorderColor, PhyS(borderWeight)), RelativeX(left), RelativeY(c.bottom - 1) + offset, RelativeX(right - 1), RelativeY(c.bottom - 1) + offset);
                }
                // Draw border at left and at top:
                e.Graphics.DrawLine(new Pen(BorderColor, PhyS(borderWeight)), RelativeX(left) - offset, RelativeY(top) - borderWeight, RelativeX(left) - offset, RelativeY(bottom - 1) + PhyS(borderWeight));
                e.Graphics.DrawLine(new Pen(BorderColor, PhyS(borderWeight)), RelativeX(left), RelativeY(top) - offset, RelativeX(Right - 1) + PhyS(borderWeight), RelativeY(top) - offset);
            }
        }
    }

    //-----------------------------------------------------
        [VisClassAtt(ToolboxCategory = "Control", ShowInToolbox = false)]
    public class PanelCell : Canvas {
        public int iColumn = 0; // Needed for cell to identify itself with a formula. See GetCanvasFormula.
        public int iRow = 0;
        PanelGrid pGrid;

        public PanelCell(int targetCount, System.Data.DataRow row, Template tpl, int iColumn, int iRow, PanelGrid pGrid)
            : base(targetCount, row, tpl) {
            this.propertyValues = pGrid.propertyValues; // Share property values, particularly Rows property.

            Anonymous = true;
            this.iColumn = iColumn; this.iRow = iRow; this.pGrid = pGrid;
            ScrollableControl scr = this.GuiInstance as ScrollableControl;
            scr.AutoScroll = false; // Default values for cells. Bottom and Right-most cells get another setting when all cells have been created.
        }

        public override string GetCanvasFormula() { // Returns the formula to be used in a component's canvas property in order to place it in this cell.
            // Used for instance by drag-and-drop in the container. Example: A PanelGrid cell has this formula: grid1!Cell(i, j).
            return pGrid.tpl.Name + " ! Cell(" + iColumn + ", " + iRow + ")";
        }
    }

    //--------------------------------------------------------
    public class VBorder : VisComponent { // Catches border drags. Painting is done in PanelGrid.Paint. 
        PanelGrid panelGrid;
        int iColumn;

        public VBorder(PanelGrid panelGrid, int iColumn) {
            //this.Bundle = panelGrid.Bundle;
            this.tpl = panelGrid.tpl;

            propertyValues = panelGrid.propertyValues; // Share properties.

            this.Anonymous = true;
            this.panelGrid = panelGrid;
            this.iColumn = iColumn;
            this.visible = true;
            this.zOrder = panelGrid.ZOrder + 1; // To catch the mouse.
            this.movable = true;
        }

        public void setPos(int pos) {
            this.left = pos - 3;
            this.width = Math.Max(6, panelGrid.BorderWeight); 
            this.right = left + width;
            this.top = panelGrid.Top; this.height = panelGrid.Height; this.bottom = top + height;
        }

        public override void MouseMove(object sender, MouseEventArgs e) {
            if (this.Moving == Move.No) return; // Not dragging.
            this.Moving = Move.Moved;
            if (!panelGrid.columnExpressions.expressions[iColumn].Init) return; // The border cannot be dragged.
            int displacement = LogicalX(e.X) - SystemItems.MouseX;
            if (panelGrid.columnWidths[iColumn] + displacement < panelGrid.minCellWidth) return; // Cell too narrow.
            int invalidWidth = panelGrid.width;
            this.Left += displacement; // Displace myself.
            panelGrid.columnWidths[iColumn] += displacement; 

            panelGrid.hPosUpdated = false;
            panelGrid.ComponentUpdated(); // Update cell sizes, etc.
            if (panelGrid.Width > invalidWidth) invalidWidth = panelGrid.width;
            panelGrid.Canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(panelGrid.left) - PhyS(panelGrid.borderWeight), RelativeY(panelGrid.top) - PhyS(panelGrid.borderWeight), 
                PhyS(invalidWidth + 2 * panelGrid.borderWeight), PhyS(panelGrid.height + 2 * panelGrid.borderWeight)));
            updateMouseVariables(e);
            panelGrid.ExecuteHandler("BordersChanging", true);
        }

        public override void MouseUp(object sender, MouseEventArgs e) {
            if (this.Moving == Move.Moved) {
                panelGrid.ExecuteHandler("BordersChanged", true);
            }
            this.Moving = Move.No;
            this.canvas.GuiInstance.Capture = false;
        }

        public override void MouseEnter(object sender, EventArgs e) {
            if (panelGrid.columnExpressions.expressions[iColumn].Init) 
                this.canvas.GuiInstance.Cursor = Cursors.VSplit;
        }

        public override void MouseLeave(object sender, EventArgs e) {
            this.canvas.GuiInstance.Cursor = Cursors.Arrow;
            this.Moving = Move.No;
        }
    }

    //--------------------------------------------------------
    public class HBorder : VisComponent { // Catches border drags. Painting is done in PanelGrid.Paint. 
        PanelGrid panelGrid;
        int iRow;

        public HBorder(PanelGrid panelGrid, int iRow) {
            this.tpl = panelGrid.tpl;

            propertyValues = panelGrid.propertyValues; // Share properties.
            this.Anonymous = true;
            this.panelGrid = panelGrid;
            this.iRow = iRow;
            this.visible = true;
            this.zOrder = panelGrid.ZOrder + 1; // To catch the mouse.
            this.movable = true;
        }

        public void setPos(int pos) {
            this.top = pos - 3;
            this.height = Math.Max(6, panelGrid.BorderWeight); 
            this.bottom = top + height;
            this.left = panelGrid.Left; this.width = panelGrid.Width; this.right = left + width; 
        }

        public override void MouseMove(object sender, MouseEventArgs e) {
            if (this.Moving == Move.No) return; // Not dragging.
            this.Moving = Move.Moved;
            this.canvas.GuiInstance.Capture = true;
            if (!panelGrid.rowExpressions.expressions[iRow].Init) return; // The border cannot be dragged.
            int displacement = LogicalY(e.Y) - LogS(SystemItems.MouseY);
            if (panelGrid.rowHeights[iRow] + displacement < panelGrid.minCellHeight) return; // Cell too narrow.
            int invalidHeight = panelGrid.height;
            this.top += displacement; // Displace myself.
            panelGrid.rowHeights[iRow] += displacement; // Compensate for scrolling and borderPixels being relative to time scale.

            panelGrid.vPosUpdated = false;
            panelGrid.ComponentUpdated(); // Update cell sizes, etc.
            if (panelGrid.height > invalidHeight) invalidHeight = panelGrid.height;
            panelGrid.Canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(panelGrid.left) - PhyS(panelGrid.borderWeight), RelativeY(panelGrid.top) - PhyS(panelGrid.borderWeight),
                PhyS(panelGrid.width) + 2 * PhyS(panelGrid.borderWeight), PhyS(invalidHeight + 2 * panelGrid.borderWeight)));
            updateMouseVariables(e);
            panelGrid.ExecuteHandler("BordersChanging", true);
        }

        public override void MouseUp(object sender, MouseEventArgs e) {
            if (this.Moving == Move.Moved) {
                panelGrid.ExecuteHandler("BordersChanged", true);
            }
            this.Moving = Move.No;
            this.canvas.GuiInstance.Capture = false;
        }

        public override void MouseEnter(object sender, EventArgs e) {
            if (panelGrid.rowExpressions.expressions[iRow].Init)
                this.canvas.GuiInstance.Cursor = Cursors.HSplit;
        }

        public override void MouseLeave(object sender, EventArgs e) {
            this.canvas.GuiInstance.Cursor = Cursors.Arrow;
            this.Moving = Move.No;
        }
    }
}
