﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using System.Threading;
using System.ComponentModel;
using System.Diagnostics;
using VisTool; //in order to integrate the control with VisTool
using System.Data;

namespace TimeScale {

    /* An HTimeScale consists of a VisComponent that paints the entire time scale including the dragable borders and the drag indicator arrows. It generates a list of intervals and 
     * three lists of tags: TopTags, Large tags and Small tags. Each tag consists of a DateTime and an x-pos. This is based on borderValues and borderPixels. 
     * Painting the time scale simply draws a tag for each of these tags, possibly with a label. The external tag lines are painted by an 
     * anonymous VisComponent in each of the external tag canvases. It uses the same tag lists as HTimeScale.
     * 
     * The tricky part is how to generate the tags. The time scale can have around 14 different looks depending on the zoom factor. They are 
     * specified in a static array of structs, "intervalTypes". Each element specifies the zoom factor it applies to, the units for the three kinds of tags 
     * with this zoom factor, the number of units in each kind of tag, plus the format of the top label and large tag label, e.g. "MMM" when a date time is to
     * be shown as just the month. 
     * Small ticks never have a label. If there is room for a label, the label should be promoted to a large tag and smaller intervals used for the small tags.
     * 
     * The basic algorithm finds the proper intervalType based on the ratio (zoom factor) between the time span of the interval and the pixel span of the interval.
     * It then interprets the data in the IntervalType to generate the tag lists. This is done in the constructor of the Interval class. HTimeScale.Paint generates the labels.
     * There is always a top tag and a large tag at the beginning of an interval and at the end of the last interval. This helps paint cutting off the labels.
     * 
     * A dragable border consists of an Anonymous component that catches the mouse events. It doesn't paint anything. Painting is done by the time scale paint.
     * Since the dragable border component is very narrow, it can easily "lose" the mouse during dragging. For this reason it sets Capture mouse to true. This causes Windows to
     * send all mouse events to this canvas. The canvas passes them on to the dragable border component.
     * 
     * Dragging of the ribbon itself is handled by the time scale component. It too uses Capture mouse. The result is that the end-user can drag the 
     * ribbon even outside the form to really stretch the ribbon. This is often convenient.
     * 
     * Dragging of the border is only allowed when the borderPixel is Init. Similarly, dragging the ribbon across a border is only allowed when borderValue is Init.
     * Normally, ribbon dragging is across the closest border (if Init), but when Alt is down, dragging is across the opposite border. The drag indicator arrows shows what is possible.
     * 
     * BorderPixels specify x-positions relative to the left border of the time scale. The intervals and tag lists use x-positions relative to the canvas. 
     * Paint transforms this x to x-pos relative to the client area of the canvas or Form (like other components).
     * 
     * This version of HTimeScale is around 550 LOC plus 150 to specify properties available to the designer (version 5.0). 
     * It replaces an old version (until version 4.14) that was 3200 LOC, comprising 4 files.
     */
    [VisClassAtt(ToolboxCategory = "Control", ShortHelp = "Time ribbon with multiple zoom areas.", Icon = "timescale.png",
        LongHelp = @"Time ribbon with multiple zoom areas. User can drag the ribbon across the closest border.
Tag lines may extend into other canvases.")]
    public class HTimeScale : VisComponent {
        public ExpressionList valueExpressions = null; // To test for Init during ribbon drag.
        public ExpressionList pixelExpressions = null; // To test for Init during border drag.
        int IntervalCount = 1000; // The number of intervals = Min(BorderValues.Length, BorderPixels.Length) - 1.
        public Interval[] Intervals;
        VisComponent oldKeyFocus; // The Alt-key indicates dragging across the far border. 
        // At mouse enter, the timescale takes the key focus to rspond to key down. Resets key focus at mouse leave.

        const int indiWidth = 22; // Size of indicator arrow. For invalidating the area.
        const int indiHeight = 12;

        const int topHeight = 18; // Layout of the ribbon. 
        const int largeHeight = 16;
        const int smallHeight = 8;
        new const int height = topHeight + largeHeight + smallHeight;
        const int indiTop = height - 12; // Top of the drag indicator arrow relative to the time scale.
        static Font labelFont = new Font("Arial", 8, FontStyle.Regular);
        public static float yearLabelWidth = calculateYearLabel() + 2; // Add 2 pix margin.
        public static float monthLabelWidth = calculateMonthLabel() + 2; // Add 2 pix margin.
        public static float dayLabelWidth = calculateDayLabel() + 2;

        static int calculateYearLabel() { // Year numbers in string format are culture dependent. The longest year number must fit into the smallest large-tag.
            var image = new Bitmap(1, 1);
            Graphics g = Graphics.FromImage(image);
            return (int)(g.MeasureString("9999", labelFont).Width + 0.5);
        }

        static int calculateMonthLabel() { // Month names are culture dependent. The longest short month name must fit into the smallest large-tag.
            var image = new Bitmap(1, 1);
            Graphics g = Graphics.FromImage(image);
            int m = 0;
            for (int i = 1; i < 13; i++) {
                string mmm = new DateTime(2014, i, 1).ToString("MMM");
                int width = (int)(g.MeasureString(mmm, labelFont).Width + 0.5);
                if (width > m) m = width;
            }
            return m;
        }

        static int calculateDayLabel() { // Day numbers in string format are culture dependent. The longest day number must fit into the smallest large-tag.
            var image = new Bitmap(1, 1);
            Graphics g = Graphics.FromImage(image);
            int m = 0;
            for (int i = 10; i < 32; i++) {
                string dd = new DateTime(2014, 1, i).ToString("d ");
                int width = (int)(g.MeasureString(dd, labelFont).Width + 0.5);
                if (width > m) m = width;
            }
            return m;
        }

        public HTimeScale(int targetCount, DataRow row, Template tpl)
            : base(targetCount, row, tpl) {
            Height = height;
            borderColor = Color.Orange; // Default value
            borderWeight = 0;
        }

        public override void ResetUpdated() { // Called by kernel at refresh to set internal dependent properties to not-updated.
            tsUpdated = false;
            IntervalCount = 1000; // Testing BorderPixels and BorderValues will gradually decrease IntervalCount.
            base.ResetUpdated();
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Constant",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Height is computed automatically.", Position = 6)]
        public override int Height { get { Height = height;  return height; } } // Height is constant. Make sure it looks updated.

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Calculated from BorderPixels (internal formula)",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Width is calculated from BorderPixels.", Position = 9)]
        public override int Width {
            get { UpdateTimeScale(WidthID); return width; }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Property, " ' Calculated from BorderPixels (internal formula)",
        PropertyCategory = "Common", ReadOnly = true, ShowInitially = false, ShortHelp = "Right is calculated from BorderPixels.", Position = 10)]
        public override int Right {
            get { UpdateTimeScale(RightID); return right; }
        }

        public static int BorderValuesID; // Property index. Set at initialization in ReflectDLL. 
        public DateTime[] borderValues = new DateTime[0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, "Now()-100, Now(), Now()+100", ShowInitially = true, Position = 1,
            ShortHelp = "The points in time of the borders. Example: 100 days before and after Now: Now()-100, Now(), Now()+100")]
        public DateTime[] BorderValues {
            get {
                CmpProp borderValuesProp = propertyValues[BorderValuesID];
                if (borderValuesProp.state >= State.Updated) return borderValues;
                TplProperty targetProp = tpl.valPropArray[BorderValuesID];

                object[] newList = EvalListProp(targetProp, borderValuesProp, borderValues);
                if (newList.Length < 2) {
                    VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "BorderValues",
                    "There must be at least two BorderValues", null, this.Index());
                    borderValues = new DateTime[] { SystemItems.Date().AddDays(-100), SystemItems.Date(), SystemItems.Date().AddDays(100) };
                    newList = new object[] { borderValues[0], borderValues[1], borderValues[2] };
                    valueExpressions = DummyList(3);
                } else if (newList.Length != borderValues.Length)
                    borderValues = new DateTime[newList.Length];
                for (int i = 0; i < borderValues.Length; i++) {
                    borderValues[i] = toDateTime(newList[i], SystemItems.Date());
                    if (i > 0 && borderValues[i - 1] >= borderValues[i]) {
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "BorderValues",
                        "BorderValues must be increasing. Number " + i + " is not larger than number " + (i - 1), null, this.Index());
                        borderValues[i] = borderValues[i-1].AddDays(1);
                    }
                }
                if (IntervalCount+1 > borderValues.Length) IntervalCount = borderValues.Length-1;
                return borderValues;
            }
            //set { borderValues = value; } // Elements can be set, not the entire list.
        }

        private static ExpressionList DummyList(int count) { // Used as default when BorderValues or BorderPixels are garbage.
            ExpressionList dummy = new ExpressionList();
            for (int i = 0; i < count; i++) {
                Expression item = new Expression(); item.Init = true;
                dummy.Add(item);
            }
            return dummy;
        }

        public static int BorderPixelsID; // Property index. Set at initialization in ReflectDLL. 
        public int [] borderPixels = new int [0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, "0, 100, 200", ShowInitially = true, Position = 2,
            ShortHelp = "The pixel positions of the borders relative to my Left, e.g. 0, 100, 200")]
        public int [] BorderPixels {
            get {
                CmpProp borderPixelsProp = propertyValues[BorderPixelsID];
                if (borderPixelsProp.state >= State.Updated) return borderPixels;
                TplProperty targetProp = tpl.valPropArray[BorderPixelsID];

                object[] newList = EvalListProp(targetProp, borderPixelsProp, borderPixels);
                if (newList.Length < 2) {
                    VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "BorderPixels",
                    "There must be at least two BorderPixels", null, this.Index());
                    borderPixels = new int [] { 0, 100, 200 }; // Default value.
                    newList = new object[] { 0, 100, 200 };
                    pixelExpressions = DummyList(3);
                } else if (newList.Length != borderPixels.Length)
                    borderPixels = new int[newList.Length];
                for (int i = 0; i < borderPixels.Length; i++) {
                    borderPixels[i] = toInt32(newList[i], 0);
                    if (i > 0 && borderPixels[i - 1] >= borderPixels[i]) {
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "BorderValues",
                        "BorderPixels must be increasing. Number " + i + " is not larger than number " + (i - 1), null, this.Index());
                        borderPixels[i] = borderPixels[i - 1] + 100;
                    }
                }
                if (IntervalCount + 1 > borderPixels.Length) IntervalCount = borderPixels.Length - 1;
                return borderPixels;
            }
            // set {} // Elements can be set, not the entire list.
        }

        public static int BackColorsID; // Property index. Set at initialization in ReflectDLL. 
        public Color[] backColors = new Color [0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, "", ShowInitially = false, Position = 3,
            ShortHelp = "Colors of the zoom areas, e.g. Color.LightGray, Color.Green, Color.Yellow. Default: White.")]
        public Color [] BackColors {
            get {
                CmpProp backColorsProp = propertyValues[BackColorsID];
                if (backColorsProp.state >= State.Updated) return backColors;
                TplProperty targetProp = tpl.valPropArray[BackColorsID];

                object[] newList = EvalListProp(targetProp, backColorsProp, backColors);
                if (newList.Length != backColors.Length)
                    backColors = new Color[newList.Length]; 
                for (int i = 0; i < newList.Length; i++) {
                    backColors[i] = toColor(newList[i], Color.White);
                }
                return backColors; 
            }
            //set { ... = value; } // Elements can be set, not the entire list.
        }

        public static int LargeTagColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color largeTagColor = Color.Black;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 4,
            ShortHelp = "Color of the large tick marks Default: Color.Black.")]
        public Color LargeTagColor {
            get {
                CmpProp largeTagColorProp = this.propertyValues[LargeTagColorID];
                if (largeTagColorProp.state >= State.Updated) return largeTagColor;
                LargeTagColor = toColor(EvalProp(LargeTagColorID, Color.Black), Color.Black);
                return largeTagColor; 
            }
            set { largeTagColor = value; SetUpdated(ref propertyValues[LargeTagColorID]); }
        }

        public static int SmallTagColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color smallTagColor = Color.Black;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 5,
            ShortHelp = "Color of the small tick marks Default: Color.Black.")]
        public Color SmallTagColor {
            get {
                CmpProp smallTagColorProp = this.propertyValues[SmallTagColorID];
                if (smallTagColorProp.state >= State.Updated) return smallTagColor;
                SmallTagColor = toColor(EvalProp(SmallTagColorID, Color.Black), Color.Black);
                return smallTagColor;
            }
            set { smallTagColor = value; SetUpdated(ref propertyValues[SmallTagColorID]); }
        }

        public static int ExtTagCanvasesID; // Property index. Set at initialization in ReflectDLL. 
        public ContainerComponent [] extTagCanvases = new ContainerComponent[0];
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 10,
            ShortHelp = "Canvases where tag lines will appear, e.g. grid1!Cell(1, 1), grid1!Cell(1,2). Tag lines have zOrder= -10")]
        public ContainerComponent[] ExtTagCanvases {
            get {
                CmpProp extTagCanvasesProp = this.propertyValues[ExtTagCanvasesID];
                if (extTagCanvasesProp.state >= State.Updated) return extTagCanvases;
                TplProperty targetProp = tpl.valPropArray[ExtTagCanvasesID];

                object[] newList = EvalListProp(targetProp, extTagCanvasesProp, extTagCanvases);
                if (newList.Length != extTagCanvases.Length)
                    extTagCanvases = new ContainerComponent[newList.Length];
                for (int i = 0; i < extTagCanvases.Length; i++) {
                    extTagCanvases[i] = toCanvas(newList[i], null);
                }
                return extTagCanvases;
            }
            //set { ... = value; } // Elements can be set, not the entire list.
        }

        public static int ExtLargeTagColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color extLargeTagColor = Color.Gray;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 11,
            ShortHelp = "Color of the large tag lines, e.g. Color.Red", LongHelp = "Color of the large tick lines, e.g. Color.Red")]
        public Color ExtLargeTagColor {
            get {
                CmpProp extLargeTagColorProp = this.propertyValues[ExtLargeTagColorID];
                if (extLargeTagColorProp.state >= State.Updated) return extLargeTagColor;
                ExtLargeTagColor = toColor(EvalProp(ExtLargeTagColorID, Color.Gray), Color.Gray);
                return extLargeTagColor;
            }
            set { extLargeTagColor = value; SetUpdated(ref propertyValues[ExtLargeTagColorID]); }
        }

        public static int ExtLargeTagWeightID; // Property index. Set at initialization in ReflectDLL. 
        public int extLargeTagWeight = 1;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 12,
            ShortHelp = "Width of the large tag lines, e.g. 2. Make width zero to avoid the lines.")]
        public int ExtLargeTagWeight {
            get {
                CmpProp extLargeTagWeightProp = this.propertyValues[ExtLargeTagWeightID];
                if (extLargeTagWeightProp.state >= State.Updated) return extLargeTagWeight;
                ExtLargeTagWeight = toPosInt32(EvalProp(ExtLargeTagWeightID, 1), 1);
                return extLargeTagWeight;
            }
            set { extLargeTagWeight = value; SetUpdated(ref propertyValues[ExtLargeTagWeightID]); }
        }

        public static int ExtLargeTagStyleID; // Property index. Set at initialization in ReflectDLL. 
        public DashStyle extLargeTagStyle = DashStyle.Solid;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 13,
            ShortHelp = "Style of the large tag lines, e.g.\"Dot\", \"Dash\", \"DashDot\". Default \"Solid\"")]
        public string ExtLargeTagStyle {
            get {
                CmpProp extLargeTagStyleProp = this.propertyValues[ExtLargeTagStyleID];
                if (extLargeTagStyleProp.state < State.Updated)
                    ExtLargeTagStyle = toString(EvalProp(ExtLargeTagStyleID, "Solid"), "Solid");
                return extLargeTagStyle.ToString(); 
            }
            set {
                if (String.Compare(value.Trim().ToLower(), "dash", true) == 0)
                    extLargeTagStyle = DashStyle.Dash;
                else if (String.Compare(value.Trim().ToLower(), "dashdot", true) == 0)
                    extLargeTagStyle = DashStyle.DashDot;
                else if (String.Compare(value.Trim().ToLower(), "dot", true) == 0)
                    extLargeTagStyle = DashStyle.Dot;
                else
                    extLargeTagStyle = DashStyle.Solid;
                SetUpdated(ref propertyValues[ExtLargeTagStyleID]);
            }
        }

        public static int ExtSmallTagColorID; // Property index. Set at initialization in ReflectDLL. 
        public Color extSmallTagColor = Color.LightGray;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 21,
            ShortHelp = "Color of the small tag lines, e.g. Color.Green")]
        public Color ExtSmallTagColor {
            get {
                CmpProp extSmallTagColorProp = this.propertyValues[ExtSmallTagColorID];
                if (extSmallTagColorProp.state >= State.Updated) return extSmallTagColor;
                ExtSmallTagColor = toColor(EvalProp(ExtSmallTagColorID, Color.LightGray), Color.LightGray);
                return extSmallTagColor;
            }
            set { extSmallTagColor = value; SetUpdated(ref propertyValues[ExtSmallTagColorID]); }
        }

        public static int ExtSmallTagWeightID; // Property index. Set at initialization in ReflectDLL. 
        public int extSmallTagWeight = 1;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 22,
        ShortHelp = "Width of the small tick lines, e.g. 1. Make width zero to avoid the lines.")]
        public int ExtSmallTagWeight {
            get {
                CmpProp extSmallTagWeightProp = this.propertyValues[ExtSmallTagWeightID];
                if (extSmallTagWeightProp.state >= State.Updated) return extSmallTagWeight;
                ExtSmallTagWeight = toPosInt32(EvalProp(ExtSmallTagWeightID, 1), 1);
                return extSmallTagWeight;
            }
            set { extSmallTagWeight = value; SetUpdated(ref propertyValues[ExtSmallTagWeightID]); }
        }

        public static int ExtSmallTagStyleID; // Property index. Set at initialization in ReflectDLL. 
        public DashStyle extSmallTagStyle = DashStyle.Solid;
        [VisPropAtt(VisPropAtt.PropertyType.Property, null, ShowInitially = false, Position = 23,
            ShortHelp = "Style of the small tag lines, e.g.\"Dot\", \"Dash\", \"DashDot\". Default \"Solid\"")]
        public string ExtSmallTagStyle {
            get {
                CmpProp extSmallTagStyleProp = this.propertyValues[ExtSmallTagStyleID];
                if (extSmallTagStyleProp.state < State.Updated)
                    ExtSmallTagStyle = toString(EvalProp(ExtSmallTagStyleID, "Solid"), "Solid");
                return extSmallTagStyle.ToString();
            }
            set {
                if (String.Compare(value.Trim().ToLower(), "dash", true) == 0)
                    extSmallTagStyle = DashStyle.Dash;
                else if (String.Compare(value.Trim().ToLower(), "dashdot", true) == 0)
                    extSmallTagStyle = DashStyle.DashDot;
                else if (String.Compare(value.Trim().ToLower(), "dot", true) == 0)
                    extSmallTagStyle = DashStyle.Dot;
                else
                    extSmallTagStyle = DashStyle.Solid;
                SetUpdated(ref propertyValues[ExtSmallTagStyleID]);
            }
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -9, PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "Occurs when user has dragged a ribbon or a border.")]
        public void BordersChanged(object sender, EventArgs e) {
            this.ExecuteHandler("BordersChanged", true);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Eventhandler, null, Position = -10, PropertyCategory = "EventHandler", ReadOnly = false,
            ShortHelp = "Occurs repeatedly while user drags a ribbon or a border.")]
        public void BordersChanging(object sender, EventArgs e) {
            //Debug.Print("HTimeScale.BordersChanging");
            this.ExecuteHandler("BordersChanging", true);
        }

        int dragInterval = 0; // During MouseMove, the interval where the mouse is.
        DateTime dragFromTime; // During MouseMove, the from time when dragging started.
        DateTime dragToTime; // During MouseMove, the to time when dragging started.
        int dragPoint; // During MouseMove, the logical X-pos where dragging started.
        DragType dragType = DragType.None; // During MouseMove, whether dragging across the left border, right or no drag possible.

        int oldDragInterval = 0; // Set at stat of dragging. For undo of the latest drag.
        DateTime oldDragFromTime; // For undo of the latest drag.
        DateTime oldDragToTime; // For undo of the latest drag.
        DragType oldDragType = DragType.None; // For undo of the latest drag.

        double minTimeInterval = 0.06; // Minimum number of days in an interval. (around 1.5 hours)
        int indiX; // The X-position of the drag indicator.
        public enum DragType { Left, Right, None };

        public override void MouseMove(object sender, MouseEventArgs e) {
            if (this.Moving == Move.No) { // Not dragging. Show drag indicator arrow.
                updateMouseVariables(e);
                SetDragType(LogicalX(e.X));
                return;
            }
            if (this.Moving == Move.Init) {
                oldDragInterval = dragInterval; oldDragType = dragType;
                oldDragFromTime = Intervals[dragInterval].fromTime; oldDragToTime = Intervals[dragInterval].toTime;
                this.Moving = Move.Moved;
            }
            this.canvas.GuiInstance.Capture = true; // Allow dragging outside the time scale and even outside the Form.
            int newDragPoint = LogicalX(e.X); // Compensate for scrolling.
            if (newDragPoint == dragPoint) return; // Moved 0 pixels.
            Interval interval = Intervals[dragInterval];

            try {
                if (dragType == DragType.Left) {
                    DateTime newfromTime;
                    if (newDragPoint >= interval.toPx - 2) return; // Increase scales too much.
                    newfromTime = interval.toTime.AddDays(-(double)(interval.toPx - dragPoint) / (interval.toPx - newDragPoint) * (interval.toTime - dragFromTime).TotalDays);
                    if (newfromTime.AddDays(minTimeInterval) >= interval.toTime) return; // Min 0.5 hour interval.
                    if (dragInterval > 0) {
                        if (Intervals[dragInterval - 1].fromTime.AddDays(minTimeInterval) >= newfromTime) return; // Min 0.5 hour interval.
                        Intervals[dragInterval - 1].toTime = newfromTime;
                        //Debug.Print("interval " + (dragInterval - 1) + "  toTime: " + newfromTime.ToLongDateString());
                    }
                    borderValues[dragInterval] = newfromTime; // CreateIntervals get data from the properties.
                } else { // Across right.
                    DateTime newtoTime;
                    if (newDragPoint <= interval.fromPx + 2) return; // Increase scales too much.
                    newtoTime = interval.fromTime.AddDays(-(double)(interval.fromPx - dragPoint) / (interval.fromPx - newDragPoint) * (interval.fromTime - dragToTime).TotalDays);
                    if (newtoTime <= interval.fromTime.AddDays(minTimeInterval)) return; // Min 0.5 hour interval.
                    if (dragInterval < Intervals.Length - 1) {
                        if (Intervals[dragInterval + 1].toTime <= newtoTime.AddDays(minTimeInterval)) return; // Min 0.5 hour interval.
                        Intervals[dragInterval + 1].fromTime = newtoTime;
                    }
                    borderValues[dragInterval + 1] = newtoTime; // CreateIntervals get data from the properties.
                }
            }
            catch (Exception) { // Dragging too far. Just ignore.
            }
            CreateIntervals();
            canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(left), RelativeY(top), PhyS(width) + 2, PhyS(height) + 1)); // Border doesn't need repainting. 
            updateMouseVariables(e);
            this.ExecuteHandler("BordersChanging", true);
        }

        void SetDragType(int x) { // x is logical xpos of mouse.
            if (IntervalCount == 0) return; // Erroneous intervals.
            int oldIndiX = indiX;
            for (int i = 0; i < IntervalCount; i++) {
                if (x <= Intervals[i].toPx) {
                    dragInterval = i;
                    dragFromTime = Intervals[i].fromTime; dragToTime = Intervals[i].toTime;
                    break;
                }
            }
            if (!valueExpressions.expressions[dragInterval].Init && !valueExpressions.expressions[dragInterval + 1].Init) {
                dragType = DragType.None; return; // Dragging not possible
            }

            if (!valueExpressions.expressions[dragInterval].Init) dragType = DragType.Right; // Left side fixed. Drag across right.
            else if (!valueExpressions.expressions[dragInterval + 1].Init) dragType = DragType.Left; // Right side fixed. Drag across left.
            else if (x - Intervals[dragInterval].fromPx < Intervals[dragInterval].toPx - x)
                dragType = SystemItems.AltPressed ? DragType.Right : DragType.Left; // Dragging in left side.
            else
                dragType = ! SystemItems.AltPressed ? DragType.Right : DragType.Left; // Dragging in right side.

            // Find indicator X-pos
            if (dragType == DragType.Left)
                indiX = (dragInterval == 0 ? Intervals[dragInterval].fromPx + 2 : Intervals[dragInterval].fromPx -10);
            else
                indiX = (dragInterval == Intervals.Length - 1 ? Intervals[dragInterval].toPx - 22 : Intervals[dragInterval].toPx-10);
            if (oldIndiX == indiX) return; // No change.
            canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(oldIndiX), RelativeY(top + indiTop), PhyS(indiWidth), PhyS(indiHeight)));
            canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(indiX), RelativeY(top + indiTop), PhyS(indiWidth), PhyS(indiHeight)));
        }

        public override void MouseDown(object sender, MouseEventArgs e) {
            dragPoint = SystemItems.MouseX;
            SetDragType(LogicalX(e.X));
            if (dragType != DragType.None) Moving = Move.Init; // Dragging possible.
        }

        public override void MouseUp(object sender, MouseEventArgs e) {
            if (this.Moving == Move.Moved) {
                this.ExecuteHandler("BordersChanged", true);
            }
            this.Moving = Move.No;
            this.canvas.GuiInstance.Capture = false;
        }

        public override void MouseEnter(object sender, EventArgs e) {
            this.oldKeyFocus = GetKeyFocus(this);
            this.SetKeyFocus(); // In order to catch the Alt-key.
            this.canvas.GuiInstance.Cursor = Cursors.NoMoveHoriz;
            indiX = -1;
            this.ExecuteHandler("MouseEnter", true); // E.g. to show a popup.
        }

        public override void MouseLeave(object sender, EventArgs e) {
            this.canvas.GuiInstance.Cursor = Cursors.Arrow;
            this.canvas.GuiInstance.Capture = false;
            this.Moving = Move.No;
            if (oldKeyFocus != null) oldKeyFocus.SetKeyFocus();
            Canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(indiX), RelativeY(top + indiTop), PhyS(indiWidth), PhyS(indiHeight))); // Border doesn't need repainting. 
            this.dragType = DragType.None; indiX = -1; // Make sure it is a change when the cursor enters again.
            this.ExecuteHandler("MouseLeave", true);
        }

        public override void KeyDown(object sender, KeyEventArgs e) {
            if (e.Alt != SystemItems.AltPressed) { // Alt down generates KeyDown all the time.
                SystemItems.AltPressed = e.Alt; // SetDragType tests SystemItems, which hasn't yet been updated.
                SetDragType(SystemItems.MouseX);
                return;
            }
            base.KeyDown(sender, e);
        }

        public override void KeyUp(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Escape) { // Undo last drag.
                if (oldDragType == DragType.Left) {
                    borderValues[oldDragInterval] = oldDragFromTime;
                } else if (oldDragType == DragType.Right) {
                    borderValues[oldDragInterval + 1] = oldDragToTime;
                } else
                    return;
                CreateIntervals();
                dragType = DragType.None;
                canvas.GuiInstance.Invalidate(new System.Drawing.
                    Rectangle(RelativeX(left), RelativeY(top), PhyS(width) + 2, PhyS(height) + 1)); // Border doesn't need repainting. 
                this.ExecuteHandler("BordersChanged", true);
                return;
            }
            if (e.Alt != SystemItems.AltPressed) {
                SystemItems.AltPressed = e.Alt;
                SetDragType(SystemItems.MouseX);
            }

            base.KeyUp(sender, e);
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(dt) Pixel position for DateTime dt, e.g. Left: HPos(now()+10)")]
        public int HPos(object time) { //Relative to the timescale
            UpdateTimeScale(0);
            DateTime dt = DateTime.Now;
            try {
                dt = Convert.ToDateTime(time);
            }
            catch { return left; } // Default is Now().
            if (IntervalCount < 1) return left; 
            for (int i = 0; i < IntervalCount; i++) {
                if (dt < Intervals[i].toTime) return Intervals[i].TimeToPixel(dt); // Default is first interval.
            }
            return Intervals[IntervalCount - 1].TimeToPixel(dt); // Default is last interval.
        }

        [VisPropAtt(VisPropAtt.PropertyType.Method, null, ShowInitially = false,
            ShortHelp = "(x) DateTime corresponding to pixel position X on the Canvas.")]
        public DateTime TimeOfX(int pos) {
            UpdateTimeScale(0);
            if (IntervalCount < 1) return DateTime.Now;
            for (int i = 0; i < IntervalCount; i++) {
                if (pos < Intervals[i].toPx) return Intervals[i].PixelToTime(pos); // Default is first interval.
            }
            return Intervals[IntervalCount - 1].PixelToTime(pos); // Default is last interval.
        }

        // Check BorderValues and BorderPixels and find the number of correct time intervals.
        // Next, create the Intervals including their tags, and create the extended tick mark components on the canvases.
        bool tsUpdated = false;
        public void UpdateTimeScale(int sourceID) {
            if (tsUpdated) return;
            tsUpdated = true;
            valueExpressions = tpl.valPropArray[BorderValuesID].expr as ExpressionList;
            pixelExpressions = tpl.valPropArray[BorderPixelsID].expr as ExpressionList;

            CmpProp sourceProp = this.propertyValues[sourceID];
            if (sourceID != 0 && sourceProp.state < State.Updated) // SourceID == 0 when there is no source property to lock, e.g. HPos.
                sourceProp.state = State.Visited; // Start internal calculation. Catch cyclical references.
            object dummy = BorderValues; dummy = BorderPixels; dummy = ExtTagCanvases;
            dummy = Left; // Needed to translate border pixels to X-pos relative to canvas.
            if (borderValues.Length != borderPixels.Length) {
                VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, "BorderValues",
                    "BorderValues and BorderPixels must have the same number of elements", null, this.Index());
            }
            if (IntervalCount < 1) {
                Intervals = new Interval[0]; return; // There were serious errors. No intervals and nothing will be painted.
            }

            CreateIntervals();
            base.Width = borderPixels[IntervalCount];
            base.Right = left + width;
            base.Height = height;

            for (int iInterval = 0; iInterval < Intervals.Length - 1; iInterval++) {
                Interval itv = Intervals[iInterval];
                //if (itv.dragableBorder != null) 
                //    itv.dragableBorder.Dispose();
                itv.dragableBorder = new DragableBorder(itv.toPx, this, iInterval); // Will place themselves on the canvas.
            }

            foreach (ContainerComponent canvas in extTagCanvases) {
                new ExtTag(canvas, this); // Creates a VisComponent on the canvas. It will paint when the canvas is painted.
            }
        }

        public void CreateIntervals() { // Used when the time scale is created and when dragging the scale or border.  
            Intervals = new Interval[IntervalCount];
            for (int i = 0; i < IntervalCount; i++) { // The constructor determines the interval type and creates the small, large and top tags.
                Intervals[i] = new Interval(borderValues[i], borderPixels[i] + left, borderValues[i + 1], borderPixels[i + 1] + left);
            }
        }

        public override void Paint(PaintEventArgs e) { // Paint all the tags and their labels.
            if (!visible) return;
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            Pen borderPen = new Pen(borderColor, 3);
            Pen largeTagPen = new Pen(largeTagColor, 1);
            Pen smallTagPen = new Pen(smallTagColor, 1);
            Brush labelBrush = new SolidBrush(Color.Black);
            Brush indicatorBrush = new SolidBrush(Color.LightGreen);
            Pen indicatorPen = new Pen(Color.Black, 1);

            for (int iInterval = 0; iInterval < Intervals.Length; iInterval++) {
                Interval interval = Intervals[iInterval];
                // Paint background
                Rectangle rec = new Rectangle(RelativeX(interval.fromPx), RelativeY(top), interval.toPx - interval.fromPx + 1, PhyS(height));
                Brush backBrush;
                if (iInterval < backColors.Length)
                    backBrush = new SolidBrush(backColors[iInterval]);
                else
                    backBrush = new SolidBrush(backColor);
                e.Graphics.FillRectangle(backBrush, rec);

                if (interval.intervalType.TopLabel != null)
                    for (int iTop = 0; iTop < interval.topTags.Count - 1; iTop++) {
                        Tag topTag = interval.topTags[iTop];
                        e.Graphics.DrawLine(largeTagPen, RelativeX(topTag.pixel), RelativeY(top), RelativeX(topTag.pixel), RelativeY(top + topHeight));
                        int labelWidth = interval.topTags[iTop + 1].pixel - topTag.pixel +1; // +1 to better indicate that a char is clipped.
                        if (labelWidth > 0) // DrawString draws anyway when labelWidth is 0.
                            e.Graphics.DrawString(topTag.dt.ToString(interval.intervalType.TopLabel), labelFont, labelBrush, 
                            new Rectangle(RelativeX(topTag.pixel + 1), RelativeY(top + 2), labelWidth, largeHeight), new StringFormat(StringFormatFlags.NoWrap));
                    }

                for (int iLarge = 0; iLarge < interval.largeTags.Count - 1; iLarge++) {
                    Tag largeTag = interval.largeTags[iLarge];
                    e.Graphics.DrawLine(largeTagPen, RelativeX(largeTag.pixel), RelativeY(top + topHeight), RelativeX(largeTag.pixel), RelativeY(top + height));

                    int labelWidth = interval.largeTags[iLarge + 1].pixel - largeTag.pixel +1;
                    if (labelWidth > 0) // DrawString draws anyway when labelWidth is 0.
                        e.Graphics.DrawString(largeTag.dt.ToString(interval.intervalType.LargeTagLabel), labelFont, labelBrush,
                        new Rectangle(RelativeX(largeTag.pixel + 1), RelativeY(top + topHeight + 2), labelWidth, largeHeight), new StringFormat(StringFormatFlags.NoWrap));
                }

                foreach (Tag tag in interval.smallTags)
                    e.Graphics.DrawLine(smallTagPen, RelativeX(tag.pixel), RelativeY(top + topHeight + largeHeight), RelativeX(tag.pixel), RelativeY(top + height));
            }
            // Draw line below all top labels:
            e.Graphics.DrawLine(smallTagPen, RelativeX(left), RelativeY(top + topHeight), RelativeX(left + width), RelativeY(top + topHeight));
            // Draw movable borders.
            for (int iInterval = 0; iInterval < Intervals.Length - 1; iInterval++) {
                Interval interval = Intervals[iInterval];
                e.Graphics.DrawLine(borderPen, RelativeX(interval.toPx), RelativeY(top), RelativeX(interval.toPx), RelativeY(top + height));
            }

            int bw = PhyS(borderWeight);
            if (borderWeight > 0) { // Draw border:
                int offset = (borderWeight + 1) / 2; // Make sure the border surrounds the component.
                e.Graphics.DrawRectangle(new Pen(borderColor, bw), RelativeX(left) - offset, RelativeY(top) - offset, PhyS(width) + bw, PhyS(height) + bw);
            }

            // Draw drag indicator:
            if (dragType == DragType.None) return;
            PointF[] points = { 
                new PointF(RelativeX(indiX), RelativeY(top+ indiTop+6)),
                new PointF(RelativeX(indiX+5), RelativeY(top+ indiTop +1)),
                new PointF(RelativeX(indiX+5), RelativeY(top+ indiTop +4)),
                new PointF(RelativeX(indiX+15), RelativeY(top+ indiTop +4)),
                new PointF(RelativeX(indiX+15), RelativeY(top+ indiTop +1)), 
                new PointF(RelativeX(indiX+20), RelativeY(top+ indiTop+6)),
                new PointF(RelativeX(indiX+15), RelativeY(top+ indiTop+11)),
                new PointF(RelativeX(indiX+15), RelativeY(top+ indiTop+8)),
                new PointF(RelativeX(indiX+5), RelativeY(top+ indiTop+8)), 
                new PointF(RelativeX(indiX+5), RelativeY(top+ indiTop+11)),
                new PointF(RelativeX(indiX), RelativeY(top+ indiTop+6))
            };
            e.Graphics.DrawLines(indicatorPen, points);
            e.Graphics.FillPolygon(indicatorBrush, points);
        }
    }

    //--------------------------------------------------
    public enum UnitType { Minute, Hour, Day, Month, Year, None} // None means don't show anything on this level.

    //------------------------------------------------
    public struct IntervalType {
        public double minPxPerDay; // Intervals with less than these pixels per day should use a lower granularity.
        public UnitType sUnit; // Whether the small unit is in hours, days, etc.
        public int sUnitCount; // The number of units per small tag.
        public int sUnitRepeat; // The number of small tags per large tag. 
        public UnitType lUnit; // Whether the large unit is in hours, days, etc.
        public int lUnitCount; // The number of units per large tag. 
        public int lUnitRepeat; // The number of large tags per top tag. If 0, the number is the number of days in the month.
        // If there is no top tag, repeat the large tag a huge number of times. Loop will stop when toTime is reached.
        public UnitType tUnit; // Whether the top unit is a day, a month or a year.
        public string LargeTagLabel, TopLabel; // Examples: "MMM", "yyyy", null. Null means show nothing. Small tags never have a label.

        public IntervalType(double maxDaysPerPixel, UnitType sUnit, int sUnitCount, int sUnitRepeat, UnitType lUnit, int lUnitCount, int lUnitRepeat, UnitType tUnit, string LargeTagLabel, string TopLabel) {
            this.minPxPerDay = maxDaysPerPixel;
            this.sUnit = sUnit; this.sUnitCount = sUnitCount; this.sUnitRepeat = sUnitRepeat;
            this.lUnit = lUnit; this.lUnitCount = lUnitCount; this.lUnitRepeat = lUnitRepeat; this.tUnit = tUnit;
            this.LargeTagLabel = LargeTagLabel;
            this.TopLabel = TopLabel;
        }
    }

    //-----------------------------------------------
    public class Interval {
        public DateTime fromTime, toTime;
        public int fromPx, toPx; // The X-pos, i.e. relative to the left of the canvas.
        public const float minPxPerTag = 7;
        double pxsPerDay;
        public IntervalType intervalType;
        public List<Tag> topTags = new List<Tag>(10);
        public List<Tag> largeTags = new List<Tag>(20);
        public List<Tag> smallTags = new List<Tag>(50);
        public DragableBorder dragableBorder;

        // Each interval will get an interval type. The interval type specifies the scale (minimum pixels per day). The unit for small tags (hour, day, month ...) followed by
        // the number of units per tag and the number of tags per large tag. The unit for large tags followed by the number of units per tag and the number of tags per top tag 
        // (0 means according to the number of days in the month). The format string for large tags (e.g. "MMM") and the format string for top tags (null when there is no top label). 
        // This is a "catalog" of available interval types. Must have decreasing pixels per day (decreasing resolution).
        public static IntervalType[] intervalTypes = {
        new IntervalType(minPxPerTag*24*6, UnitType.Minute, 15, 4, UnitType.Hour, 1, 24, UnitType.Day, "HH:mm", "d.MMM yyyy"), // Small tag: 15 minutes.
        new IntervalType(minPxPerTag*24, UnitType.Hour, 1, 6, UnitType.Hour, 6, 4, UnitType.Day, "HH:mm", "d.MMM yyyy"), // Small tag: 1 hour.
        new IntervalType(minPxPerTag*4, UnitType.Hour, 6, 4, UnitType.Day, 1, 0, UnitType.Month, "d ", "MMM yyyy"), // Small tag: 6 hours. 
        new IntervalType(HTimeScale.dayLabelWidth/1, UnitType.Hour, 12, 2, UnitType.Day, 1, 0, UnitType.Month, "d ", "MMM yyyy"), // Small tag: 12 hours.
        new IntervalType(Math.Max(HTimeScale.dayLabelWidth/5, minPxPerTag), UnitType.Day, 1, 5, UnitType.Day, 5, 0, UnitType.Month, "d ", "MMM yyyy"), // Small tag: 1 day.
        new IntervalType(HTimeScale.dayLabelWidth/10, UnitType.Day, 5, 2, UnitType.Day, 10, 0, UnitType.Month, "d ", "MMM yyyy"), // Small tag: 5 days.
        new IntervalType(HTimeScale.monthLabelWidth/30, UnitType.Day, 10, 3, UnitType.Month, 1, 12, UnitType.Year, "MMM", "yyyy"), // Small tag: 10 days.
        new IntervalType(HTimeScale.monthLabelWidth/90, UnitType.Month, 1, 3, UnitType.Month, 3, 4, UnitType.Year, "MMM", "yyyy"), // Small tag: 1 month.
        new IntervalType(minPxPerTag/90, UnitType.Month, 3, 4, UnitType.Year, 1, 1000, UnitType.None, "yyyy", null), // Small tag: 3 months.
        new IntervalType(minPxPerTag/365, UnitType.Year, 1, 5, UnitType.Year, 5, 1000, UnitType.None, "yyyy", null), // Small tag: 1 year.
        new IntervalType(HTimeScale.yearLabelWidth/(365*10), UnitType.Year, 5, 2, UnitType.Year, 10, 1000, UnitType.None, "yyyy", null), // Small tag: 5 years.
        new IntervalType(minPxPerTag/(365*10), UnitType.Year, 10, 5, UnitType.Year, 50, 1000, UnitType.None, "yyyy", null), // Small tag: 10 years.
        new IntervalType(minPxPerTag/(365*20), UnitType.Year, 20, 5, UnitType.Year, 100, 1000, UnitType.None, "yyyy", null), // Small tag: 20 years.
        new IntervalType(minPxPerTag/(365*50), UnitType.Year, 50, 4, UnitType.Year, 200, 1000, UnitType.None, "yyyy", null), // Small tag: 50 years.
        new IntervalType(minPxPerTag/(365*100), UnitType.Year, 100, 5, UnitType.Year, 500, 1000, UnitType.None, "yyyy", null), // Small tag: 100 years.
        };

        public Interval(DateTime fromTime, int fromPx, DateTime toTime, int toPx) {  // Pixels relative to canvas left.
            this.fromTime = fromTime; this.toTime = toTime;
            this.fromPx = fromPx; this.toPx = toPx;
            pxsPerDay = (toPx - fromPx)/(toTime - fromTime).TotalDays;
            DateTime curTopTag, curLargeTag, curSmallTag; // DateTime for the start of the tag.
            for (int i = 0; i < intervalTypes.Length; i++) { // Find the interval type with the finest, suitable granularity.
                intervalType = intervalTypes[i];
                if (intervalType.minPxPerDay <= pxsPerDay) break;
            }
            //Debug.Print("Small tag: " + minPxPerTag / intervalType.minPxPerDay);

            curTopTag = topStart(); // Round down to nearest top unit.
            topTags.Add(new Tag(fromPx, fromTime)); // Add top tag at the border.
            largeTags.Add(new Tag(fromPx, fromTime)); // Add large tag at the border.
            while (curTopTag < toTime) { // Top tag loop.
                if (curTopTag > fromTime) 
                    topTags.Add(new Tag(TimeToPixel(curTopTag), curTopTag));
                DateTime nextTopTag = AddUnit(curTopTag, intervalType.tUnit, 1);

                curLargeTag = curTopTag; // Make a large tag also for the top tag.
                int repeatLarge = intervalType.lUnitRepeat;
                if (repeatLarge <= 0) repeatLarge = DateTime.DaysInMonth(curTopTag.Year, curTopTag.Month)/intervalType.lUnitCount;
                for (int iLarge = 0; iLarge <= repeatLarge; iLarge++ ) { // Large tag loop. Note: 5 periods have 6 end points.
                    if (curLargeTag > fromTime && curLargeTag < toTime && curLargeTag < nextTopTag) 
                        largeTags.Add(new Tag(TimeToPixel(curLargeTag), curLargeTag));
                    DateTime nextLargeTag = AddUnit(curLargeTag, intervalType.lUnit, intervalType.lUnitCount);

                    curSmallTag = AddUnit(curLargeTag, intervalType.sUnit, intervalType.sUnitCount);
                    for (int iSmall = 1; iSmall < intervalType.sUnitRepeat; iSmall++ ) { // Small tag loop. No small tag for the large tag.
                        if (curSmallTag > fromTime && curSmallTag < toTime && curSmallTag < nextLargeTag && curSmallTag < nextTopTag) 
                            smallTags.Add(new Tag(TimeToPixel(curSmallTag), curSmallTag));
                        curSmallTag = AddUnit(curSmallTag, intervalType.sUnit, intervalType.sUnitCount); ;
                    }
                    curLargeTag = nextLargeTag;
                    if (curLargeTag >= toTime) break; // When there is no top tag, lUnitRepeat and repeatLarge are huge. Repeat until toTime.
                }
                curTopTag = nextTopTag;
            }
            topTags.Add(new Tag(toPx, toTime)); // Add top tag at the end of the interval.
            largeTags.Add(new Tag(toPx, toTime)); // Add large tag at the end of the interval.
        }

        DateTime topStart() {
            switch (intervalType.tUnit) {
            case UnitType.Day: return new DateTime(fromTime.Year, fromTime.Month, fromTime.Day); 
            case UnitType.Month: return new DateTime(fromTime.Year, fromTime.Month, 1);
            case UnitType.Year: return new DateTime(fromTime.Year, 1, 1);
            default: {
                int year = fromTime.Year - fromTime.Year % (intervalType.lUnitCount);
                if (year < 1) year = 1; // DateTime starts with #01-01-01 0:0#
                return new DateTime(year, 1, 1); 
                // UnitType.None. Use large tag as unit.
            }
            }
        }

        DateTime AddUnit(DateTime dt, UnitType unitType, int unitCount) {
            try {
                switch (unitType) {
                case UnitType.Minute: return dt.AddMinutes(unitCount);
                case UnitType.Hour: return dt.AddHours(unitCount);
                case UnitType.Day: return dt.AddDays(dt.Day == 1 && unitCount > 1 ? unitCount - 1 : unitCount);  // Days are counted 1, 5, 10
                case UnitType.Month: return dt.AddMonths(unitCount);
                case UnitType.Year: return dt.AddYears(unitCount);
                default: return this.toTime.AddDays(1); // UnitType.None. Make sure the top loop terminates immediately.
                }
            }
            catch (Exception) {
                return dt; // Ignore excessive values.
            }
        }

        public int TimeToPixel(DateTime dt) { // Assumes dt is inside this interval.
            return (int)((dt - fromTime).TotalDays * pxsPerDay + fromPx); // TotalDays includes fractions.
        }

        public DateTime PixelToTime(int px) { // Assumes px is inside this interval.
            return fromTime.AddDays((px - fromPx) / pxsPerDay); 
        }
    }

    //-------------------------------------------------------
    public struct Tag {
        public Tag(int px, DateTime dt) { this.pixel = px; this.dt = dt; }
        public int pixel; // The X-pos, i.e. relative to the left of the canvas.
        public DateTime dt; 
    } 

    //--------------------------------------------------------
    public class DragableBorder : VisComponent { // Catches border drags.. 
        HTimeScale timeScale;
        int interval;
        public DragableBorder(int xPos, HTimeScale timeScale, int interval) {
            this.tpl = timeScale.tpl;

            propertyValues = timeScale.propertyValues; // Share properties.
            //propertyValues = new CmpProp[tpl.valPropArray.Length];
            //for (int pi = 0; pi < tpl.valPropArray.Length; pi++) {
            //    propertyValues[pi] = new CmpProp(tpl.valPropArray[pi], this);
            //    propertyValues[pi].state = State.Updated;
            //}
            this.Anonymous = true;
            this.zOrder = timeScale.ZOrder + 1; // Position in front of timeScale..
            this.timeScale = timeScale;
            this.interval = interval;
            this.top = timeScale.Top; 
            this.left = xPos - 3; 
            this.height = timeScale.height;
            this.width = 5; // To catch the mouse.
            this.movable = true;
            this.AddToCanvas(timeScale.Canvas); // Will evaluate Top, etc. to resolve internal dependencies.
            //Debug.Print("Canvas Controls: " + timeScale.canvas.GuiInstance.Controls.Count);
        }

        public override void MouseMove(object sender, MouseEventArgs e) {
            int bw = PhyS(borderWeight);
            if (this.Moving == Move.No) return; // Not dragging.
            this.Moving = Move.Moved;
            this.canvas.GuiInstance.Capture = true;
            if (! timeScale.pixelExpressions.expressions[interval + 1].Init) return; // BorderPixels cannot be changed.
            if (timeScale.borderPixels[interval] + timeScale.left > LogicalX(e.X) - Interval.minPxPerTag) 
                return; // Moved too far left.
            if (timeScale.borderPixels[interval + 2] + timeScale.left < LogicalX(e.X) + Interval.minPxPerTag) 
                return; // Moved toofar right.
            this.Left = LogicalX(e.X) - 3; // Move myself.
            timeScale.borderPixels[interval + 1] = LogicalX(e.X) - timeScale.left; // Compensate for scrolling and borderPixels being relative to time scale.
            timeScale.CreateIntervals();
            timeScale.Canvas.GuiInstance.Invalidate(new System.Drawing.
                Rectangle(RelativeX(timeScale.left) - bw / 2, RelativeY(timeScale.top - borderWeight / 2), PhyS(timeScale.width) + bw + 1, PhyS(timeScale.height) + bw + 1));
            updateMouseVariables(e);
            timeScale.ExecuteHandler("BordersChanging", true);
        }

        public override void MouseUp(object sender, MouseEventArgs e) {
            if (this.Moving == Move.Moved) {
                timeScale.ExecuteHandler("BordersChanged", true);
            }
            this.Moving = Move.No;
            this.canvas.GuiInstance.Capture = false;
        }

        public override void MouseEnter(object sender, EventArgs e) {
            this.canvas.GuiInstance.Cursor = Cursors.VSplit;
        }

        public override void MouseLeave(object sender, EventArgs e) {
            //Debug.Print("MouseLeave: " + this.Moving.ToString());
            this.canvas.GuiInstance.Cursor = Cursors.Arrow;
            this.Moving = Move.No;
        }
    }
    //--------------------------------------------------------
    public class ExtTag : VisComponent { // One anonymous component. Paints all the vertical tick lines on a canvas. 
        public ExtTag(ContainerComponent extTickCanvas, HTimeScale timeScale) { 
            // The canvas to put the tick marks on and the time scale holding the properties.
            this.tpl = extTickCanvas.tpl; // Borrow template from the canvas. Sets the Canvas property of Me.

            propertyValues = timeScale.propertyValues; // Share properties.
            this.Anonymous = true;
            this.zOrder = -10; // Paint behind the default zOrder.
            this.extTickCanvas = extTickCanvas;
            this.timeScale = timeScale;
            this.top = 0; 
            this.left = 0; // Canvas.Paint will paint an Anonymous component independent of its position. ExtTag.Paint optiimizes as needed.
            this.height = 0; // Not used, but assign canvas refers to it. Final virtual height is not available until paint time. 
            this.width = 0; 
            this.AddToCanvas(extTickCanvas); 
        }

        ContainerComponent extTickCanvas;
        HTimeScale timeScale;

        public override void Paint(PaintEventArgs e) {
            Tag tag;
            //Debug.Print("ExtTag paint");
            Rectangle r = e.ClipRectangle;
            e.Graphics.SmoothingMode = SmoothingMode.HighSpeed;
            Pen extLargeTagPen = new Pen(timeScale.extLargeTagColor);
            extLargeTagPen.Width = PhyS(timeScale.extLargeTagWeight);
            extLargeTagPen.DashStyle = timeScale.extLargeTagStyle;

            Pen extSmallTagPen = new Pen(timeScale.extSmallTagColor);
            extSmallTagPen.Width = PhyS(timeScale.extSmallTagWeight);
            extSmallTagPen.DashStyle = timeScale.extSmallTagStyle;
            int finalHeight = PhyS(Math.Max(canvas.height, canvas.virtualHeight + canvas.paddingBottom));
            int relX;

            if (timeScale.extLargeTagWeight > 0) //paint large extending tick marks
                foreach (Interval interval in timeScale.Intervals)
                    for (int i = 0; i < interval.largeTags.Count - 1; i++) { // The last largeTag is overlapping the first tag in the next interval.
                        tag = interval.largeTags[i];
                        relX = RelativeX(tag.pixel);
                        if (relX + extLargeTagPen.Width < r.X || relX - extLargeTagPen.Width > r.Right) continue;
                        e.Graphics.DrawLine(extLargeTagPen, relX, RelativeY(0), relX, RelativeY(finalHeight));
                    }

            if (timeScale.extSmallTagWeight > 0) //paint small extending tick marks
                foreach (Interval interval in timeScale.Intervals)
                    for (int i = 0; i < interval.smallTags.Count; i++) { // No overlap.
                        tag = interval.smallTags[i];
                        relX = RelativeX(tag.pixel);
                        if (relX + extSmallTagPen.Width < r.X || relX - extSmallTagPen.Width > r.Right) continue;
                        e.Graphics.DrawLine(extSmallTagPen, relX, RelativeY(0), relX, RelativeY(finalHeight));
                    }
        }
    }    
}
