﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Linq;
using VisTool;
using uVisStudio.VisDesigner;
using uVisStudio.VisExplorer;
using uVisStudio.VisPropertyBox;

namespace uVisStudio {
    public class Controller {
        /* Contains a method for each kind of design change, e.g. TemplateDrawn and ChangePropertyFormula.
         * The methods are called when the designer changes something, e.g. through the property grid.. 
         * Each method creates a command object, calls its Execute method, and if successfull, adds the command object to the 
         * UndoManager's list of Commands. Examples of command objects: TemplateDrawnCommand and ChangePropertyCommand.
         * 
         * The last command can be undone through a call to the UndoManager's Undo. It calls the Undo method in the last command object
         * and moves the command to the list of PoppedCommands. The command can be redone through a call to the UndoManagers ReDo.
         * 
         * A command object's Execute and Undo methods call the necessary Refresh methods and updates of property grids.
         * A command object may save references to template properties because template properties are replaced at TreeChange.
         * It must address the property by name. It may save references to templates because they survive including their property collections.
         */
        private static Controller instance = new Controller();
        public static Controller Instance { get { return instance; } }

        public bool isSomeCopyProperties = false;

        public List<PropertyGridLine> listSelectedProperties = new List<PropertyGridLine>();

        private Controller() { }

        public bool TemplateDropped(ComponentInfo componentInfo, ContainerComponent container, Point relativePoint) {
            var command = new TemplateDrawnCommand(componentInfo, container, relativePoint, 0, 0);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public bool TemplateDrawn(ComponentInfo componentInfo, ContainerComponent container, Point relativePoint, int width, int height) {
            var command = new TemplateDrawnCommand(componentInfo, container, relativePoint, width, height);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public void Deselect() {
            SelectionBox.DeleteAll();
        }

        public bool CopyTemplates(List<MoveTransaction> transactions) {
            var command = new TemplatesCopiedCommand(transactions);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public bool DeleteTemplates(List<Template> templates) {
            if (templates.Count != 0) {
                var command = new DeleteTemplateCommand(templates);
                bool res = command.TryExecute();
                if (res)
                    UndoManager.AddCommand(command);
                return res;
            }
            return false;
        }

        public bool MoveTemplates(List<MoveTransaction> transactions) {
            var command = new TemplatesMovedCommand(transactions);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public bool ResizeTemplates(List<ResizeTransaction> transactions) {
            var command = new TemplatesResizedCommand(transactions);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public void SaveAll(bool showMessage) {
            foreach (var pair in VisEngine.Forms) {
                string frmName = pair.Key; int cloneID = 0;
                FormTemplate form = VisEngine.GetFormTemplate(frmName, ref cloneID);
                if (form != null) SaveFile(form, showMessage);
            }
        }

        public bool SaveFile(FormTemplate frm, bool showMessage) { // Returns cancel, i.e. true if form closing is abandoned.
            if (!frm.isDirty) return false;
            if (showMessage) {
                var reply = MessageBox.Show("Want to save form: " + frm.Name + "?", "Save vis-file", MessageBoxButtons.YesNoCancel);
                if (reply == DialogResult.Cancel) return true;
                if (reply == DialogResult.Yes) Save(frm);
                frm.isDirty = false; // SaveFile may be called again by VisForm.Close while closing a form bundle.
            } else {
                Save(frm); frm.isDirty = false;
            }
            return false;
        }

        private void Save(FormTemplate formTemplate) {
            // Earlier wrote templates in the sequence they were loaded, that is the sequence in formTemplate.Templates.
            // In large projects this is inconvenient. Better save templates in alphabetical sequence.
            var lines = new List<string>();
            var tplList = formTemplate.Templates.ToList();
            // Ensure that the Form is first.
            tplList.Sort((a, b) => (a.Value is FormTemplate ? "0" : a.Key).CompareTo(b.Value is FormTemplate ? "0" : b.Key));
            foreach (var t in tplList) {
                WriteTemplate(t.Value, lines);
            }
            Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
            File.WriteAllLines(formTemplate.Name + ".vis", lines);
        }

        private void WriteTemplate(Template tpl, List<string> lines) {
            var templateName = tpl.Name;
            var controlName = tpl.Info.Type.Name;

            TplProperty tplProp = tpl.valPropArray[VisComponent.NameID]; //Info.ValuePropertyByName["name"].propIndex];
            if (tpl is FormTemplate) {
                Token s = tplProp.startChain;
                if (s != null && s.next != null)
                    lines.Add(s.next.space.Trim()); // Write initial form comment.
                //else
                //    MessageBox.Show("The name property of Form " + templateName + " has no tokens. Comments will be lost.");
            } else
                lines.Add(controlName + ": " + tplProp.Formula);
            this.WriteTemplateProperties(tpl, lines);
            lines.Add("-----------------------------------");
        }

        private void WriteTemplateProperties(Template template, List<string> lines) {

            List<PropertyBoxMenu.CategoryData> categories = PropertyBoxMenu.CreateCategories(template.Info);
            foreach (TplPropValue tplProp in template.valPropArray) {
                if (tplProp.cmpPropIndex == VisComponent.NameID) continue;
                if (String.IsNullOrWhiteSpace(tplProp.Formula)) continue; // Ignore properties with empty formulas.
                if (tplProp.PropInfo.VisAttribute.ReadOnly) continue;
                string line = tplProp.Name + ": " + tplProp.Formula;
                lines.Add(line);
            }
            var orderedDesignProps = from v in template.DesignProperties orderby v.Key ascending select v;
            foreach (var valuePair in orderedDesignProps) {
                // Save also designer properties with empty formulas, which often are simple variables without a formula.
                string line = valuePair.Key + ": " + valuePair.Value.Formula;
                lines.Add(line);
            }
            foreach (TplPropEvent tplProp in template.evtPropArray) {
                if (String.IsNullOrWhiteSpace(tplProp.Formula)) continue; // Ignore properties with empty formulas.
                string line = tplProp.Name + ": " + tplProp.Formula;
                lines.Add(line);
            }
        }

        public void OpenFile(string file) {
            var command = new OpenFileCommand(file);
            command.TryExecute();
        }

        public void AddFolder(string path) {
            var directory = Path.GetDirectoryName(path);
            var files = Directory.GetFiles(directory);

            if (Path.GetExtension(path).ToUpper().Equals(".VISM")) {
                this.OpenFile(path);
                return;
            }

            foreach (var file in files) {
                if (Path.GetExtension(file).ToUpper().Equals(".VISM")) {
                    this.OpenFile(file);
                    this.OpenFile(path);
                    return;
                }
            }
        }

        public void CreateFile(string file) {
            var command = new CreateFileCommand(file);
            if (command.TryExecute()) {
                UndoManager.AddCommand(command);
                StudioBase.Explorer.UpdateExplorer();
            }
        }

        public void CopyFile(string file) {
            var command = new CopyFileCommand(file);
            if (command.TryExecute()) {
                UndoManager.AddCommand(command);
                StudioBase.Explorer.UpdateExplorer();
            }
        }

        public bool RenameFile(string file, string newNameWithExtension) {
            var command = new RenameFileCommand(file, newNameWithExtension);
            bool res = command.TryExecute();
            if (res) {
                UndoManager.AddCommand(command);
                StudioBase.Explorer.UpdateExplorer();
            }
            return res;
        }

        public void DeleteFile(string file) {
            Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
            if (File.Exists(file)) File.Delete(file);
            StudioBase.Explorer.UpdateExplorer();
        }

        public void EditFile(string file) {

        }

        public bool ChangeTemplatePropertyName(TplProperty property, string newName) {
            if (property.Name.Equals(newName)) return false;
            var command = new ChangePropertyNameCommand(property, newName);
            bool res = command.TryExecute();
            if (res) 
                UndoManager.AddCommand(command);
            return res;
        }

        public bool ChangePropertyFormula(TplProperty property, string oldFormula, string newFormula) {
            var command = new ChangePropertyCommand(property, oldFormula, newFormula);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public bool ChangeTemplateType(Template tpl, string oldFormula, string newFormula) {
            var command = new ChangeTemplateTypeCommand(tpl, oldFormula, newFormula);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }

        public bool AddTemplateProperty(Template tpl, PropInfo propInfo) {
            ICommand command = null;
            TplProperty tplProp;
            bool res;
            if (propInfo == null) {
                command = new AddPropertyCommand(tpl, "des", "");
                res = command.TryExecute();
                if (res)
                    UndoManager.AddCommand(command);
                return res;
            }
            if (propInfo is ValueInfo)
                tplProp = tpl.valPropArray[propInfo.propIndex];
            else
                tplProp = tpl.evtPropArray[propInfo.propIndex];
            if (tplProp.Formula == null) { // No formula yet.
                String value = propInfo.VisAttribute.DefaultFormula == null ? "" : propInfo.VisAttribute.DefaultFormula;
                command = new AddPropertyCommand(tpl, propInfo.Name, value);
                res = command.TryExecute();
                if (res)
                    UndoManager.AddCommand(command);
                return res;
            }
            return false;
        }

        public bool RemoveTemplateProperty(Template template, string name) {
            var command = new RemovePropertyCommand(template, name);
            bool res = command.TryExecute();
            if (res)
                UndoManager.AddCommand(command);
            return res;
        }
    }
}
