﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using VisLibrary;
using VisTool;
using Brushes = System.Windows.Media.Brushes;
using KeyEventArgs = System.Windows.Input.KeyEventArgs;
using UserControl = System.Windows.Controls.UserControl;
using System.IO;
using System.Linq;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Threading;
using System.Text.RegularExpressions;

namespace WinStudio.VisPropertyBox {

    public partial class PropertyLine : UserControl {
        public PropertyLine(PropertyBoxTemplate propertyBox, Property property)
            : this() {
            this.PropertyBox = propertyBox;
            this.UpdateData(property);
        }

        // The PropertyLine contains a name part (a text box), a formula part (RichTextBox) and an instance value (a read-only textbox).
        // When editing starts, the RichTextBox is set to the kernels property formula. Editing takes place in the RichTextBox.
        // At commit, focus lost, etc., the RichTextBox is saved in the kernel's property formula. 
        // When the popup (Intellisense) is open, the formula part before the curser is handled by GetSuggestions, which temporarily 
        // saves it as a kernel formula, compiles it as a regular formula, creates the suggestions, and restores the kernel's formula.

        public int cutLeft = 0, cutRight = 0; // Number of ident characters before and after cursor:
        public string formulaBeforeCursor; // Set when formula editor got focus or when text changed
        public string formulaAfterCursor; // Set when formula editor got focus or when text changed

        private bool popUpGettingFocus = false; // True while popUp opens and the rich text box loses its focus.
        private bool popUpWhenFocus = VisEngine.intellisensePopUpInitially; // The popUp closes when losing focus. May open when formula editor gets focus.
        private Controller controller = Controller.Instance;
        public PropertyBoxTemplate PropertyBox;
        public Property Property;
        PropertyDialog propertyDialog; // The large formula dialog box

        private List<TagToken> tagTokens = new List<TagToken>(); // Error tokens to show in the rich text boxes.

        public PropertyLine() {
            this.InitializeComponent();
            this.Background = ColorConstants.Normal();
            this.uniformGrid.Background = ColorConstants.Normal();

            this.formulaEditor.TextChanged += formulaEditor_TextChanged; // Formula editor is created by xaml.
            this.formulaEditor.KeyUp += formulaEditor_KeyUp;
            this.formulaEditor.GotKeyboardFocus += new KeyboardFocusChangedEventHandler(formulaEditor_GotKeyboardFocus);
            this.formulaEditor.LostKeyboardFocus += new KeyboardFocusChangedEventHandler(formulaEditor_LostKeyboardFocus);
            // Why are these events handled differently than the name editor below? SL 11-03-13

            this.autoListbox.KeyUp += autoListbox_KeyUp; // Autolistbox is created bu xaml.
            this.autoListbox.GotFocus += autoListbox_GotFocus;
            this.autoListbox.LostFocus += autoListbox_LostFocus;
            this.autoListbox.PreviewMouseDown += new MouseButtonEventHandler(autoListbox_PreviewMouseDown);
            this.autoListbox.MouseUp += new MouseButtonEventHandler(autoListbox_MouseUp);

            this.nameEditor.KeyUp += nameEditor_KeyUp;
            this.nameEditor.LostFocus += nameEditor_LostFocus;
            this.MouseDoubleClick += PropertyMouseDoubleClick;
        }

        //void formulaEditor_PreviewKeyDown(object sender, KeyEventArgs e) {
        //if (e.Key == Key.V && Keyboard.Modifiers == ModifierKeys.Control && PropertyBox.isSomeCopyProperties) {
        //    e.Handled = true;

        //    if (PropertyBox.listSelectedProperties.Count > 0) {
        //        foreach (PropertyGUI pg in PropertyBox.listSelectedProperties) {
        //            this.controller.CopyPasteTemplateProperty(this.Property.Template, pg.Property.TemplateProperty);
        //        }
        //        PropertyBox.CreateProperties(this.Property.Template);
        //    } else {
        //        this.controller.CopyPasteTemplateProperty(this.Property.Template, this.Property.TemplateProperty);
        //        PropertyBox.CreateProperties(this.Property.Template);
        //    }
        //    PropertyBox.listSelectedProperties.Clear();
        //}
        //}

        void autoListbox_PreviewMouseDown(object sender, MouseButtonEventArgs e) {
            //debug.Print("autoListBox_PreviewMouseUp");
            if (!this.autoListbox.IsFocused)
                this.popUpGettingFocus = true; // Will get focus
        }

        void autoListbox_MouseUp(object sender, MouseButtonEventArgs e) {
            //debug.Print("autoListBox_MouseUp");
            insertSuggestion(sender, Key.A); //Key.A is not used in this case.
        }

        public void UpdateData(Property property) {
            this.Property = property;
            this.nameEditor.Text = (this.Property is NameProperty) ? "Name" : (this.Property is TypeProperty) ? "Type" : property.GetGUIName();
            string formula = (this.Property is TypeProperty) ? property.GetGUIType() : property.GetFormula();
            this.SetRichTextFormula(formula, -1); 

            this.instanceValue.Text = property.GetVisComponentValue();
            this.instanceValue.TextWrapping = TextWrapping.NoWrap;
            this.HighlightErrors();

            this.nameEditor.IsReadOnly = false;
            this.formulaEditor.IsReadOnly = false;
            var visAttribute = new VisPropAtt(VisPropAtt.PropertyType.Property, "");

            var component = this.Property.VisComponent;
            if (component != null) {
                var componentType = component.GetType().Name;
                var componentInfo = TypeInfo.ComponentByName[componentType];
                var propertyName = property.GetKernelName();
                if (componentInfo.PropertyByName.ContainsKey(propertyName))
                    visAttribute = componentInfo.PropertyByName[propertyName].VisAttribute;
            }
            this.instanceValue.ToolTip = this.instanceValue.Text;
            this.formulaEditor.ToolTip = visAttribute.Tooltip;
            this.nameEditor.ToolTip = visAttribute.Tooltip;
            if (visAttribute.ReadOnly) {
                this.instanceValue.IsReadOnly = true;
                this.nameEditor.IsReadOnly = true;
            }
        }

        public void Select(Error error) {
            foreach (var tagToken in this.tagTokens) {
                if (tagToken.TokenError.error.Equals(error)) {
                    var range = tagToken.TextRange;
                    this.formulaEditor.CaretPosition = range.End;
                    this.formulaEditor.UpdateLayout();
                    this.formulaEditor.Focus();
                }
            }
        }

        public void Commit(bool refresh) { // Put the document text into the template formula. Make it undoable.
            string newFormula = this.GetRichText(); // Get the rich text as a simple string
            //Debug.Print("Commit, old: " + this.Property.TemplateProperty.Formula + " new: " + newFormula);
            if (this.Property is TypeProperty) {
                if (this.Property.GetGUIType() == newFormula)
                    return;
                controller.ChangeTemplateType(Property.Template, Property.GetGUIType(), newFormula);
            } else
                if (this.Property.TemplateProperty.Formula == newFormula) return;
                    controller.ChangePropertyFormula(Property.TemplateProperty, this.Property.TemplateProperty.Formula, newFormula);
                UndoManager.Refresh(Property.TemplateProperty.tpl);
                this.HighlightErrors();
            //if (refresh) { // Designer may want refresh even if nothing has changed, e.g. after Undo.
        }

        public void SetNameEditor(string name) {
            this.nameEditor.Text = name;
        }

        private string GetRichText() { // Gets the rich text as a simple string
            var text = new TextRange(formulaEditor.Document.ContentStart, formulaEditor.Document.ContentEnd).Text;
            return text.Replace("\r\n", "");
        }

        public void SetRichTextFormula(string formula, int cursorPos) { // Cursor pos < 0 means keep the old position.
            TextPointer cursor;
            if (cursorPos < 0 && this.formulaEditor.Document != null) { // Get the old cursor position.
                cursor = this.formulaEditor.CaretPosition;
                cursorPos = cursor.DocumentStart.GetOffsetToPosition(cursor)-1; // Inconsistent character indexes.            
            }
            var document = new FlowDocument();
            var paragraph = new Paragraph();
            paragraph.Inlines.Add(formula.Replace("\r\n", ""));
            document.Blocks.Clear();
            document.Blocks.Add(paragraph);
            this.formulaEditor.Document = document; // Calls FormulaEditorTextChanged, which should not call ShowPopup yet.

            Paragraph p = (Paragraph)this.formulaEditor.Document.Blocks.FirstBlock;
            try {
                cursor = p.ContentStart;
                this.formulaEditor.CaretPosition = cursor.GetPositionAtOffset(cursorPos);
            }
            catch {
                this.formulaEditor.CaretPosition = this.formulaEditor.Document.ContentEnd;
            }
        }

        private void PropertyMouseDoubleClick(object sender, MouseButtonEventArgs e) {
            if (propertyDialog != null) propertyDialog.Close();
            propertyDialog = new PropertyDialog(this);
            propertyDialog.Show();
        }

        private void nameEditor_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("NameEditorKeyUp: " + e.Key.ToString());
            if (e.Key == Key.Enter) { // Enter. Commit change.
                controller.ChangeTemplatePropertyName(Property.TemplateProperty, this.nameEditor.Text);
            }
        }

        private void nameEditor_LostFocus(object sender, RoutedEventArgs e) {
            //debug.Print("nameEditor_LostFocus");
            controller.ChangeTemplatePropertyName(Property.TemplateProperty, this.nameEditor.Text);
        }

        void formulaEditor_GotKeyboardFocus(object sender, KeyboardFocusChangedEventArgs e) {
            Debug.Print("formulaEditor_GotKeyboardFocus. popUpWhenFocus: " + popUpWhenFocus.ToString());
            StudioBase.formulaEditorFocus = true;
            if (this.popUpWhenFocus) this.ShowPopup(formulaBeforeCursor);

            // Show that property is selected: See comments above
        }

        void formulaEditor_LostKeyboardFocus(object sender, KeyboardFocusChangedEventArgs e) {
            Debug.Print("formulaEditor_LostKeyboardFocus " + (this.popUpGettingFocus ? "ignore" : ""));
            if (this.popUpGettingFocus) return;
            this.popup.IsOpen = false;
            Commit(true); // Checks whether the formula has changed
            StudioBase.formulaEditorFocus = false; // If done before commit, the application gets focus during Refresh.
        }

        private void formulaEditor_TextChanged(object sender, System.Windows.Controls.TextChangedEventArgs e) {
            //debug.Print("FormulaEditorTextChanged");
            // Will be called at any visible change to the text. Also when the property grid is initialized.
            if (!this.popup.IsOpen) return;
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text.Replace("\r\n", "");
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text.Replace("\r\n", "");
            this.ShowPopup(formulaBeforeCursor);
        }

        private void formulaEditor_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("FormulaEditorKeyUp: " + e.Key.ToString());
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text.Replace("\r\n", ""); // Multiline formula? SL 11-03-13
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text.Replace("\r\n", "");
            //}
            if (e.Key == Key.Enter && Keyboard.Modifiers == ModifierKeys.Shift) { // Shift-Enter: Line break
                var caretPosition = this.formulaEditor.CaretPosition.InsertLineBreak();
                this.formulaEditor.CaretPosition = caretPosition;
            } else if (e.Key == Key.C && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-C
                PropertyBox.isSomeCopyProperties = true;
            } else if (e.Key == Key.V && Keyboard.Modifiers == ModifierKeys.Control && PropertyBox.isSomeCopyProperties) { // Ctrl-V
                e.Handled = true;
                if (PropertyBox.listSelectedProperties.Count > 0) {
                    foreach (PropertyLine pg in PropertyBox.listSelectedProperties) {
                        this.controller.CopyPasteTemplateProperty(this.Property.Template, pg.Property.TemplateProperty);
                    }
                    PropertyBox.CreateProperties(this.Property.Template);
                    PropertyBox.listSelectedProperties.Clear();
                } else {
                    this.controller.CopyPasteTemplateProperty(this.Property.Template, this.Property.TemplateProperty);
                    PropertyBox.CreateProperties(this.Property.Template);
                }
            } else if (e.Key == Key.Z && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Z
                //Debug.Print("Rich - Ctrl-Z");
                this.Commit(false); // Save latest changes before StudioWindow makes undo.
            } else if (e.Key == Key.Y && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Y
                //Debug.Print("Rich - Ctrl-Y");
                this.Commit(false); // Save latest changes before StudioWindow makes redo.
            } else if (e.Key == Key.S && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-S. Save form.
                var formTemplate = Property.Template.FormTpl;
                this.popup.IsOpen = false;
                this.Commit(true);
                this.controller.SaveFile(formTemplate, false);
            } else if (this.popup.IsOpen && (e.Key == Key.Left || e.Key == Key.Right)) // Arrow left or right
                this.ShowPopup(formulaBeforeCursor);
            else if (e.Key == Key.Enter)  // Enter: Commit and refresh..
                this.Commit(true);
            else if (e.Key == Key.Escape) { // Esc. Close popup
                this.popUpWhenFocus = false;
                this.popup.IsOpen = false;
            } else if (e.Key == Key.Down && this.popup.IsOpen) { // Arrow down. Move into popup list
                //refreshTimer.Stop();
                this.popUpGettingFocus = true;
                if (this.autoListbox.Items.Count > 0) this.autoListbox.SelectedItem = this.autoListbox.Items[0];
                this.autoListbox.Focus();
            } else if (!this.popup.IsOpen && e.Key == Key.J && Keyboard.Modifiers == ModifierKeys.Control) { // Activate Intellisense with Ctrl+J
                this.popUpWhenFocus = true;
                this.ShowPopup(formulaBeforeCursor);
            } else if (e.Key == Key.F2) { // Open property dialog box with F2
                if (propertyDialog != null) propertyDialog.Close();
                propertyDialog = new PropertyDialog(this);
                propertyDialog.Show();
            }
        }

        private void autoListbox_GotFocus(object sender, RoutedEventArgs e) {
            //Debug.Print("AutoListboxGotFocus");
            this.popUpGettingFocus = false;
            this.popUpWhenFocus = true;
        }

        private void autoListbox_LostFocus(object sender, RoutedEventArgs e) {
            //debug.Print("AutoListboxLostFocus");
            //this.formulaEditor.Focus();
            //this.popup.IsOpen = false; // If formula editor gets focus, it will open it again.
        }

        private void autoListbox_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("AutoListboxKeyUp: " + e.Key.ToString());
            if (e.Key == Key.Enter || e.Key == Key.Tab || e.Key == Key.Space) {
                insertSuggestion(sender, e.Key);
            }
            if (Key.Escape == e.Key || Key.Back == e.Key || e.Key == Key.Tab || e.Key == Key.Space) {
                this.formulaEditor.Focus();
            }
        }

        private void insertSuggestion(object sender, Key e) {
            string newFormula = "";
            if (this.autoListbox.SelectedItem != null) {
                cutRight = cutLeft = 0;
                var formulaItem = (FormulaItem)this.autoListbox.SelectedItem;
                string item = "";
                if (this.Property is TypeProperty) { // Type is not a general formula, but a name selected from a list.
                    newFormula = formulaItem.Formula;
                    this.SetRichTextFormula(newFormula, -1); // Will call FormulaEditorTextChanged.
                } else {
                    item = formulaItem.Formula;
                    if (e == Key.Space)
                        item += " ";
                    if (pi.la.kind == Parser.tk.CutIdent) cutLeft = pi.la.val.Length;
                    newFormula = formulaBeforeCursor; // Build up the new formula. Cut away the cutIdent.
                    if (cutLeft > 0) newFormula = newFormula.Remove(newFormula.Length - cutLeft);
                    newFormula = newFormula + item; // Add the chosen suggestion.
                    if (formulaAfterCursor.Length > 0 && Char.IsLetter(formulaAfterCursor[0])) cutRight = 1; // An identifier may start right after the cursor.
                    char c;
                    while (cutRight < formulaAfterCursor.Length && (Char.IsLetterOrDigit(c = formulaAfterCursor[cutRight]) || c == '_')) cutRight++;
                    // Skip the first part of text after the cursor, any ident or the tail of a cut ident.
                    newFormula = newFormula + formulaAfterCursor.Substring(cutRight, formulaAfterCursor.Length - cutRight).Replace("\r\n", "");
                    // Include the rest of the text after the cursor.

                    int cursorPos = formulaBeforeCursor.Length - cutLeft + item.Length + 1; // Cursor position seems to count characters from 1.
                    this.popup.IsOpen = false; // Avoid calling ShowPopup until the new rich text is ready
                    this.SetRichTextFormula(newFormula, cursorPos); // Will call FormulaEditorTextChanged.
                    ShowPopup(newFormula);
                    this.formulaEditor.Focus();
                    //Commit(false); Formula is still in the rich text box.
                }
            }
        }

        private void ShowPopup(string curText) {
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text;
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text;
            //debug.Print("ShowPopup " + formulaBeforeCursor);

            if (this.Property is NameProperty) return; //No suggestions for the Name property.
            //Debug.Print("Current Formula = " + curText);
            this.popup.IsOpen = true;
            this.popup.StaysOpen = false; // Makes popup close when focus moves away from the formula editor. 
            this.popup.Width = this.formulaEditor.ActualWidth;
            this.popup.Height = this.autoListbox.Height;

            this.autoListbox.Items.Clear();
            bool usePrefix = false;
            List<Suggestion> listSuggestion = this.GetSuggestions(formulaBeforeCursor);
            for (int i = 0; i < listSuggestion.Count; i++) {
                Suggestion s = listSuggestion[i];
                System.Windows.Controls.Image icon = new System.Windows.Controls.Image();
                var iconPath = VisEngine.ProgramDirectory + "\\Icons\\IntelliIcons\\" + s.IconName + ".bmp";
                if (File.Exists(iconPath)) {
                    var bitmap = new Bitmap(iconPath);
                    icon.Source = bitmap.GetImage();
                    icon.Stretch = Stretch.UniformToFill;
                }
                FormulaItem f = new FormulaItem((usePrefix ? s.Prefix + s.uVisOperator + s.Formula : s.Formula), icon.Source, s.Description, s.Prefix, s.ToolTipText);
                this.autoListbox.Items.Add(f);
            }

            if (listSuggestion.Count == 0) {
                FormulaItem f = new FormulaItem("No match", null, "", "", "");
                this.autoListbox.Items.Add(f);
            }
            this.autoListbox.SelectedIndex = 0;
            // Don't set focus. Focus will be set when the cursor moves into the list
        }

        public void HighlightErrors() {
            var errors = Property.TemplateProperty.tokenErrors;
            var paragraph = (Paragraph)this.formulaEditor.Document.Blocks.FirstBlock;
            this.tagTokens.Clear();

            foreach (var tokenError in errors) {
                var range = FindTokenRange(paragraph.ContentStart, tokenError.t);
                if (range == null) continue;
                tagTokens.Add(new TagToken(range, tokenError));
            }
            foreach (var tagToken in this.tagTokens) {
                var range = tagToken.TextRange;
                range.ApplyPropertyValue(TextElement.BackgroundProperty, new SolidColorBrush(Colors.Orange));
                range.ApplyPropertyValue(TextElement.ForegroundProperty, new SolidColorBrush(Colors.DarkBlue));
            }
            this.UpdateLayout();
        }

        private TextRange FindTokenRange(TextPointer position, Token token) {
            var start = position.GetPositionAtOffset(token.pos + 1);
            var end = position.GetPositionAtOffset(token.pos + 1 + token.val.Length);
            if (token.kind == Parser.tk.EOL)
                end = position.GetPositionAtOffset(token.pos + 2 + token.val.Length);
            if (start == null || end == null) return null;
            var text = new TextRange(start, end).Text;
            return new TextRange(start, end);
        }

        private static PathInfo pi;

        private List<Suggestion> GetSuggestions(string curText) {
            //Special case for the Type
            if (this.Property is TypeProperty) {
                List<Suggestion> typeSuggestions = new List<Suggestion>();
                foreach (KeyValuePair<string, ComponentInfo> t in TypeInfo.ComponentByName) {
                    if ("ContainerComponent" == t.Key || "VisForm" == t.Key) continue;
                    Suggestion s = new Suggestion(t.Key, "", "", "Control Type", "Supported component types in uVis", 0, "template");
                    typeSuggestions.Add(s);
                }
                typeSuggestions.Sort(delegate(Suggestion s1, Suggestion s2) { return s1.Formula.CompareTo(s2.Formula); });
                return typeSuggestions;
            }

            // Temporarily insert a formula and compile it. Restore at the end.
            VisLibrary.ErrorLogging.ClearAllErrors(this.Property.Template.FormTpl.Name);
            string oldFormula = this.Property.TemplateProperty.Formula;
            VisEngine.SetTemplateFormula(this.Property.Template, this.Property.TemplateProperty.Name, curText);
            Parser curParser = new Parser(this.Property.Template, new StringReader(curText), this.Property.TemplateProperty, true); // Use CutID's
            curParser.Formula(this.Property.TemplateProperty);
            pi = Parser.pi;
            var result = VisTool.VisIntelliSense.getSuggestions(pi, this.Property.Template, this.Property.TemplateProperty, curText);
            VisEngine.SetTemplateFormula(this.Property.Template, this.Property.TemplateProperty.Name, oldFormula);
            return result;
        }
    }

    public class TagToken {
        public TextRange TextRange;
        public TokenError TokenError;

        public TagToken(TextRange textRange, TokenError tokenError) {
            this.TextRange = textRange;
            this.TokenError = tokenError;
        }
    }
}
