﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using VisTool;
using Brushes = System.Windows.Media.Brushes;
using KeyEventArgs = System.Windows.Input.KeyEventArgs;
using UserControl = System.Windows.Controls.UserControl;
using System.IO;
using System.Linq;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Threading;
using System.Text.RegularExpressions;

namespace uVisStudio.VisPropertyBox {

    public partial class PropertyGridLine : UserControl {
        public PropertyGridLine(PropertyBoxTemplate propertyBox, StudioProperty property)
            : this() {
            this.PropertyBox = propertyBox;
            this.UpdateData(property);
        }

        public class PopupCategory {
            public string name;
            public List<NameItem> nameItems = new List<NameItem>(40);
            public PopupCategory(string name) { this.name = name; }
        }
        public static List<PopupCategory> categories = new List<PopupCategory>(20) // The known categories in a designer-friendly sequence. SL 24-04-13
        { new PopupCategory("Operator"), new PopupCategory("Keyword"), new PopupCategory("Prefix"), new PopupCategory("Field"), 
            new PopupCategory("Relation"), new PopupCategory("Aggregate"), new PopupCategory("SQLkeyword"), 
            new PopupCategory("Common"), new PopupCategory("Special"), new PopupCategory("Designer"), 
            new PopupCategory("Template"), new PopupCategory("Form"), new PopupCategory("Table"), new PopupCategory("Enum"), 
            new PopupCategory("System") };

        /* The PropertyGridLine contains a name part (a text box), a formula part (RichTextBox) and an instance value (a read-only textbox).
        * When editing starts, the RichTextBox is set to the kernel's property formula. Editing takes place in the RichTextBox.
        * At commit, focus lost, etc., the RichTextBox is saved in the kernel's property formula (Commit). 
        * When popup (autoListBox, Intellisense) is open, the formula part before the curser is handled by GetSuggestions, which temporarily 
        * saves it as a kernel formula, compiles it as a regular formula, creates the suggestions, and restores the kernel's formula.
        * Opening, selecting and closing the autoListBox is handled in formulaEditor_PreviewKeyDown and autoListbox_KeyUp.
        */
        public string formulaBeforeCursor; // Set when formula editor got focus or when text changed.
        public Token startChainBeforeCursor; // Allows scanning the tokens in formulaBeforeCursor.
        public string formulaAfterCursor; // Set when formula editor got focus or when text changed.
        //Token cursorToken=null; // The token where the cursor is. As long as the curser token is the same, we don't need to calculate possibleEntities.
        Dictionary<string, List<Suggestion>> possibleEntities; // The list of possible entities at the cursor point.

        private bool popUpGettingFocus = false; // True while popUp opens and the rich text box loses its focus.
        private bool popUpWhenFocus = VisEngine.intellisensePopUpInitially; // The popUp closes when losing focus. May open when formula editor gets focus.
        private Controller controller = Controller.Instance;
        public PropertyBoxTemplate PropertyBox;
        public StudioProperty Property;
        PropertyDialog propertyDialog; // The large formula dialog box
        public List<TagToken> tagTokens = new List<TagToken>(); // Error tokens to show in the rich text boxes.

        public PropertyGridLine() {
            this.InitializeComponent();
            this.Background = ColorConstants.Normal();
            this.uniformGrid.Background = ColorConstants.Normal();

            this.formulaEditor.TextChanged += formulaEditor_TextChanged; // Formula editor is created by xaml.
            this.formulaEditor.KeyUp += formulaEditor_KeyUp; // KeyDown doesn't respond to Ctrl+J, etc.
            this.formulaEditor.PreviewKeyDown += formulaEditor_PreviewKeyDown;
            this.formulaEditor.GotKeyboardFocus += new KeyboardFocusChangedEventHandler(formulaEditor_GotKeyboardFocus);
            this.formulaEditor.LostKeyboardFocus += new KeyboardFocusChangedEventHandler(formulaEditor_LostKeyboardFocus);
            // Why are these events handled differently than the name editor below? SL 11-03-13

            this.autoListbox.KeyUp += autoListbox_KeyUp; // Autolistbox is created by xaml. AutoListbox.Items contains NameItems.
            this.autoListbox.GotFocus += autoListbox_GotFocus;
            this.autoListbox.LostFocus += autoListbox_LostFocus;
            this.autoListbox.PreviewMouseDown += new MouseButtonEventHandler(autoListbox_PreviewMouseDown);
            this.autoListbox.MouseUp += new MouseButtonEventHandler(autoListbox_MouseUp);

            this.GotFocus += PropertyGotFocus;
            this.nameEditor.KeyUp += nameEditor_KeyUp;
            this.nameEditor.LostFocus += nameEditor_LostFocus;
            this.MouseDoubleClick += PropertyMouseDoubleClick;
        }

        void autoListbox_PreviewMouseDown(object sender, MouseButtonEventArgs e) {
            //debug.Print("autoListBox_PreviewMouseUp");
            if (!this.autoListbox.IsFocused)
                this.popUpGettingFocus = true; // Will get focus
        }

        void autoListbox_MouseUp(object sender, MouseButtonEventArgs e) {
            //debug.Print("autoListBox_MouseUp");
            insertSuggestion(sender, Key.A); //Key.A is not used in this case.
        }

        public void UpdateData(StudioProperty property) {
            this.Property = property;
            //var icon = new System.Windows.Controls.Image(); // Doesn't show. SL 23-04-13.
            //icon.Width = 16; icon.Height = 16;
            //icon.Source = new BitmapImage(new Uri(VisEngine.ProgramDirectory + "\\Icons\\" + "field.png"));
            //this.propImage = icon; 
            VisPropAtt visAttribute = this.Property.attr;
            this.nameEditor.ToolTip = visAttribute.LongHelp == null ? visAttribute.ShortHelp : visAttribute.LongHelp;
            this.nameEditor.Text = (this.Property is NameProperty) ? "Name" : (this.Property is TypeProperty) ? "Type" : property.TemplateProperty.Name;
            this.instanceValue.Text = property.GetVisComponentValue();
            this.instanceValue.ToolTip = this.instanceValue.Text;
            this.instanceValue.TextWrapping = TextWrapping.NoWrap;
            this.instanceValue.IsReadOnly = true;

            Template tpl = this.Property.Template;
            if (visAttribute.ReadOnly) {
                this.SetRichTextFormula(visAttribute.DefaultFormula, this.Property.TemplateProperty, -1);
                this.formulaEditor.IsReadOnly = true;
                this.nameEditor.IsReadOnly = true;
                this.formulaEditor.ToolTip = "The property is read-only. You cannot type a formula.";
            } else {
                string formula = property.GetFormula(); // (this.Property is TypeProperty) ? property.GetGUIType() : property.GetFormula();
                this.SetRichTextFormula(formula, property.TemplateProperty, -1);
                this.HighlightErrors();
                this.formulaEditor.IsReadOnly = false;
                this.nameEditor.IsReadOnly = false;
                this.formulaEditor.ToolTip = "Ctrl+J:\tShow suggestions\n"+
                    "Tab:\tInsert suggestion\n"+
                    "Enter:\tInsert plus close\n" +
                    "F2, F3:\tEdit in dialog box";
            }

            //var visAttribute = new VisPropAtt(VisPropAtt.PropertyType.Property, "");
            //var component = this.Property.visComponent;
            //if (component != null) {
            //    var componentType = component.GetType().Name;
            //    var componentInfo = VisEngine.ComponentByName[componentType];
            //    var propertyName = property.GetKernelName();
            //    if (componentInfo.PropertyByName.ContainsKey(propertyName))
            //        visAttribute = componentInfo.PropertyByName[propertyName].VisAttribute;
            //}
            //this.nameEditor.ToolTip = visAttribute.LongHelp;
            //if (visAttribute.ReadOnly) {
            //    this.instanceValue.IsReadOnly = true;
            //    this.nameEditor.IsReadOnly = true;
            //}
        }

        public void Select(Error error) {
            foreach (var tagToken in this.tagTokens) {
                if (tagToken.TokenError.error.Equals(error)) {
                    var range = tagToken.TextRange;
                    this.formulaEditor.CaretPosition = range.End;
                    this.formulaEditor.UpdateLayout();
                    this.formulaEditor.Focus();
                }
            }
        }

        public void Commit(bool refresh) { // When the document text has changed, put it into the template formula, make it undoable and refresh.
            // When the document text hasn't changed, refresh when refresh is true.
            if (this.Property.attr.ReadOnly) return;
            string newFormula = this.GetRichText(); // Get the rich text as a simple string, end trimmed.
            Template tpl = this.Property.Template;
            if (tpl.FormTpl.updating)
                return; // Commit is often repeated, e.g. for Enter and LostFocus, which may be activated at Dispose in Refesh().
            TplProperty tplProp = this.Property.TemplateProperty;
            //if (tpl.FormTpl.multiClone(tpl, false)) {
            //    if (tplProp.Formula == newFormula || String.IsNullOrWhiteSpace(tplProp.Formula) && String.IsNullOrWhiteSpace(newFormula)) return;
            //    this.SetRichTextFormula(tplProp.Formula, tplProp, 0); // Reset formula before losing focus to answer MsgBox.
            //    MessageBox.Show("Multiple clones of " + tpl.Name + " are open.\nCannot edit formulas.");
            //    return;
            //}
            if (this.Property is TypeProperty) {
                newFormula = newFormula.Trim();
                if (this.Property.TemplateProperty.Formula == newFormula) return;
                //if (this.Property.GetGUIType() == newFormula) return;
                controller.ChangeTemplateType(Property.Template, Property.GetGUIType(), newFormula);
                tplProp.Formula = tpl.Info.Type.Name; // Set formula to correct type name.
                //this.SetRichTextFormula(tplProp.Formula, tplProp, 0); // Reset formula.
            } else {
                if (this.Property.TemplateProperty.Formula == newFormula ||
                    String.IsNullOrWhiteSpace(tplProp.Formula) && String.IsNullOrWhiteSpace(newFormula)) { // No change.
                    if (refresh) UndoManager.Refresh(Property.Template); // OK, but which RefreshMode has been set?
                    return;
                } else if (controller.ChangePropertyFormula(Property.TemplateProperty, this.Property.TemplateProperty.Formula, newFormula)) {
                    //Debug.Print("Commit, old: " + this.Property.TemplateProperty.Formula + " new: " + newFormula);
                    return;
                    }
                this.SetRichTextFormula(tplProp.Formula, tplProp, 0); // Reset formula.
                if (refresh) UndoManager.Refresh(Property.Template);
            }
            //UndoManager.Refresh(Property.Template); ChangeProperty refreshes.
            //this.HighlightErrors();
            //if (refresh) { // Designer may want refresh even if nothing has changed, e.g. after Undo.
        }

        public void SetNameEditor(string name) {
            this.nameEditor.Text = name;
        }

        public string GetRichText() { // Gets the rich text as a simple string
            string text = new TextRange(formulaEditor.Document.ContentStart, formulaEditor.Document.ContentEnd).Text;
            return text.TrimEnd(); //" \r\n";   //.Replace("\r\n", "");
        }

        public void SetRichTextFormula(string formula, TplProperty tplProp, int cursorPos) { // Cursor pos < 0 means keep the old position.
            // tplProp is only used to check for token errors. The formula text is given by formula.
            if (formula == null) formula = "  ";
            foreach (TokenError v in tplProp.tokenErrors) {
                // If there is an error in the EOL-token and its space has few visible chars, add two spaces for error mark.
                if (v.t.EOF && v.t.space.TrimEnd().Length < 2) formula = formula.TrimEnd() + "  ";
            }
            TextPointer cursor;
            if (cursorPos < 0 && this.formulaEditor.Document != null) { // Get the old cursor position.
                cursor = this.formulaEditor.CaretPosition;
                cursorPos = cursor.DocumentStart.GetOffsetToPosition(cursor) - 1; // Inconsistent character indexes.            
            }
            var document = new FlowDocument();
            var paragraph = new Paragraph();
            paragraph.Inlines.Add(formula);
            document.Blocks.Clear();
            document.Blocks.Add(paragraph);
            this.formulaEditor.Document = document; // Calls FormulaEditorTextChanged, which should not call ShowPopup yet.

            Paragraph p = (Paragraph)this.formulaEditor.Document.Blocks.FirstBlock;
            cursor = p.ContentStart;
            TextPointer current = cursor.GetPositionAtOffset(cursorPos);
            if (current != null)
                this.formulaEditor.CaretPosition = cursor.GetPositionAtOffset(cursorPos);
            else
                this.formulaEditor.CaretPosition = this.formulaEditor.Document.ContentEnd;
        }

        private void PropertyMouseDoubleClick(object sender, MouseButtonEventArgs e) {
            //if (propertyDialog != null) propertyDialog.Close();
            //propertyDialog = new PropertyDialog(this);
            //propertyDialog.Show();
        }

        private void PropertyGotFocus(object sender, RoutedEventArgs e) {
            if (this.popup.IsOpen) return; // No reason to change select indication.
            PropertyBox.SelectedProperty = this;
            if (Keyboard.Modifiers != (ModifierKeys.Control)) {
                if (PropertyBox.listSelectedProperties.Count > 0)
                    foreach (PropertyGridLine pg in PropertyBox.listSelectedProperties)
                        pg.nameEditor.Background = Brushes.White;
                PropertyBox.listSelectedProperties.Clear();
                this.nameEditor.Background = Brushes.LightBlue;
                PropertyBox.listSelectedProperties.Add(this);
            } else {
                this.nameEditor.Background = Brushes.LightBlue;
                if (!PropertyBox.listSelectedProperties.Contains(this))
                    PropertyBox.listSelectedProperties.Add(this);
            }
            UpdateLayout();
        }

        private void nameEditor_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("NameEditorKeyUp: " + e.Key.ToString());
            if (e.Key == Key.Enter) { // Enter. Commit change.
                CommitName(Property.TemplateProperty, this.nameEditor.Text);
            }
        }

        private void nameEditor_LostFocus(object sender, RoutedEventArgs e) {
            //debug.Print("nameEditor_LostFocus");
            CommitName(Property.TemplateProperty, this.nameEditor.Text);
        }

        private void CommitName(TplProperty tplProp, string newName) { // New name is the text editor's text.
            Template tpl = tplProp.tpl;
            if (tpl.FormTpl.updating || tplProp.Name == null) // Updating or dangling tplProp.
                return; // CommitName is often repeated, e.g. for Enter and LostFocus, which may be activated at Dispose in Refesh().
            if (!controller.ChangeTemplatePropertyName(tplProp, newName))
                this.nameEditor.Text = tplProp.Name; // Reset name.
        }

        void formulaEditor_GotKeyboardFocus(object sender, KeyboardFocusChangedEventArgs e) {
            //debug.Print("formulaEditor_GotKeyboardFocus. popUpWhenFocus: " + popUpWhenFocus.ToString());
            if (this.popup.IsOpen) return;
            StudioBase.formulaEditorFocus = true;
            //if (this.popUpWhenFocus) this.ShowPopup(formulaBeforeCursor);
            // When we try pop up here, focus behaves strange when we have focused on other applications. When we again click at the formula, nothing happens.
            // We first have to click at Studio's title bar. At mouse down in the title bar, the formula gets focus and the formula popup appears. 
            // When releasing the button, the formula loses focus, but no event appears. Yet the popup disappears.
        }

        void formulaEditor_LostKeyboardFocus(object sender, KeyboardFocusChangedEventArgs e) {
            //debug.Print("formulaEditor_LostKeyboardFocus " + (this.popUpGettingFocus ? "ignore" : ""));
            if (this.popUpGettingFocus) return;
            //debug.Print("formulaEditor_LostKeyboardFocus " + this.formulaBeforeCursor);
            this.popup.IsOpen = false;
            Commit(false); // Checks whether the formula has changed
            StudioBase.formulaEditorFocus = false; // If done before commit, the application gets focus during Refresh.
        }

        private void formulaEditor_TextChanged(object sender, System.Windows.Controls.TextChangedEventArgs e) {
            //debug.Print("FormulaEditorTextChanged");
            // Will be called at any visible change to the text. Also when the property grid is initialized.
            if (!this.popup.IsOpen) return;
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text; //.Replace("\r\n", "");
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text; //.Replace("\r\n", "");
            this.ShowPopup(formulaBeforeCursor);
        }

        private void formulaEditor_PreviewKeyDown(object sender, KeyEventArgs e) {
            //debug.Print("FormulaEditor_PreviewKeyDown: " + e.Key.ToString());
            if (e.Key == Key.Down && this.popup.IsOpen) { // Arrow down. Move into popup list
                this.popUpGettingFocus = true;
                //if (this.autoListbox.Items.Count > 0) this.autoListbox.SelectedItem = this.autoListbox.Items[0];
                this.autoListbox.Focus();
            } else if (this.popup.IsOpen && (e.Key == Key.Tab)) { // Tab with open popup: Insert suggestion.
                insertSuggestion(sender, e.Key);
                e.Handled = true; // Avoid default handling, which moves focus to next.
            } else if (this.popup.IsOpen && e.Key == Key.Enter) { // Enter with open popup: Insert suggestion and close popup.
                insertSuggestion(sender, Key.Tab);
                this.popup.IsOpen = false;
                e.Handled = true; // Avoid default handling, which moves focus to next.
            }
        }

        private void formulaEditor_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("FormulaEditorKeyDown: " + e.Key.ToString() + " " + Keyboard.Modifiers.ToString());
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text; //.Replace("\r\n", ""); // Multiline formula? SL 11-03-13
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text; //.Replace("\r\n", "");
            //}
            if (e.Key == Key.Enter && Keyboard.Modifiers == ModifierKeys.Shift) { // Shift-Enter: Line break
                var caretPosition = this.formulaEditor.CaretPosition.InsertLineBreak();
                this.formulaEditor.CaretPosition = caretPosition;

            // Ctrl-C, Ctrl-V: The rich textbox handles copy and paste itself. The code below seems intended for copying and pasting entire properties.
            //} else if (e.Key == Key.C && Keyboard.Modifiers == ModifierKeys.Control) {
            //    PropertyBox.isSomeCopyProperties = true;
            //    formulaEditor.Copy();
            //} else if (e.Key == Key.V && Keyboard.Modifiers == ModifierKeys.Control && PropertyBox.isSomeCopyProperties) { // Ctrl-V
            //    e.Handled = true; // Paste has been done automatically before we get here.
            //    if (PropertyBox.listSelectedProperties.Count > 0) {
            //        foreach (PropertyGridLine pg in PropertyBox.listSelectedProperties) {
            //            //this.controller.CopyPasteTemplateProperty(this.Property.Template, pg.Property.TemplateProperty); ??
            //            pg.formulaEditor.Paste();
            //        }
            //        //PropertyBox.CreateProperties(this.Property.Template); ??
            //        //PropertyBox.listSelectedProperties.Clear(); ??
            //    } else {
            //        this.controller.CopyPasteTemplateProperty(this.Property.Template, this.Property.TemplateProperty);
            //        PropertyBox.CreateProperties(this.Property.Template);
            //    }
            } else if (e.Key == Key.Z && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Z
                //Debug.Print("Rich - Ctrl-Z");
                this.Commit(false); // Save latest changes before StudioWindow makes undo.
            } else if (e.Key == Key.Y && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-Y
                //Debug.Print("Rich - Ctrl-Y");
                this.Commit(false); // Save latest changes before StudioWindow makes redo.
            } else if (e.Key == Key.S && Keyboard.Modifiers == ModifierKeys.Control) { // Ctrl-S. Save form.
                var formTemplate = Property.Template.FormTpl;
                //debug.Print("formulaEditor_KeyUp S " + this.formulaBeforeCursor);
                this.popup.IsOpen = false;
                this.Commit(true);
                this.controller.SaveFile(formTemplate, false); // Also done by StudioWindow
            } else if (this.popup.IsOpen && (e.Key == Key.Left || e.Key == Key.Right)) // Arrow left or right
                this.ShowPopup(formulaBeforeCursor);
            else if (e.Key == Key.Enter)  // Enter: Commit and refresh..
                this.Commit(true);
            else if (e.Key == Key.Escape) { // Esc. Close popup
                //this.popUpWhenFocus = false;
                //Debug.Print("formulaEditor_Keyup Esc " + this.formulaBeforeCursor);
                this.popup.IsOpen = false;
            // if (e.Key == Key.Down && this.popup.IsOpen)  // Arrow down. Move into popup list is handled by PreviewKeyDown so that KeyUp is handled by autoListbox.
            } else if (!this.popup.IsOpen && e.Key == Key.J && Keyboard.Modifiers == ModifierKeys.Control) { // Activate Intellisense with Ctrl+J
                //this.popUpWhenFocus = true;
                this.ShowPopup(formulaBeforeCursor);
            } else if (e.Key == Key.F2) { // Open property dialog box with F2
                if (propertyDialog != null) propertyDialog.Close();
                propertyDialog = new PropertyDialog(this, 8.50);
                propertyDialog.Show();
            } else if (e.Key == Key.F3) { // Open property dialog box with F3 (large font for presentation).
                if (propertyDialog != null) propertyDialog.Close();
                propertyDialog = new PropertyDialog(this, 18.0);
                propertyDialog.Show();
            }
        }

        private void autoListbox_GotFocus(object sender, RoutedEventArgs e) {
            //debug.Print("AutoListboxGotFocus");
            this.popUpGettingFocus = false;
            //this.popUpWhenFocus = true;
        }

        private void autoListbox_LostFocus(object sender, RoutedEventArgs e) {
            //debug.Print("AutoListboxLostFocus");
            //this.formulaEditor.Focus();
            //this.popup.IsOpen = false; // If formula editor gets focus, it will open it again.
        }

        private void autoListbox_KeyUp(object sender, KeyEventArgs e) {
            //debug.Print("AutoListboxKeyUp: " + e.Key.ToString());
            if (e.Key == Key.Enter || e.Key == Key.Tab) {
                if ( !insertSuggestion(sender, e.Key)) return;
            }
            if (e.Key == Key.Enter || Key.Escape == e.Key || e.Key == Key.Tab) {
                this.formulaEditor.Focus();
                if (e.Key == Key.Enter) this.popup.IsOpen = false;
            }
        }

        private bool insertSuggestion(object sender, Key e) { // False if nothing was inserted.
            int cutLeft = 0, cutRight = 0; // Number of ident characters before and after cursor.
            string newFormula = "", newSpace = "", item = "";
            var nameItem = (NameItem)this.autoListbox.SelectedItem;
            if (nameItem != null) {
                if (!nameItem.insertPossible)
                    return false;
                if (this.Property is TypeProperty) { // Type is not a general formula, but a name selected from a list.
                    newFormula = nameItem.Name;
                    this.SetRichTextFormula(newFormula, this.Property.TemplateProperty, -1); // Will call FormulaEditorTextChanged.
                } else { // Normal property
                    string selectedDot = nameItem.dotOperator;
                    if (selectedDot == null) selectedDot = "";
                    //if (pi.dot == null) { // Beginning of path. May or may not have an ident.
                    //    if (pi.ident != null || (selectedDot != "." && selectedDot != "!")) item = selectedDot; // Or dot needed for uniqueness??
                    //} else 
                    newSpace = pi.la.space;
                    if (pi.dot != null) {
                        cutLeft += pi.dot.val.Length;
                        if (pi.dot.next.kind == Parser.TkKind.Ident) // pi.dot belongs to an identifier, not to a cut ident. 
                            item = selectedDot;
                        else
                            pi.dot.val = selectedDot;
                        // The dot-operator is after the last ident. Change it to the selected dot. Will show when the formula is generated from the token chain.
                    }

                    item += newSpace + (nameItem.PrefixNeeded ? (nameItem.Prefix + nameItem.Name) : nameItem.Name);
                    if (pi.la.cutToken) cutLeft += pi.la.val.Length + pi.la.space.Length;
                    newFormula = Parser.GetFormulaFromTokens(startChainBeforeCursor); // The dot-operator may have changed in the chain.
                    // Build up the new formula. Cut away the cutToken.
                    int cursorPos = newFormula.Length;
                    if (cutLeft > 0) newFormula = newFormula.Remove(newFormula.Length - cutLeft);
                    newFormula += item; // Add the chosen suggestion.
                    cursorPos = cursorPos - cutLeft + item.Length + 1; // Cursor position seems to count characters from 1.

                    if (formulaAfterCursor.Length > 0 && Char.IsLetter(formulaAfterCursor[0])) cutRight = 1; // An identifier may start right after the cursor.
                    char c;
                    while (cutRight < formulaAfterCursor.Length && (Char.IsLetterOrDigit(c = formulaAfterCursor[cutRight]) || c == '_')) cutRight++;
                    // Skip the first part of text after the cursor, any ident or the tail of a cut ident.
                    newFormula = newFormula + formulaAfterCursor.Substring(cutRight, formulaAfterCursor.Length - cutRight).TrimEnd(); //.Replace("\r\n", "");
                    // Include the rest of the text after the cursor. Remove blanks at end of old formula. Editor inserts them

                    //debug.Print("Insert suggestion " + this.formulaBeforeCursor);
                    this.popup.IsOpen = false; // Avoid calling ShowPopup until the new rich text is ready
                    this.SetRichTextFormula(newFormula, this.Property.TemplateProperty, cursorPos); // Will call FormulaEditorTextChanged.
                    ShowPopup(newFormula);
                    this.formulaEditor.Focus();
                    //Commit(false); Formula is still in the rich text box.
                }
                return true;
            }
            return false;
        }

        private void ShowPopup(string curText) {
            formulaBeforeCursor = new TextRange(this.formulaEditor.Document.ContentStart, this.formulaEditor.CaretPosition).Text;
            formulaAfterCursor = new TextRange(this.formulaEditor.CaretPosition, this.formulaEditor.Document.ContentEnd).Text;
            //debug.Print("ShowPopup " + formulaBeforeCursor);

            if (this.Property is NameProperty) return; // No suggestions for the Name property.
            //Debug.Print("Current Formula = " + curText);
            this.popup.IsOpen = true;
            this.popup.StaysOpen = false; // Makes popup close when focus moves away from the formula editor. 
            this.popup.Width = this.nameEditor.ActualWidth + this.formulaEditor.ActualWidth;
            this.popup.Height = this.autoListbox.Height;

            this.autoListbox.Items.Clear();
            this.BuildSuggestions(); // Compiles and generates the suggestions

            this.autoListbox.SelectedIndex = 0;
            // Don't set focus. Focus will be set when the cursor moves into the list
        }

        public void HighlightErrors() {
            if (Property.TemplateProperty == null) return;
            var errors = Property.TemplateProperty.tokenErrors;
            var paragraph = (Paragraph)this.formulaEditor.Document.Blocks.FirstBlock;
            this.tagTokens.Clear();

            foreach (var tokenError in errors) {
                var range = FindTokenRange(paragraph.ContentStart, tokenError.t);
                if (range == null) continue;
                tagTokens.Add(new TagToken(range, tokenError));
            }
            foreach (var tagToken in this.tagTokens) {
                var range = tagToken.TextRange;
                range.ApplyPropertyValue(TextElement.BackgroundProperty, new SolidColorBrush(Colors.Orange));
                range.ApplyPropertyValue(TextElement.ForegroundProperty, new SolidColorBrush(Colors.DarkBlue));
            }
            this.UpdateLayout();
        }

        private TextRange FindTokenRange(TextPointer position, Token token) {
            TextPointer start, end;
            if (token.kind == Parser.TkKind.EOL) { // Use a space plus an orange character to mark the area. EOL has val.Lenght == 0.
                // SetRichTextFormula has inserted additional spaces after the previous token wshen needed.
                    start = position.GetPositionAtOffset(token.pos - token.space.Length + 2);
                    end = position.GetPositionAtOffset(token.pos - token.space.Length + 3);
            } else {
                start = position.GetPositionAtOffset(token.pos + 1); // pos is zero-based, TextPointer is 1-based. 
                end = position.GetPositionAtOffset(token.pos + 1 + token.val.Length);
            }
            if (start == null || end == null) return null;
            var text = new TextRange(start, end).Text; // For debugging.
            return new TextRange(start, end);
        }

        private static PathInfo pi;

        private void BuildSuggestions() {
            // Compiles the formula before the cursor, gets the kernel's suggestions, selects the relevant ones and adds them to the autolist.

            //Special case for the Type property
            if (this.Property is TypeProperty) {
                if (this.Property.Template is FormTemplate) {
                    this.autoListbox.Items.Add(new NameItem(null, null, null, false, "You canot change the type of a form", false, "", "", 0));
                    return;
                }
                List<NameItem> typeSuggestions = new List<NameItem>();
                foreach (KeyValuePair<string, ComponentInfo> t in VisEngine.ComponentByName) {
                    var a = t.Value.ClassAttribute;
                    if (a != null && a.ShowInToolbox) {
                        string iconPath = VisEngine.ProgramDirectory + "\\Icons\\" + a.Icon;
                        System.Windows.Media.ImageSource image = null;
                        if (File.Exists(iconPath)) {
                            var bitmap = new Bitmap(iconPath);
                            image = bitmap.GetImage();
                        }
                        typeSuggestions.Add(new NameItem(image, null, null, false, t.Key, true, a.LongHelp == null ? a.ShortHelp : a.LongHelp, a.ShortHelp, 0));
                    }
                }
                typeSuggestions.Sort(delegate(NameItem s1, NameItem s2) { return s1.Name.CompareTo(s2.Name); });
                foreach (NameItem n in typeSuggestions) this.autoListbox.Items.Add(n);
                return;
            }

            // Temporarily insert formulaBeforeCursor as the formula and compile it. Restore at the end.
            // Optimization to consider: When cursor is still in the same ident, skip recompiling and getting SuggestedNames. They are still the same. 
            VisTool.ErrorLogging.ClearAllErrors(this.Property.Template.FormTpl.Name);
            //PropertyBox.RemoveErrors(); // remove errors from all property gridlines.

            string oldFormula = this.Property.TemplateProperty.Formula;
            if (VisEngine.CreateOrSetTemplateProperty(this.Property.Template, this.Property.TemplateProperty.Name, formulaBeforeCursor) == null)
                return; // E.g. trying to edit a clone.
            Parser curParser = new Parser(this.Property.Template, new StringReader(formulaBeforeCursor), this.Property.TemplateProperty, true); // Use cutTokens
                TplProperty tplProp = this.Property.TemplateProperty;
            VisTool.Error error = null;
            try {
                curParser.Formula(tplProp); // Compile formula until cursor.
            } catch (CutException) {};
            pi = Parser.pi;
            error = ErrorLogging.GetError(tplProp);
            if (error != null) {
                pi.state = PathState.AfterError;
                //this.HighlightErrors();
                //StudioGrid dg = PropertyBoxManager.FindStudioGrid(tplProp.tpl);
                //dg.Grid.ShowError(error);
            }

            startChainBeforeCursor = this.Property.TemplateProperty.startChain; // Will be used at insertSuggestion.
            VisIntelliSense.match = ((pi.la.cutToken && pi.la.kind == Parser.TkKind.Ident) ? pi.la.val : "").ToLower(); // Use any cut ident as a search criterion
            //if (error != null) {
            //    VisIntelliSense.possibleNames = new Dictionary<string, List<Suggestion>>();
            //    possibleEntities = new Dictionary<string, List<Suggestion>>();
            //    Suggestion mes = new Suggestion("No suggestion because of an earlier error", null, "", "", "Operator", null, null, 1, null, null);
            //    possibleEntities.Add("Error", new List<Suggestion>(1) { mes });
            //} else
            possibleEntities = VisIntelliSense.GetSuggestions(pi, this.Property.Template, this.Property.TemplateProperty, formulaBeforeCursor);
            VisEngine.CreateOrSetTemplateProperty(this.Property.Template, this.Property.TemplateProperty.Name, oldFormula);

            foreach (var c in categories) c.nameItems.Clear(); // Remove earlier nameItems.

            // Insert the suggestions starting with "match" as nameItems in the PopupCategories.
            bool found = false;
            foreach (var nameItem in possibleEntities) {
                if (nameItem.Key.StartsWith(VisIntelliSense.match)) {
                    found = true;
                    insertItems(nameItem.Value);
                }
            }
            // If none, insert the suggestions containing "match". Consider inserting also those containing "match" except the last character??
            if (!found) {
                foreach (var nameItem in possibleEntities) {
                    if (nameItem.Key.Contains(VisIntelliSense.match)) {
                        insertItems(nameItem.Value);
                    }
                }
            }

            // Compact operators into one NameItem in the autoListbox.
            int countItems = 0;
            PopupCategory p = categories[0];
            if (p.nameItems.Count != 0) { 
                countItems++;
                p.nameItems.Sort(delegate(NameItem s1, NameItem s2) {
                    return s1.pos.CompareTo(s2.pos) == 0 ? s1.Name.CompareTo(s2.Name) : s1.pos.CompareTo(s2.pos);
                });
                string t = p.nameItems[0].Name; 
                // Make a one-line list of operators.
                for (int i=1; i < p.nameItems.Count; i++) t += "  " + p.nameItems[i].Name;
                this.autoListbox.Items.Add(new NameItem(null, null, null, false, t, false, null, null, 0));
            }

            // Sort other PopupCategories and add each NameItem to the autolist.
            for (int i = 1; i < categories.Count; i++) { // Normal NameItems
                p = categories[i];
                p.nameItems.Sort(delegate(NameItem s1, NameItem s2) { 
                    return s1.pos.CompareTo(s2.pos) == 0? s1.Name.CompareTo(s2.Name) : s1.pos.CompareTo(s2.pos); });
                foreach (var f in p.nameItems) {
                    countItems++;
                    this.autoListbox.Items.Add(f);
                }
            }

            // Show "No match" when nothing is shown.
            if (countItems == 0) {
                NameItem f = new NameItem(null, null, null, false, "No suggestion", false, "", "", 0);
                this.autoListbox.Items.Add(f);
            }
        }

        void insertItems(List<Suggestion> list) {
            bool unique=true;
            bool tblPrefixNeeded = false;
            if (list.Count > 1) {
                for (int i = 1; i < list.Count; i++)
                    // Unique may remain true if all suggestions have the same prefix and dotOperator and only differ on the table prefix.
                    if (list[i].prefix != list[0].prefix || list[i].dotOperator != list[0].dotOperator) {
                        unique = false; break;
                    }
                for (int i = 0; i < list.Count; i++) {
                    if (list[i].tblPrefix == null) continue;
                    for (int j = i + 1; j < list.Count; j++)
                        if (list[j].tblPrefix!=null && list[j].tblPrefix != list[i].tblPrefix) tblPrefixNeeded = true;
                    if (tblPrefixNeeded) break;
                }
            }
           
            foreach (Suggestion s in list) {
                int categoryIndex = categories.FindIndex(delegate(PopupCategory c) { return c.name.ToLower() == s.category.ToLower(); });
                if (categoryIndex < 0) {
                    categoryIndex = categories.Count;
                    categories.Add(new PopupCategory(s.name.ToLower()));
                }
                //for (int i = 0; i < listSuggestion.Count; i++) {
                //    Suggestion s = listSuggestion[i];
                //    System.Windows.Controls.Image icon = new System.Windows.Controls.Image();
                //    var iconPath = VisEngine.ProgramDirectory + "\\Icons\\IntelliIcons\\" + s.IconName + ".bmp";
                //    if (File.Exists(iconPath)) {
                //        var bitmap = new Bitmap(iconPath);
                //        icon.Source = bitmap.GetImage();
                //        icon.Stretch = Stretch.UniformToFill;
                //    }
                //    FormulaItem f = new FormulaItem(s.name, icon.Source, s.description, s.prefix, s.toolTipText);

                string optional ="";
                string mandatory = "";
                if (s.tblPrefix == null || !tblPrefixNeeded) {
                    //optional = s.prefix + s.dotOperator;
                    optional = s.prefix;
                    if (pi.ident != null && !pi.ident.cutToken && pi.dot == null) 
                        // There is already a step in the path, but no dot-operator after it. The dot-operator is mandatory.
                        mandatory = s.dotOperator; 
                    else optional = optional + s.dotOperator;
                    mandatory = mandatory + s.name;
                } else {
                    optional = s.prefix + ".";
                    mandatory = s.tblPrefix + s.dotOperator + s.name;
                }
                string itemName = s.longExplanation != null && s.longExplanation.StartsWith("(") ? mandatory : ""; // E.g. show (x) as Sin(x) in the long explanation.
                var f = new NameItem(null, s.dotOperator, optional, !unique, mandatory, true, itemName + s.longExplanation, 
                    s.shortExplanation, s.position);
                categories[categoryIndex].nameItems.Add(f);
            }
        }
    }

    public class TagToken {
        public TextRange TextRange;
        public TokenError TokenError;

        public TagToken(TextRange textRange, TokenError tokenError) {
            this.TextRange = textRange;
            this.TokenError = tokenError;
        }
    }
}
